(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: symtable.ml,v 1.1.1.1 1998/04/20 11:32:11 lefessan Exp $ *)

(* To assign numbers to globals and primitives *)

open Misc
open Asttypes
open Lambda
open Emitcode
open Opcodes

(* Functions for batch linking *)

type error =
    Undefined_global of string
  | Unavailable_primitive of string

exception Error of error

(* Tables for numbering objects *)

type 'a numtable =
  { num_cnt: int;               (* The next number *)
    num_tbl: ('a, int) Tbl.t } (* The table of already numbered objects *)

let empty_numtable = { num_cnt = 0; num_tbl = Tbl.empty }

let find_numtable nt key =
  Tbl.find key nt.num_tbl

let enter_numtable nt key =
  let n = !nt.num_cnt in
  nt := { num_cnt = n + 1; num_tbl = Tbl.add key n !nt.num_tbl };
  n

let incr_numtable nt =
  let n = !nt.num_cnt in
  nt := { num_cnt = n + 1; num_tbl = !nt.num_tbl };
  n

(* Global variables *)

let global_table = ref(empty_numtable : Ident.t numtable)
and literal_table = ref([] : (int * structured_constant) list)

let names = ref []
let toplevel = ref false

let slot_for_getglobal id =
  try
    find_numtable !global_table id
  with Not_found ->
    raise(Error(Undefined_global(Ident.name id)))

let slot_for_setglobal id =
  enter_numtable global_table id

let slot_for_literal cst =
  let n = incr_numtable global_table in
  literal_table := (n, cst) :: !literal_table;
  n

(* The C primitives *)

let c_prim_table = ref(empty_numtable : string numtable)

let num_of_prim name =
  try
    find_numtable !c_prim_table name
  with Not_found ->
    if !Clflags.custom_runtime
    then enter_numtable c_prim_table name
    else raise(Error(Unavailable_primitive name))

let require_primitive name =
  if name.[0] <> '%' then begin num_of_prim name; () end

let all_primitives () =
  let prim = Array.create !c_prim_table.num_cnt "" in
  Tbl.iter (fun name number -> prim.(number) <- name) !c_prim_table.num_tbl;
  prim

let output_primitive_names outchan =
  let prim = all_primitives() in
  for i = 0 to Array.length prim - 1 do
    output_string outchan prim.(i); output_char outchan '\000'
  done

open Printf

let output_primitive_table outchan =
  let prim = all_primitives() in
  for i = 0 to Array.length prim - 1 do
    fprintf outchan "extern long %s();\n" prim.(i)
  done;
  fprintf outchan "typedef long (*primitive)();\n";
  fprintf outchan "primitive cprim[] = {\n";
  for i = 0 to Array.length prim - 1 do
    fprintf outchan "  %s,\n" prim.(i)
  done;
  fprintf outchan "  (primitive) 0 };\n";
  fprintf outchan "char * names_of_cprim[] = {\n";
  for i = 0 to Array.length prim - 1 do
    fprintf outchan "  \"%s\",\n" prim.(i)
  done;
  fprintf outchan "  (char *) 0 };\n"


let output_table outchan table null =
  let array = Array.create table.num_cnt null in
    Tbl.iter (fun name number -> array.(number) <- name) table.num_tbl;
    output_value outchan array

let input_vm_info inchan =
  Array.iter (function v -> enter_numtable c_prim_table v; ())
    (input_value inchan);
  Array.iter (function v -> enter_numtable global_table v; ())
    (input_value inchan);
  literal_table := input_value inchan;
  names := input_value inchan

let output_vm_info outchan =
  output_table outchan !c_prim_table "";
  output_table outchan !global_table (Ident.create "()");
  output_value outchan (!literal_table);
  output_value outchan (!names)

(* Initialization for batch linking *)

let init () =
  (* Enter the predefined exceptions *)
  Array.iter 
    (fun name -> 
      let id =
        try List.assoc name Predef.builtin_values
        with Not_found -> fatal_error "Symtable.init" in
      let c = slot_for_setglobal id in
      let cst = Const_block(Tag.predef_exn, [Const_base(Const_string name)]) 
      in
      literal_table := (c, cst) :: !literal_table)
    Runtimedef.builtin_exceptions;
  (* Enter the known C primitives *)
  Array.iter (fun x -> enter_numtable c_prim_table x; ())
             Runtimedef.builtin_primitives;
  names := []

(* Must use the unsafe String.set here because the block may be
   a "fake" string as returned by Meta.static_alloc. *)

let patch_int buff pos n =
  String.unsafe_set buff pos (Char.unsafe_chr n);
  String.unsafe_set buff (pos + 1) (Char.unsafe_chr (n asr 8));
  String.unsafe_set buff (pos + 2) (Char.unsafe_chr (n asr 16));
  String.unsafe_set buff (pos + 3) (Char.unsafe_chr (n asr 24))

(* Relocate a block of object bytecode *)
exception FoundName of string
let slot_for_initial_data id compunit pos crc code =
  let reloc = ref [] in
  let fields = ref [] in
  let counter = ref (-1) in (* incremented before usage *)
  let prims = ref [] in
  let modules = ref [] in
  let name = compunit.cu_name^crc in
  let _ =
    names := (id,name):: !names;
    List.iter 
      (function (reloc_info,pos) ->
         reloc :=  (Const_base (Const_int (pos/4)))::!reloc;
         match reloc_info with
           GETMETHOD m ->
             incr counter;
             fields := (Const_base (Const_string m)) :: !fields;
             Debug.print2 'c' "Relocating methode %s at %d\n" m !counter;
             patch_int code (pos+4) !counter 

         | MODULE id ->
               patch_int code (pos+4)
                 (let name = 
                    if not !toplevel then
                      try
                        List.assoc id !names
                      with
                        Not_found ->
                          raise(Error(Undefined_global(Ident.name id)))
                    else
                      begin
                    (* as we are in the toplevel, we don't know the map
                       from module idents to module names, thus we must
                       first use the map from idents to slots in the
                       global table, and then iter on the Reloc.module_htable
                       to find the corresponding module *)
                        let slot = slot_for_getglobal id in
                          try
                            Hashtbl.iter (
                              fun name module_data ->
                                if module_data.Reloc.interface = slot then
                                  raise (FoundName module_data.Reloc.module_name)
                            ) Reloc.module_htable;
                            raise(Error(Undefined_global(Ident.name id)))
                          with
                              FoundName name -> name
                      end
                  in
                    try
                      List.assoc name !modules
                    with
                        Not_found ->
                          incr counter;
                          fields := (Const_base (Const_string name))::!fields;
                          modules := (name,!counter)::!modules;
                          !counter
                 )
           | THISMODULE -> ()
(* constantes *)
           | CONSTANT sc  ->
               incr counter;
               fields := sc :: !fields;
               patch_int code (pos+4) !counter 
(* primitives *)
           | CCALL (n,prim) ->
               patch_int code 
                 (pos + 4*n)
                 (try
                    List.assoc prim !prims
                  with
                      Not_found ->
                        if !Clflags.dyn_required then
                            require_primitive prim;
                        incr counter;
                        fields := (Const_base (Const_string prim)) :: !fields;
                        prims := (prim,!counter) :: !prims;
                        !counter)
      )
      compunit.cu_dyn_reloc
  in
  let fields = (!fields) @ [
  (* exceptions *) Const_base (Const_int 0);
  (* migrate *) Const_base (Const_int
                              (if compunit.cu_migrable then 1 else 0));
  (* classes *) Const_base (Const_int 0) ;
  (* interface *) Const_base (Const_int (-1)) ;
  (* bytecode *) (Const_base (Const_string 
                                (if compunit.cu_migrable then code else "")));
  (* reloc *) (Const_block (0,!reloc));
  (* end_code *) Const_base (Const_int 0) ;
  (* begin_code *) Const_base (Const_int 0) ;
  (* longueur *) Const_base (Const_int (compunit.cu_codesize/4)) ;
  (* offset *) Const_base (Const_int (pos/4)) ; 
  (* module_slot *) Const_base (Const_int 0) ;
  (* module_name *) Const_base (Const_string name)
  ]  in
    slot_for_literal (Const_block (0,List.rev fields))

(* on doit d'abord creer initial_data *)
let patch_object buff compunit = 
  Debug.print1 'c' "Patch object %s\n" compunit.cu_name;
  let crc = Digest.string buff in
  List.iter
    (function
        (Reloc_literal sc, pos) ->
          patch_int buff pos (slot_for_literal sc)
      | (Reloc_getglobal id, pos) ->
          patch_int buff pos (slot_for_getglobal id)
      | (Reloc_setglobal id, pos) ->
          patch_int buff pos (slot_for_setglobal id)
      | (Reloc_primitive name, pos) ->
          patch_int buff pos (num_of_prim name)
      | (Reloc_initial_data id,pos) ->
          patch_int buff pos (slot_for_initial_data id compunit pos crc buff)
    ) compunit.cu_reloc

    
let toplevel_patch_object buff reloc = 
  List.iter
    (function
        (Reloc_literal sc, pos) ->
          patch_int buff pos (slot_for_literal sc)
      | (Reloc_getglobal id, pos) ->
          patch_int buff pos (slot_for_getglobal id)
      | (Reloc_setglobal id, pos) ->
          patch_int buff pos (slot_for_setglobal id)
      | (Reloc_primitive name, pos) ->
          patch_int buff pos (num_of_prim name)
      | _ -> failwith "Symtable.toplevel_patch_object: bad reloc info"
    ) reloc


(* Translate structured constants *)

let rec transl_const = function
    Const_base(Const_int i) -> Obj.repr i
  | Const_base(Const_char c) -> Obj.repr c
  | Const_base(Const_string s) -> Obj.repr s
  | Const_base(Const_float f) -> Obj.repr(float_of_string f)
  | Const_pointer i -> Obj.repr i
  | Const_block(tag, fields) ->
      let block = Obj.new_block tag (List.length fields) in
      let pos = ref 0 in
      List.iter
        (fun c -> Obj.set_field block !pos (transl_const c); incr pos)
        fields;
      block
  | Const_float_array fields ->
      transl_const
        (Const_block(0, List.map (fun f -> Const_base(Const_float f)) fields))

(* Build the initial table of globals *)

let initial_global_table () =
  let glob = Array.create !global_table.num_cnt (Obj.repr 0) in
  List.iter
    (fun (slot, cst) -> glob.(slot) <- transl_const cst)
    !literal_table;
  literal_table := [];
  glob

(* Save the table of globals *)

let output_global_map oc =
  output_value oc !global_table

(* Functions for toplevel use *)

(* Update the in-core table of globals *)

let update_global_table get_global_data realloc_global_data =
  let ng = !global_table.num_cnt in
  if ng > Array.length(get_global_data()) then realloc_global_data ng;
  let glob = get_global_data() in
  List.iter
    (fun (slot, cst) -> glob.(slot) <- transl_const cst)
    !literal_table;
  literal_table := []

(* Initialize the linker for toplevel use *)

let init_toplevel available_primitives =
  toplevel := true;
  (* Read back the known global symbols from the executable file *)
  let ic = open_in_bin Sys.argv.(0) in
  let pos_trailer =
    in_channel_length ic - 20 - String.length Config.exec_magic_number in
  seek_in ic pos_trailer;
  let code_size = input_binary_int ic in
  let prim_size = input_binary_int ic in
  let data_size = input_binary_int ic in
  let symbol_size = input_binary_int ic in
  let debug_size = input_binary_int ic in
  seek_in ic (pos_trailer - debug_size - symbol_size);
  global_table := (input_value ic : Ident.t numtable);
  close_in ic;
  (* Enter the known C primitives *)
  Array.iter (fun x -> enter_numtable c_prim_table x; ())
             (available_primitives())

(* Find the value of a global identifier *)

let get_global_position id = slot_for_getglobal id

let get_global_value get_global_data id =
  (get_global_data()).(slot_for_getglobal id)
let assign_global_value get_global_data id v =
  (get_global_data()).(slot_for_getglobal id) <- v

(* Save and restore the current state *)

type global_map = Ident.t numtable

let current_state () = !global_table

let restore_state st = global_table := st

let hide_additions st =
  if st.num_cnt > !global_table.num_cnt then
    fatal_error "Symtable.hide_additions";
  global_table :=
    { num_cnt = !global_table.num_cnt;
      num_tbl = st.num_tbl }

(* "Filter" the global map according to some predicate.
   Used to expunge the global map for the toplevel. *)

let filter_global_map p gmap =
  let newtbl = ref Tbl.empty in
  Tbl.iter
    (fun id num -> if p id then newtbl := Tbl.add id num !newtbl)
    gmap.num_tbl;
  {num_cnt = gmap.num_cnt; num_tbl = !newtbl}

(* Error report *)

open Format

let report_error = function
    Undefined_global s ->
      print_string "Reference to undefined global `"; print_string s;
      print_string "'"
  | Unavailable_primitive s ->
      print_string "The external function `"; print_string s;
      print_string "' is not available"
