(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: translmod.ml,v 1.1.1.1 1998/04/20 11:32:11 lefessan Exp $ *)

(* Translation from typed abstract syntax to lambda terms,
   for the module language *)

open Misc
open Asttypes
open Types
open Typedtree
open Lambda
open Translobj
open Translcore
open Translclass

(* defined in modules.h *)
let interface_field = 8

(* Compile a coercion *)

let rec apply_coercion restr arg =
  match restr with
    Tcoerce_none ->
      arg
  | Tcoerce_structure pos_cc_list ->
      name_lambda arg (fun id ->
        Lprim(Pmakeblock(0, Immutable),
              List.map (apply_coercion_field id) pos_cc_list))
  | Tcoerce_functor(cc_arg, cc_res) ->
      let param = Ident.create "funarg" 
      and module_data' = Ident.create "thismodule" in
      name_lambda arg (fun id ->
        Lfunction(Curried, [param;module_data'],
          apply_coercion cc_res
            (Lapply(Lvar id, [apply_coercion cc_arg (Lvar param);
                              Lvar module_data']))))
  | Tcoerce_primitive p ->
      fatal_error "Translmod.apply_coercion"

and apply_coercion_field id (pos, cc) =
  match cc with
    Tcoerce_primitive p -> transl_primitive p
  | _ -> apply_coercion cc (Lprim(Pfield pos, [Lvar id]))

(* Compose two coercions
   apply_coercion c1 (apply_coercion c2 e) behaves like
   apply_coercion (compose_coercions c1 c2) e. *)

let rec compose_coercions c1 c2 =
  match (c1, c2) with
    (Tcoerce_none, c2) -> c2
  | (c1, Tcoerce_none) -> c1
  | (Tcoerce_structure pc1, Tcoerce_structure pc2) ->
      let v2 = Array.of_list pc2 in
      Tcoerce_structure
        (List.map (fun (p1, c1) ->
                let (p2, c2) = v2.(p1) in (p2, compose_coercions c1 c2))
             pc1)
  | (Tcoerce_functor(arg1, res1), Tcoerce_functor(arg2, res2)) ->
      Tcoerce_functor(compose_coercions arg2 arg1,
                      compose_coercions res1 res2)
  | (_, _) ->
      fatal_error "Translmod.compose_coercions"

(* Record the primitive declarations occuring in the module compiled *)

let primitive_declarations = ref ([] : string list)

(* Compile a module expression *)

let rec transl_module module_data cc mexp =
  match mexp.mod_desc with
    Tmod_ident path ->
      apply_coercion cc (transl_path path)
  | Tmod_structure str ->
      transl_structure module_data 0 [] cc str
  | Tmod_functor(param, mty, body) ->
      let module_data' = Ident.create "thismodule" in
      begin match cc with
        Tcoerce_none ->
          Lfunction(Curried, [param;module_data'], 
                    transl_module (Lvar module_data') Tcoerce_none body)
      | Tcoerce_functor(ccarg, ccres) ->
          let param' = Ident.create "funarg" in
          Lfunction(Curried, [param';module_data'],
            Llet(Alias, param, apply_coercion ccarg (Lvar param'),
              transl_module (Lvar module_data') ccres body))
      | _ ->
          fatal_error "Translmod.transl_module"
      end
  | Tmod_apply(funct, arg, ccarg) ->
      apply_coercion cc
        (Lapply(transl_module module_data Tcoerce_none funct, 
                [transl_module module_data ccarg arg; module_data]))
  | Tmod_constraint(arg, mty, ccarg) ->
      transl_module module_data (compose_coercions cc ccarg) arg

and transl_structure module_data tag fields cc = function
    [] ->
      begin match cc with
        Tcoerce_none ->
          Lprim(Pmakeblock(tag, Immutable),
                List.map (fun id -> Lvar id) (List.rev fields))
      | Tcoerce_structure pos_cc_list ->
          let v = Array.of_list (List.rev fields) in
          Lprim(Pmakeblock(tag, Immutable),
                List.map
                  (fun (pos, cc) ->
                    match cc with
                      Tcoerce_primitive p -> transl_primitive p
                    | _ -> apply_coercion cc (Lvar v.(pos)))
                  pos_cc_list)
      | _ ->
          fatal_error "Translmod.transl_structure"
      end
  | Tstr_eval expr :: rem ->
      Lsequence(transl_exp expr, 
                transl_structure module_data tag fields cc rem)
  | Tstr_value(rec_flag, pat_expr_list) :: rem ->
      let ext_fields = rev_let_bound_idents pat_expr_list @ fields in
      transl_let rec_flag pat_expr_list 
        (transl_structure module_data tag ext_fields cc rem)
  | Tstr_primitive(id, descr) :: rem ->
      begin match descr.val_kind with
        Val_prim p -> primitive_declarations :=
                        p.Primitive.prim_name :: !primitive_declarations
      | _ -> ()
      end;
      transl_structure module_data tag fields cc rem
  | Tstr_type(decls) :: rem ->
      transl_structure module_data tag fields cc rem
  | Tstr_exception(id, decl) :: rem ->
      Llet(Strict, id, transl_exception module_data id decl,
           transl_structure module_data tag (id :: fields) cc rem)
  | Tstr_module(id, modl) :: rem ->
      Llet(Strict, id, transl_module module_data Tcoerce_none modl,
           transl_structure module_data tag (id :: fields) cc rem)
  | Tstr_modtype(id, decl) :: rem ->
      transl_structure module_data tag fields cc rem
  | Tstr_open path :: rem ->
      transl_structure module_data tag fields cc rem
  | Tstr_class cl_list :: rem ->
      List.fold_right
        (fun (id, cl) re ->
           Llet(Strict, id, class_stub, re))
        cl_list
        (List.fold_right
           (fun (id, cl) re ->
              Lsequence(transl_class id cl module_data, re))
           cl_list
         (transl_structure module_data tag
            ((List.rev (List.map fst cl_list)) @ fields) cc rem))

(* Compile an implementation *)

let transl_implementation module_name str cc =
  reset_labels ();
  primitive_declarations := [];
  let module_id = Ident.create_persistent module_name in
    Lsequence(Lbeginmodule module_id,
              Lprim(Psetmodule module_id, 
                    [transl_label_init
                          (transl_structure 
                             Lthismodule Tag.interface [] cc str)]))

(* A variant of transl_structure used to compile toplevel structure definitions
   for the native-code compiler. Store the defined values in the fields
   of the global as soon as they are defined, in order to reduce register
   pressure.
   "map" is a table from idents to (position in global block, coercion).
   "prim" is a list of (position in global block, primitive declaration). *)
let asm_prim name =
  try
    transl_path
      (fst (Env.lookup_value (Longident.Ldot (Longident.Lident "Asm", name))
              Env.empty))
  with Not_found ->
    fatal_error ("Primitive Asm." ^ name ^ " not found.")

let transl_store_structure module_name module_data glob map prims str =
  let rec transl_store = function
    [] ->
      if !Clflags.mixte then
        Lapply(asm_prim "export_module",
               [Lconst (Const_base(Const_string module_name));
                 Lbeginmodule glob])
      else
        lambda_unit
  | Tstr_eval expr :: rem ->
      Lsequence(transl_exp expr, transl_store rem)
  | Tstr_value(rec_flag, pat_expr_list) :: rem ->
      transl_let rec_flag pat_expr_list
        (store_idents glob map (let_bound_idents pat_expr_list)
          (transl_store rem))
  | Tstr_primitive(id, descr) :: rem ->
      begin match descr.val_kind with
        Val_prim p -> primitive_declarations :=
                        p.Primitive.prim_name :: !primitive_declarations
      | _ -> ()
      end;
      transl_store rem
  | Tstr_type(decls) :: rem ->
      transl_store rem
  | Tstr_exception(id, decl) :: rem ->
      Llet(Strict, id, transl_exception module_data id decl,
           store_ident glob map id (transl_store rem))
  | Tstr_module(id, modl) :: rem ->
      Llet(Strict, id, transl_module module_data Tcoerce_none modl,
           store_ident glob map id (transl_store rem))
  | Tstr_modtype(id, decl) :: rem ->
      transl_store rem
  | Tstr_open path :: rem ->
      transl_store rem
  | Tstr_class cl_list :: rem ->
      List.fold_right
        (fun (id, cl) re ->
           Llet(Strict, id, class_stub, re))
        cl_list
        (List.fold_right
           (fun (id, cl) re ->
              Lsequence(transl_class id cl module_data, re))
           cl_list
           (store_idents glob map (List.map fst cl_list) (transl_store rem)))

  and store_ident glob map id cont =
    try
      let (pos, cc) = Ident.find_same id map in
      let init_val = apply_coercion cc (Lvar id) in
      Lsequence
       (Lprim(Psetfield(pos, false), [Lprim(Pgetglobal glob, []); init_val]),
        cont)
    with Not_found ->
      cont

  and store_idents glob map idlist cont =
    List.fold_right (store_ident glob map) idlist cont

  and store_primitive (pos, prim) cont =
    Lsequence(Lprim(Psetfield(pos, false),
                    [Lprim(Pgetglobal glob, []); transl_primitive prim]),
              cont)
  in
    List.fold_right store_primitive prims (transl_store str)

(* Build the list of value identifiers defined by a toplevel structure *)

let rec defined_idents = function
    [] -> []
  | Tstr_eval expr :: rem -> defined_idents rem
  | Tstr_value(rec_flag, pat_expr_list) :: rem ->
      let_bound_idents pat_expr_list @ defined_idents rem
  | Tstr_primitive(id, descr) :: rem -> defined_idents rem
  | Tstr_type decls :: rem -> defined_idents rem
  | Tstr_exception(id, decl) :: rem -> id :: defined_idents rem
  | Tstr_module(id, modl) :: rem -> id :: defined_idents rem
  | Tstr_modtype(id, decl) :: rem -> defined_idents rem
  | Tstr_open path :: rem -> defined_idents rem
  | Tstr_class cl_list :: rem ->
      List.map fst cl_list @ defined_idents rem

(* Transform a coercion and the list of value identifiers built above
   into a table id -> (pos, coercion), with [pos] being the position
   in the global block where the value of [id] must be stored,
   and [coercion] the coercion to be applied to it.
   A given identifier may appear several times
   in the coercion (if it occurs several times in the signature); remember
   to assign it the position of its last occurrence.
   Also buid a list of primitives and their positions in the global block,
   and the total size of the global block. *)

let build_ident_map restr idlist =
  match restr with
    Tcoerce_none ->
      let rec build_map pos map = function
        [] ->
          (map, [], pos)
      | id :: rem ->
          build_map (pos+1) (Ident.add id (pos, Tcoerce_none) map) rem
      in build_map 0 Ident.empty idlist
  | Tcoerce_structure pos_cc_list ->
      let idarray = Array.of_list idlist in
      let rec build_map pos map prims = function
        [] ->
          (map, prims, pos)
      | (source_pos, Tcoerce_primitive p) :: rem ->
          build_map (pos+1) map ((pos, p) :: prims) rem
      | (source_pos, cc) :: rem ->
          build_map (pos+1) (Ident.add idarray.(source_pos) (pos, cc) map)
                    prims rem
      in build_map 0 Ident.empty [] pos_cc_list
  | _ ->
      fatal_error "Translmod.build_ident_map"
        
(* Compile an implementation using transl_store_structure 
   (for the native-code compiler). *)

let transl_store_implementation module_name str restr =
  reset_labels ();
  primitive_declarations := [];
  let module_id = Ident.create_persistent module_name in
  let (map, prims, size) = build_ident_map restr (defined_idents str) in
  (size, transl_label_init
     (transl_store_structure module_name Lthismodule module_id map prims str))

(* Compile a sequence of expressions *)

let rec make_sequence fn = function
    [] -> lambda_unit
  | [x] -> fn x
  | x::rem ->
      let lam = fn x in Lsequence(lam, make_sequence fn rem)

(* Compile a toplevel phrase *)

let transl_toplevel_item set_toplevel_value str = 
  let rec transl = function
      [] -> lambda_unit
    | [Tstr_eval expr] ->
        transl_exp expr
    | (Tstr_eval expr)::rem ->
        Lsequence(transl_exp expr,
                  transl rem)
    | (Tstr_value(rec_flag, pat_expr_list))::rem ->
        let idents = let_bound_idents pat_expr_list in
        let lam =
          transl_let rec_flag pat_expr_list
            (Lsequence
               (make_sequence (fun id -> set_toplevel_value id (Lvar id))
                  idents,
                transl rem
                  ) 
               )
        in
        lam
  | (Tstr_primitive(id, descr))::rem  -> transl rem
  | (Tstr_type(decls))::rem  -> transl rem
  | (Tstr_exception(id, decl))::rem  ->
      Llet(Strict, id, transl_exception Lthismodule id decl,
           Lsequence(set_toplevel_value id (Lvar id),
                     transl rem))
  | (Tstr_module(id, modl))::rem  ->
      Llet(Strict, id, transl_module Lthismodule Tcoerce_none modl,
           Lsequence(set_toplevel_value id (Lvar id),
                     transl rem))
  | (Tstr_modtype(id, decl))::rem  -> transl rem
  | (Tstr_open path)::rem  -> transl rem
  | (Tstr_class cl_list)::rem  ->
      List.fold_right
        (fun (id, cl) re ->
          Llet(Strict, id, class_stub, re))
        cl_list
        (List.fold_right
           (fun (id, cl) re ->
             Lsequence(transl_class id cl Lthismodule, re)
               )
           cl_list
           (Lsequence
              (make_sequence
                 (fun (id, cl) -> set_toplevel_value id (Lvar id)) cl_list,
               transl rem)
              )
           )

  in
  transl str

(* Mixing between bytecode en native code *)
      
let mixte_interface module_name =
  reset_labels ();
  primitive_declarations := [];
  let module_id = Ident.create_persistent module_name in
  Lsequence(Lbeginmodule module_id,
            Lprim(Psetmodule module_id, 
                  [Lapply (asm_prim "import_module",    
                           [Lconst (Const_base(Const_string module_name))]
                             )]))

let module_num = ref 0
let transl_toplevel_definition toplevel_id get_toplevel_slot str =
  let set_toplevel_value  id value =
    Lprim(Parraysetu Paddrarray,
          [Lvar toplevel_id;
            Lconst(Const_base(Const_int (get_toplevel_slot id)));
            value])
  in
  reset_labels ();
  if !Clflags.join then
    begin
      incr module_num;
      let module_id = Ident.create_persistent 
          ("toplevel"^(string_of_int !module_num)) in
      Lsequence(Lbeginmodule module_id,
                transl_label_init
                  (transl_toplevel_item set_toplevel_value str)
                  )
    end
  else
    transl_label_init 
      (transl_toplevel_item set_toplevel_value str)

(* L'ideal serait de ne plus utiliser d'identificateurs globaux. Pour
cela, on compile la phrase dans un environnement de fonction dit
TOPLEVEL, qui est cree suivant necessite par le toplevel lors de
l'appel de la fonction.

Les valeurs des identificateurs crees sont stockes dans une table speciale
nommee Toplevel_table.

La fonction cree est sous la forme 
 (fun Toplevel_table ->
    let ...
    in
      Toplevel_table.(i+0) <- id0;
      Toplevel_table.(i+1) <- id1;
      Toplevel_table.(i+2) <- id2)
   
*)
