/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: fail.c,v 1.1.1.1 1998/04/20 11:32:11 lefessan Exp $ */

/* Raising exceptions from C. */

/* Gros probleme si on melange code natif et bytecode:
   Quelle exception lancer si on est dans une fonction C ?
   Les exceptions natives doivent-elles etre catchables depuis le bytecode, 
   et reciproquement ?

 */

#include "alloc.h"
#include "fail.h"
#include "io.h"
#include "gc.h"
#include "memory.h"
#include "mlvalues.h"
#include "signals.h"
#include "stacks.h"
#include "common.h"


struct longjmp_buffer * external_raise;
value exn_bucket;

void mlraise(value v)
{
  if (in_bytecode){
    Assert(! async_signal_mode);
    Unlock_exn();
    exn_bucket = v;
    siglongjmp(external_raise->buf, 1);
  } else
    asm_mlraise(v);
}

void raise_constant(value tag)
{
  value bucket;
  Begin_root (tag);
    bucket = alloc (1, 0);
    Field(bucket, 0) = tag;
  End_roots ();
  mlraise(bucket);
}

void raise_with_arg(value tag, value arg)
{
  value bucket;
  Begin_roots2 (tag, arg);
    bucket = alloc (2, 0);
    Field(bucket, 0) = tag;
    Field(bucket, 1) = arg;
  End_roots ();
  mlraise(bucket);
}

void raise_with_2args(value tag, value arg1,value arg2)
{
  value bucket;
  Begin_roots3 (tag, arg1, arg2);
    bucket = alloc (3, 0);
    Field(bucket, 0) = tag;
    Field(bucket, 1) = arg1;
    Field(bucket, 2) = arg2;
  End_roots ();
  mlraise(bucket);
}

void raise_with_string(value tag, char *msg)
{
  value vmsg;
  Begin_root(tag);
    vmsg = copy_string(msg);
  End_roots();
  raise_with_arg(tag, vmsg);
}

void failwith (char *msg)
{
  if(in_bytecode)
    raise_with_string(Field(global_data, FAILURE_EXN), msg);
  else
    raise_with_string((value) Failure, msg);
}

void invalid_argument (char *msg)
{
  if(in_bytecode)
    raise_with_string(Field(global_data, INVALID_EXN), msg);
  else
    fatal_error_arg("Fatal_error: Invalid_argument \"%s\"\n", msg);
}

/* To raise Out_of_memory, we can't use raise_constant,
   because it allocates and we're out of memory...
   We therefore build the bucket by hand.
   This works OK because the exception value for Out_of_memory is also
   statically allocated out of the heap. */

static struct {
  header_t hdr;
  value exn;
} out_of_memory_bucket;

void raise_out_of_memory(void)
{
  out_of_memory_bucket.hdr = Make_header(1, 0, White);
  if(in_bytecode)
    out_of_memory_bucket.exn = Field(global_data, OUT_OF_MEMORY_EXN);
  else
    out_of_memory_bucket.exn = (value) Out_of_memory;
  mlraise((value) &(out_of_memory_bucket.exn));
}

void raise_stack_overflow(void)
{
  raise_constant(Field(global_data, STACK_OVERFLOW_EXN));
}

void raise_sys_error(value error,value msg)
{
  if(in_bytecode)
    raise_with_2args(Field(global_data, SYS_ERROR_EXN), error, msg);
  else
    raise_with_2args((value) Sys_error, error, msg);
}

void raise_end_of_file(void)
{
  if(in_bytecode)
    raise_constant(Field(global_data, END_OF_FILE_EXN));
  else
    raise_constant((value) End_of_file);
}

void raise_zero_divide(void)
{
  if(in_bytecode)
    raise_constant(Field(global_data, ZERO_DIVIDE_EXN));
  else
    raise_constant((value) Division_by_zero);
}

void raise_not_found(void)
{
  if(in_bytecode)
    raise_constant(Field(global_data, NOT_FOUND_EXN));
  else
    raise_constant((value) Not_found);
}

