/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*             Fabrice Le Fessant & Natacha Duval                      */
/*                projet Para, INRIA Rocquencourt                      */
/*                                                                     */
/*  Copyright 1997 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: weak.c,v 1.1.1.1 1998/04/20 11:32:11 lefessan Exp $ */

/* Operations on weak arrays */

#include "alloc.h"
#include "fail.h"
#include "memory.h"
#include "mlvalues.h"

value weak_list_head = 0;


/* if a weak set with a bigger size is full, it triggers Major GC */
#define WEAKSET_TRIGGER_GC 8192
#define WEAKSET_THRESHOLD  32

#define None_val 1
#define Some_tag 0

/* and now, something better:
  Every free slot in the weak array is the index of the next free slot.
  We only have to verify that no Val_int is inserted in the weakarray.

  The set operation is a bit expensive if the slot used is a free one.
  But older Weak.t should not suffer from that since Weak.create creates
  a Weak.t with no free slots...
 */

value weak_create (value len)        /* ML */
{
  mlsize_t size, i;
  value res;

  size = Long_val (len) + 2;
  if (size > Max_wosize) invalid_argument ("Weak.create");
  res = alloc_shr (size, Weak_tag);
  for (i = 2; i < size; i++)
    Field (res, i) = 0;
  Field(res,1) = Val_int(size);
  Field (res, 0) = weak_list_head;
  weak_list_head = res;
  return res;
}

value weakarray_create (value len)        /* ML */
{
  mlsize_t size, i;
  value res;

  size = Long_val (len) + 2;
  if (size > Max_wosize) invalid_argument ("Weak.create");
  res = alloc_shr (size, Weak_tag);
  for (i = 1; i < size; i++)
    Field (res, i) = Val_int(i+1);

  Field (res, 0) = weak_list_head;
  weak_list_head = res;
  return res;
}

value weakarray_realloc(value ar,value new_size)   /* ML */
{
  value res;
  mlsize_t size, i;
  mlsize_t sz = Wosize_val(ar);
      
  size = 2+Int_val(new_size);
  if (size > Max_wosize || size <= sz) 
    invalid_argument ("Weak.realloc: invalid size");

  Begin_root(ar); /* (ar) */
  res = alloc_shr (size, Weak_tag);
  End_roots(); /* (ar) */

  for (i = 1; i < sz; i++) /* copy the first set */
    initialize(&Field (res, i),Field(ar,i));
  for (; i < size; i++) 
    Field (res, i) = Val_int(i+1);

  Field(res,0) = weak_list_head;
  weak_list_head = res;

  return res;
}


value weakarray_add(value ar,value el) /* ML */
{
  mlsize_t free = Int_val(Field(ar,1));
  mlsize_t sz = Wosize_val(ar);
  value tmp_free;

  DPRINT1("Weak_add %ld\n",free);
  if(free == sz){/* the set is full */
    if(Wosize_val(ar) > WEAKSET_TRIGGER_GC){
      Begin_roots2(ar,el); /* (ar,el) */
      finish_major_cycle();
      End_roots(); /* (ar,el) */
      free = Int_val(Field(ar,1));
    }
    if(free == sz) /* the set is still full */
      failwith("Weak.add: set full");
  }
  tmp_free = Field (ar, free);
  if(el == None_val)
    Field(ar,free) = 0;
  else {
    el = Field(el,0);
    if(Is_long(el)) 
      invalid_argument("Can't add integers in weak arrays");
    Modify (&Field (ar, free), el);
  }
  DPRINT1("Weakset_add %ld\n",free);
  Field(ar,1) = tmp_free;

  return Val_int(free-2);
}

value weakarray_free(value ar,value n) /* ML */
{
  mlsize_t offset = Long_val (n) + 2;
  value tmp_free;

  tmp_free = Field(ar,offset);
  if(Is_long(tmp_free)) return Val_unit;
  
  Field(ar,offset) = Field(ar,1);
  Field(ar,1) = Val_int(offset);

  return Val_unit;
}

value weak_set (value ar, value n, value el)     /* ML */
{
  mlsize_t offset = Long_val (n) + 2;
  value tmp_free;
  mlsize_t free;
                                                   Assert (Is_in_heap (ar));
  if (offset < 2 || offset >= Wosize_val (ar)) invalid_argument ("Weak.set");
  tmp_free = Field (ar, offset);
  if(Is_long(tmp_free)){ /* an empty slot */
    if(el == None_val) /* that's OK */
      Field (ar, offset) = 0;
    else
      Modify (&Field (ar, offset), Field (el, 0));
    for(free=1;Field(ar,free)!=Val_int(offset);free = Int_val(Field(ar,free)));
    Field(ar,free) = tmp_free;
    return Val_unit;
  }
  Field(ar,offset) = 0;
  if (el != None_val){                  Assert (Wosize_val (el) == 1);
    Modify (&Field (ar, offset), Field (el, 0));
  }
  
  return Val_unit;
}


#define Setup_for_gc
#define Restore_after_gc

value weak_get (value ar, value n)        /* ML */
{
  mlsize_t offset = Long_val (n) + 2;
  value res;
  value elt;

  Assert (Is_in_heap (ar));
  if (offset < 2 || offset >= Wosize_val (ar)) invalid_argument ("Weak.get");
  elt = Field (ar, offset);
  if (Is_long(elt) || elt==0){
    res = None_val;
    DPRINT("Weak.get : collected\n");
  }else{
    if (gc_phase == Phase_mark) darken (elt, NULL);
    DPRINT1("Weak.get : alive %lx\n",elt);
    Begin_root(elt);
      res = alloc (1, Some_tag);
    End_roots ();
    Field (res, 0) = elt;
  }
  return res;
}

#undef Setup_for_gc
#undef Restore_after_gc
 

/*
   Weaksets are not indexed
*/

value weakset_create (value len)        /* ML */
{
  mlsize_t size, i;
  value res;

  size = Long_val (len) + 2;
  if (size > Max_wosize) invalid_argument ("Weakset.create");
  res = alloc_shr (size, Weakset_tag);
  for (i = 2; i < size; i++) Field (res, i) = 0;
  Field (res, 1) = Val_int(2);
  Field (res, 0) = weak_list_head;
  weak_list_head = res;
  return res;
}

value weakset_add(value ar,value el) /* ML */
{
  mlsize_t free = Int_val(Field(ar,1));
  mlsize_t sz = Wosize_val(ar);

  DPRINT1("Weakset_add %ld\n",free);
  if(free == sz){/* the set is full */
    if(Wosize_val(ar) > WEAKSET_TRIGGER_GC){
      Begin_roots2(ar,el); /* (ar,el) */
      finish_major_cycle();
      End_roots(); /* (ar,el) */
      free = Int_val(Field(ar,1));
    }
    if(free >= sz-(sz/WEAKSET_THRESHOLD)){/* the set should be reallocated */
      value res;
      mlsize_t size, i;
      
      size = 2*sz+2;
      if (size > Max_wosize) invalid_argument ("Weak set full");

      Begin_roots2(ar,el);  /* (ar,el) */
      res = alloc_shr (size, Weakset_tag);
      End_roots(); /* (ar,el) */

      for (i = 2; i < free; i++) /* copy the first set */
	initialize(&Field (res, i),Field(ar,i));
      for (; i < size; i++) Field (res, i) = 0;

      Field(ar,1) = Val_int(2); /* now, ar is empty */
      Field(res,0) = weak_list_head;
      weak_list_head = res;
      ar = res;
    }
  }
  Modify (&Field (ar, free), el);
  free++;
  DPRINT1("Weakset_add %ld\n",free);
  Field(ar,1) = Val_int(free);

  return ar;
}

#define Setup_for_gc
#define Restore_after_gc

value weakset_array (value ar) /* ML */
{
  value res,elt;
  mlsize_t i,len;

  len = Int_val(Field(ar,1));

  if(len==2)
    return Atom(0);
  else 
    if(len<Max_young_wosize){
      Begin_root(ar);
      res = alloc(len-2,0);
      End_roots();
      for(i=2;i<len;i++){
	elt = Field (ar, i);
	if (gc_phase == Phase_mark)  {
	  DPRINT("weakset_array :\n");
	  darken (elt, NULL);
	}
	Field(res,i-2) = elt;
      }
    } else {
      Begin_root(ar);
      res = alloc_shr(len-2,0);
      End_roots();
      
      for(i=2;i<len;i++){
	elt = Field (ar, i);
	if (gc_phase == Phase_mark){
	  DPRINT("weakset_array :\n");
	  darken (elt, NULL);
	}
	initialize(&Field(res,i-2),elt);
      }
    }
  return res;
}

value weakset_remove (value ar, value el) /* ML */
{
  mlsize_t i,len;
  value elt;

  len = Int_val(Field(ar,1));
  for(i=2;i<len;i++){
    elt=Field (ar, i);
    if (gc_phase == Phase_mark) {DPRINT("weakset_remove :\n");
    darken (elt, NULL);}
    if(elt == el){
      Modify(&Field (ar, i),Field (ar, len-1));
      Field(ar,1) = Val_int(len-1);
      return Val_true;
    }
  }
  return Val_false;
}

#undef Setup_for_gc
#undef Restore_after_gc
 

value weakset_info (value ar) /* ML */
{
  value res;

  Begin_root(ar);
  res = alloc(2,0);
  End_roots();

  Field(res,0) = Val_int(Wosize_val(ar));
  Field(res,1) = Val_int(Int_val(Field(ar,1))-2);

  return res;
}

