(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: optcompile.ml,v 1.1.1.1 1998/04/20 11:32:12 lefessan Exp $ *)

(* The batch compiler *)

open Misc
open Config
open Format
open Typedtree

(* Initialize the search path.
   The current directory is always searched first,
   then the directories specified with the -I option (in command-line order),
   then the standard library directory. *)

let init_path () =
  load_path :=
    "" :: List.rev (Config.standard_library :: 
                    (if !Clflags.join_safe then
                      Filename.concat Config.standard_library "join" 
                      :: !Clflags.include_dirs
                    else
                      !Clflags.include_dirs));
  Env.reset_cache()

(* Return the initial environment in which compilation proceeds. *)

let initial_env () =
  init_path();
  try
    if !Clflags.nopervasives
    then Env.initial
    else Env.open_pers_signature "Pervasives" Env.initial
  with Not_found ->
    fatal_error "cannot open Pervasives.cmi"

(* Optionally preprocess a source file *)

let preprocess sourcefile tmpfile =
  match !Clflags.preprocessor with
    None -> sourcefile
  | Some pp ->
      let comm = pp ^ " " ^ sourcefile ^ " > " ^ tmpfile in
      if Ccomp.command comm <> 0 then begin
        Printf.eprintf "Preprocessing error\n";
        flush stderr;
        exit 2
      end;
      tmpfile

let remove_preprocessed inputfile =
  match !Clflags.preprocessor with
    None -> ()
  | Some _ -> remove_file inputfile

(* Parse a file or get a dumped syntax tree in it *)

exception Outdated_version

let parse_file inputfile parse_fun ast_magic =
  let ic = open_in_bin inputfile in
  let is_ast_file =
    try
      let buffer = String.create (String.length ast_magic) in
      really_input ic buffer 0 (String.length ast_magic);
      if buffer = ast_magic then true
      else if String.sub buffer 0 9 = String.sub ast_magic 0 9 then
        raise Outdated_version
      else false
    with
      Outdated_version ->
        fatal_error "Ocaml and preprocessor have incompatible versions"
    | _ -> false
  in
  let ast =
    try
      if is_ast_file then begin
        Location.input_name := input_value ic;
        input_value ic
      end else begin
        seek_in ic 0;
        Location.input_name := inputfile;
        parse_fun (Lexing.from_channel ic)
      end
    with x -> close_in ic; raise x
  in
  close_in ic;
  ast

let print_signature_to_file filename sg crcs =
  let oc = open_out filename in
  Format.set_formatter_out_channel oc;
  Format.print_string "(* Auto-generated mli file *)";
  Format.print_newline ();
  let crc = Printtyp.signature_ext sg filename crcs in
  print_newline();
  Format.set_formatter_out_channel stdout;
  close_out oc;
  crc


(* Compile a .mli file *)

let interface sourcefile =
  let prefixname = Filename.chop_extension sourcefile in
  let modulename = String.capitalize(Filename.basename prefixname) in
  let inputfile = preprocess sourcefile (prefixname ^ ".ppi") in
  let ast = parse_file inputfile Parse.interface ast_intf_magic_number in
  let sg = Typemod.transl_signature (initial_env()) ast in
  if !Clflags.print_types then (Printtyp.signature sg; print_newline());
  Env.save_signature sg modulename prefixname print_signature_to_file;
  remove_preprocessed inputfile

(* Compile a .ml file *)

let print_if flag printer arg =
  if !flag then begin printer arg; print_newline() end;
  arg

let implementation sourcefile =
  Clflags.native_code := true;
  let prefixname = Filename.chop_extension sourcefile in
  let modulename = String.capitalize(Filename.basename prefixname) in
  let inputfile = preprocess sourcefile (prefixname ^ ".ppo") in
  let ast = parse_file inputfile Parse.implementation ast_impl_magic_number in
  let (str, sg, finalenv) = Typemod.type_structure (initial_env()) ast in
  if !Clflags.print_types then (Printtyp.signature sg; print_newline());
  let coercion =
    let mli_file = prefixname ^ ".mli" in
      let cmi_file = prefixname ^ ".cmi" in
        if Sys.file_exists mli_file then 
          let dclsig = Env.read_signature modulename cmi_file in
            Includemod.compunit sourcefile sg mli_file dclsig
        else 
          begin
            Typemod.check_nongen_schemes finalenv str;
            if Env.save_signature sg modulename prefixname
                print_signature_to_file then
              Tcoerce_none
            else
              let dclsig = Env.read_signature modulename cmi_file in
              Includemod.compunit sourcefile sg mli_file dclsig
          end  
  in
  Compilenv.reset modulename;
  let (compunit_size, lam) =
    Translmod.transl_store_implementation modulename str coercion in
  Asmgen.compile_implementation prefixname compunit_size
    (print_if Clflags.dump_lambda Printlambda.lambda
      (Simplif.simplify_lambda
        (print_if Clflags.dump_rawlambda Printlambda.lambda lam)));
  Compilenv.save_unit_info (prefixname ^ ".cmx");
  remove_preprocessed inputfile;
  Clflags.native_code := false


let c_file name =
  if Ccomp.compile_file_native name <> 0 then exit 2

let mixte_implementation sourcefile =
  Clflags.native_code := true;
  Clflags.mixte := true;
  let prefixname = Filename.chop_extension sourcefile in
  let modulename = String.capitalize(Filename.basename prefixname) in

  let inputfile = preprocess sourcefile (prefixname ^ ".ppo") in
  let ast = parse_file inputfile Parse.implementation ast_impl_magic_number in
  let (str, sg, finalenv) = Typemod.type_structure (initial_env()) ast in
  if !Clflags.print_types then (Printtyp.signature sg; print_newline());
  let coercion =
    let mli_file = prefixname ^ ".mli" in
    let cmi_file = prefixname ^ ".cmi" in
    if Sys.file_exists mli_file then 
      let dclsig = Env.read_signature modulename cmi_file in
      Includemod.compunit sourcefile sg mli_file dclsig
    else 
      begin
        Typemod.check_nongen_schemes finalenv str;
        if Env.save_signature sg modulename prefixname
            print_signature_to_file then
          Tcoerce_none
        else
          let dclsig = Env.read_signature modulename cmi_file in
          Includemod.compunit sourcefile sg mli_file dclsig
      end  
  in
  Compilenv.reset modulename;
  let (compunit_size, lam) =
    Translmod.transl_store_implementation modulename str coercion in
  Asmgen.compile_implementation prefixname compunit_size
    (print_if Clflags.dump_lambda Printlambda.lambda
       (Simplif.simplify_lambda
          (print_if Clflags.dump_rawlambda Printlambda.lambda lam)));
  Compilenv.save_unit_info (prefixname ^ ".cmx");
  remove_preprocessed inputfile;
  Clflags.native_code := false;

  let objfile = prefixname ^ ".cmo" in
  let oc = open_out_bin objfile in
  try
    let cmi_file = prefixname ^ ".cmi" in
    let sg = Env.read_signature modulename cmi_file in
    Typemod.mixte_signature sg;
    Emitcode.to_file oc modulename
      (print_if Clflags.dump_instr Printinstr.instrlist
         (Bytegen.compile_implementation modulename
            (print_if Clflags.dump_lambda Printlambda.lambda
               (Simplif.simplify_lambda
                  (print_if Clflags.dump_rawlambda Printlambda.lambda
                     (Translmod.mixte_interface modulename ))))));
    close_out oc
  with x ->
    close_out oc;
    remove_file objfile;
    raise x
    
