(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open ml

(* test the speed of pure message-passing *)

let benchsize = 256
let logsize = 10

do shutdown.ready()

let ping(x) = reply to ping

let counter(n) = 
  let tick(x) | count(n) = if n=1 then { done() } else { count(n-1) }
  and done() | wait() = reply to wait in
  count(n) | tick("essai") | reply wait,tick

let print_int(x) = 
  let pad(x) = if x<100000 then ( print_string(" ");pad(x*10) ) ; reply in
  pad(x);print(x);reply
;;

let display(msg_size,msg_number,time) =
  print_int(msg_number);
  print_string(" x ");
  print_int(msg_size);
  print_string("-msg  time=");
  print_string(string_of_float(time));
  print_newline();
  reply
;;

let chrono(job)  =
  let start = unix.gettimeofday() in
  job();
  let stop = unix.gettimeofday() in
  reply ( stop -. start )
  
spawn { 
  if ns.server then { 
    ns.register("ping",ping); 
    ns.register("counter",counter);
  } else {
    let rping = ns.lookup("ping") in
    let rcounter = ns.lookup("counter") in

    let test_ping(msg,size,ping) =
      (* sequential function-calls with one large argument, no result *)
      let x = string.create(msg) in
      let loop(n) = if n<size then ( ping(x);loop(n+1) ); reply to loop in
      let job() = loop(0); reply in
      let time = chrono(job) in
      display(msg,size,time);
      reply in

    let test_ping_four(msg,size,ping) =
      (* sequential function-calls with one large argument, no result *)
      let x = string.create(msg) in
      let loop(n,m) = if n<m then ( ping(x);loop(n+1,m) ); reply to loop in
      let job() = 
        let one() | two() | three() | four() = reply to four in
        let s = size/4 in
          loop(0,s);one()
        | loop(s,s+s);two()
        | loop(s+s,s+s+s);three()
        | loop(s+s+s,s+s+s+s);four();reply in
      let time = chrono(job) in
      display(msg,size,time);
      reply  in

    let test_parallel_ping(msg,size,ping) =
      let x = string.create(msg) in
      let tick() | count(n) = if n<size then { count(n+1) } else { done() }
      and done() | wait() = reply to wait in
      let loop(n) = if n = size then reply to loop else
        { ping(x);tick() | loop(n+1);reply to loop } in
      let job() = count(1) | loop(0);wait(); reply in
      let time = chrono(job) in
      display(msg,size,time);
      reply in
      
    let test_counter(msg,size,counter) =
      (* asynchronous messages with one large argument *)
      let wait,ping = counter(size) in
      let x = string.create(msg) in
      let loop(n) = if n<size then { ping(x) | loop(n+1) } in
      let job() = 
        loop(0) | wait();reply in
      let time = chrono(job) in
      display(msg,size,time);
      reply in

    let bench(name,test,f) =
      let loop(msg,n) = 
        test(msg,benchsize,f); 
        if n<logsize then loop(msg*2,n+1); 
        reply in
      print_string(" --- ");print_string(name); print_string(" ---");print_newline();
      loop(1,0);
      reply in

    print_newline();print_string(" --- PING (sequential) ---");print_newline();
    bench("local",test_ping,ping);
    bench("remote",test_ping,rping);
    print_newline();print_string(" --- PING (four processes) ---");print_newline();
    bench("local",test_ping_four,ping);
    bench("remote",test_ping_four,rping);
    print_newline();print_string(" --- PING (parallel) ---");print_newline();
    bench("local",test_parallel_ping,ping);
    bench("remote",test_parallel_ping,rping);
    print_newline();print_string(" --- POST ---");print_newline();
    bench("local",test_counter,counter);
    bench("remote",test_counter,rcounter);
    shutdown.global() }}
    

