(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

open Jtypes;;

let create module_data =
  let self_loc = Jprims.self_location () in
  Jprims.create_new_location Space.local_space self_loc module_data
;;


(*
  Il faut transformer tous ces objets locaux en objets distants. 
Cependant, il ne faut pas transmettre a l'exterieur l'information qu'ils
ont migres, tant que cette migration n'est pas achevee. Le plus simple
est de creer des scions vers ces objets pour l'espace dans lequel ils
vont migrer. Les locators presents dans ces scions continuent de faire
croire que les objets sont locaux, alors qu'ils sont distants.

Le message d'acquittement de la migration devra simplement changer ces
locators pour qu'ils pointent vers le nouvel espace.

GC: on cre des stubs pour des scions qui n'existent pas encore. Quel timestamp
leur donner ? A priori, on peut mettre celui du dernier message recu si 
disponible (space.last_ts).
*)

let marshal dest_space location =
  let modules = ref [] in
  let ids = ref [] in
  let rec marshal_loc location =
    let (locator,_,_,_,_) = Obj.magic location in
    let (thread_stubs, defs, locs, module_data, attachements, vars) =
      Jprims.marshal_location dest_space location
    in
    let threads = Array.map (fun stub -> 
      stub, Jprims.marshal_thread stub) thread_stubs
    in
    let defs = Array.map (fun stub -> stub, Jprims.marshal_def stub) defs in
    let attachements = 
      List.map (function module_data -> 
        if not (List.mem module_data.Reloc.module_name !modules) then
          modules := module_data.Reloc.module_name :: !modules;
        (module_data,Jprims.get_interface module_data)
        ) attachements in
    let locs = Array.map (fun stub -> marshal_loc stub) locs in
    if not (List.mem module_data.Reloc.module_name !modules) then
      modules := module_data.Reloc.module_name :: !modules;
    LocationBloc (location, threads, defs, locs, attachements,module_data,vars)
  in
  let m = marshal_loc location in
  !modules, m
;;

let insert_stub space ts pointer =
  let (Some old_locator,position,local, _, old_gc_stub) = Obj.magic pointer in
  old_locator.distance <- 10000;
  let gc_stub = Some (pointer,space,old_locator.age+1) in
  Jprims.set_stub_locator pointer 
    (Some { 
     age = old_locator.age+1;
     distance = 1;
     otype = old_locator.otype;
     uid = old_locator.uid
   });
  Jprims.set_stub_stub pointer gc_stub;
  Whashtbl.add space.stubs_table (old_locator.uid,old_locator.age+1)
    (Some gc_stub);
  Debug.print1 'l' "Insert gc_stub for %s\n" (Space.string_of_uid old_locator.uid)

let rec insert_migration_stubs space ts msg = 
  let insert stub = insert_stub space ts (Obj.repr stub) in
  let rec insert_for_location =
    function 
        (LocationBloc (location, threads, defs, locs, mods, module_data,vars)) ->
          Jprims.set_stub_space location space;
          insert location;
          Array.iter (fun (stub,_) -> insert stub) threads;
          Array.iter (fun (stub,_) -> insert  stub) defs;
          Array.iter insert_for_location locs
  in
  insert_for_location msg

(* on doit ajouter un scion vers l'objet,ainsi que incrementer l'age du 
locator , et mettre la distance a 0 *)

let insert_scion space ts pointer =
  let (Some old_locator,position,local,_,old_gc_stub) = Obj.magic pointer in
  (match old_locator.otype with
    Location (_,oo_desc) ->
      old_locator.otype <- Location (Space.local_id,oo_desc)
  | _ -> ());
  (try
    let scion = Hashtbl.find Space.scions_table 
        (old_locator.uid,old_locator.age)
    in
    scion.old_stub <- old_gc_stub;
    Hashtbl.remove Space.scions_table (old_locator.uid,old_locator.age)
  with
    Not_found -> ());
  old_locator.age <- old_locator.age + 1;
  old_locator.distance <- 0;
  let scion = {
    object = Obj.repr pointer;
    cyclic_ts = 0;
    scion_age = old_locator.age;
    old_stub = None;
    gc_scion_counter = 1
  }
  in
  let gc_scion = {
    scion_ts = ts;
    source = space;
    scion = scion
  } in
  Hashtbl.add Space.scions_table (old_locator.uid,old_locator.age) scion;
  Hashtbl.add space.scions_table (old_locator.uid,old_locator.age) gc_scion;
  Jprims.set_stub_stub pointer None

let insert_location_scions space ts modules m =
  let ts = space.next_ts in
  space.next_ts <- space.next_ts + 1;
  let insert stub = insert_scion space ts (Obj.repr stub) 
  in
  let rec insert_location (LocationBloc
                             (location, threads, defs, locs,
                              attachements, module_data, vars)) =
    List.iter (function (module_data,interface) ->
      Jprims.set_interface module_data interface) attachements;
    insert location;
    Array.iter (function (def_stub,ldef)  -> insert def_stub) defs;
    Array.iter (function (thread_stub,thread) -> insert thread_stub) threads;
    Array.iter insert_location locs
  in
  insert_location m

let unmarshal space ts modules m dest =
  let insert stub = insert_scion space ts (Obj.repr stub) in
  let rec unmarshal_loc dest =
    fun
      (LocationBloc
         (location, threads, defs, locs, attachements, module_data,vars)) ->
           Jprims.unmarshal_location dest location module_data 
             (List.map (function (module_data,interface) ->
               Jprims.set_interface module_data interface;
               module_data) attachements) vars;
           insert location;
           Array.iter (function ((def_stub,ldef) as paire) -> 
             Jprims.test_stub def_stub "unmarshal def ";
             Jprims.unmarshal_def paire;
             insert def_stub) defs;
           Array.iter (function (thread_stub,thread) as paire -> 
             Jprims.unmarshal_thread paire;
             insert thread_stub) threads;
           Array.iter (unmarshal_loc location) locs;
  in
  unmarshal_loc dest m
;;
    
(*
  Dans le halt, on voit bien la distinction entre la location toplevel et les
autres sous-locationns. Le language devrait peut-etre aussi faire cette 
distinction: 
  location = sous-location 
  space = toplevel-location

  Cela semble plus approprie a l'etude des failures, qui ne concernent vraiment
que les locations toplevel. De plus, la migration pourrait alors etre limite 
correctement a go_space et go_location, le second etant purement local.

*)

(* 
 halt_toplevel: terminaison propre d'un runtime.

Il faut prevenir les espaces voisins de la coupure prochaine des canaux,
des failures de locations dans l'espace. Il faut aussi short-cutter de
facon intensive toutes les SSPC indirectes passant par l'espace.
*)

let halt_toplevel' location =
  Jprims.set_critical_section true;
  Hashtbl.iter (fun space_id space ->
    ()    )  Space.spaces;
  Jprims.set_critical_section false;
  exit 0

let halt_toplevel location = failwith "Toplevel halt not implemented"

let halt_location location =
  if location ==  Space.local_location then
    halt_toplevel location
  else
    Jprims.halt_location location
