(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

open Jtypes;;

let string_of_channel (_,_,stub,ord) =
  let (locator,position,local,_,gc_stub) = Obj.magic stub in
  (if (Obj.magic position) == None then "<failed>" else "")^
  (match locator with
    None -> "Local "
  | Some locator -> Space.string_of_uid locator.uid
        )^" : "^(string_of_int ord)


(* fail safe *)
let send_asynchrone (channel : channel) (arg : 'a) =
  Jprims.critical_section (function () ->
    let (_,_,automata, ord) = channel in
    let (locator,location,local,_,_) = Obj.magic automata in
    if not ((Obj.magic local) == None) then 
        Jprims.trigger_def (automata, ord, arg)
    else
      if not ((Obj.magic location) == None) then
        match locator with
          Some locator ->  
            let (_,space,_,_,_) = Obj.magic location in
            if not ((Obj.magic space) == None) then
              Jmarshal.send space (Asynchrone (locator.uid, ord, arg))
        | None -> failwith "No locator for remote stub"
              )        
    
let send_synchrone automata ord arg =
  Jprims.set_critical_section true;
  try
    let (locator,location,local,_,_) = Obj.magic automata in
    match
      if not ((Obj.magic local) == None) then 
        let thread_stub = Jprims.self_stub () in
        Jprims.trigger_def_and_wait (automata, ord, (thread_stub, arg))
      else
        if (Obj.magic location) == None then Jprims.wait_on_stub () else
        match locator with
          Some locator ->
            let thread_stub = Jprims.self_stub () in
            let (_,space,_,_,_) = Obj.magic location in
            if not ((Obj.magic space) == None) then
              Jmarshal.send space 
                (Synchrone (locator.uid,ord, arg, thread_stub));
            Jprims.wait_on_stub ()
        | None -> failwith "No locator for remote stub"
    with
      Reply v -> Obj.magic v
    | Exception e -> raise e
  with
    e -> 
      Jprims.set_critical_section false;
      raise e
;;

(* ==========  DEBUT hack !!!!  ========== *)
open Obj

let automata = (magic  374848: automata)
let ord = 298274
let f = repr (send_synchrone automata ord)

let _ = 
  if
    (size f <> 4) ||
    (field f 2) <> (repr automata) ||
    (field f 3) <> (repr ord)
  then
    failwith "Bad representation of synchronous channels"
let code_t = (magic (field f 0): code_t)
let env_t = (magic (field f 1): env_t)
let _ = Jprims.register_send_sync code_t env_t
(* ==========   FIN hack !!!!  ========== *)

let hard_reply_to thread_stub v =
  let (locator,location,local,_,_) = Obj.magic thread_stub in
  if not ((Obj.magic local) == None) then 
    Jprims.reply_thread_stub thread_stub v
  else 
    if not ((Obj.magic location) == None) then
      (match locator with
        Some locator ->
          let (_,space,_,_,_) = Obj.magic location in
          if not ((Obj.magic space) == None) then
            Jmarshal.send space
              (ReplyThread (locator.uid, Obj.repr v)) 
      | None -> failwith "No locator for remote stub"
            )
        
let reply_to_and_exit thread_stub v = 
  Jprims.set_critical_section  true;
  try
    hard_reply_to thread_stub v;
    Thread.exit ()
  with
      e ->
        Jprims.set_critical_section false;
        (* on ne peut pas relancer l'exception ! *)
        prerr_string "Warning: exception in reply_to_and_exit: Thread.exit\n";
        Thread.exit ();
        raise e

let reply_to thread_stub v =
  Jprims.critical_section 
    (function () -> hard_reply_to thread_stub v)

let tail_synchrone (channel:channel) arg thread_stub =
  Jprims.set_critical_section true;
  try
    let (_,_,automata, ord) = channel in
    let (locator,location,local,_,_) = Obj.magic automata in
    if not ((Obj.magic local) == None) then 
      Jprims.exit_trigger_def_and_wait (automata, ord, (thread_stub, arg))
    else
      if (Obj.magic location) == None then Jprims.wait_on_stub () else
      match locator with
        Some locator ->
          let (_,space,_,_,_) = Obj.magic location in
          if not ((Obj.magic space) == None) then
            Jmarshal.send space 
              (Synchrone (locator.uid,ord, arg, thread_stub));
          Jprims.exit_wait_on_stub ()
        | None -> failwith "No locator for remote stub"
  with
    e -> 
      Jprims.set_critical_section  false;
                (* on ne peut pas relancer l'exception ! *)
      prerr_string "Warning: exception in tail_synchrone: Thread.exit\n";
      Thread.exit ();
      raise e


let appterm f arg thread_stub =
  let o = repr f in
  if size o == 4 && field o 0 == repr code_t && field o 1 == repr env_t then
    tail_synchrone f arg thread_stub
  else
    reply_to_and_exit thread_stub ((magic f) arg)
    
let rec invoke_no_reply object meth_name label args thread_stub =
  Jprims.critical_section
    (fun () ->
      let (locator, space, local, _, gc_stub) = Obj.magic object in
      if not ((Obj.magic local) == None) then
        begin
          Thread.create_in (Obj.magic object:Jtypes.location)
            (fun () ->
              reply_to_and_exit thread_stub
                (try
                  Debug.print1 'o' "N args %d\n" (List.length args);
                  Reply (
                  Mop.invoke_methode 
                    (Mop.lookup object label)
                    object (List.rev args)
                    )    
                with
                  e -> Exception e
                      )
                ) ();
          ()
        end
      else
      (* l'objet est distant *)
        if not ((Obj.magic space) == None)  then
          match locator with
            Some locator ->
              Space.send space
	        (Jmarshal.marshal space 
	           (Invoke (locator.uid, meth_name,label,args,thread_stub)))
          | None -> failwith "No locator for remote stub"
                )

let create_ext_buckets class_sig =
  Mop.create_ext_buckets class_sig 
    (fun nparams meth_name label ->
      Debug.print2 'o' "Method %s with  %d params\n" meth_name nparams;
      (Obj.magic
         (fun object ->
           let rec iter nparams args =
             Debug.print1 'o' "Iter %d params\n" nparams;
             if nparams = 0 then
               (* ici, on est sur de rien : l'objet peut etre local
                  ou peut avoir migre *)
               let thread_stub = Jprims.self_stub () in
               Jprims.set_critical_section true;
               invoke_no_reply object meth_name label args thread_stub;
               match Jprims.wait_on_stub () with
                 Jtypes.Reply v -> Obj.magic v
               | Jtypes.Exception e -> raise e
             else
               Obj.magic (fun arg -> iter (nparams-1) (arg::args))
           in
           iter nparams []
             ): Mop.item)
        )

let remote_ext_buckets space class_id class_sig =
  try
    Hashtbl.find space.ext_buckets class_id
  with
    Not_found ->
      let ext_buckets = create_ext_buckets class_sig
      in
      Hashtbl.add space.ext_buckets class_id ext_buckets;
      ext_buckets

external sys_exit : exn -> 'a = "join_sys_exit"

let send_exception exn array =
  try
  let array = (Obj.magic array: thread_stub array)
  in
  Jprims.set_critical_section  true;
  Debug.print "send exception";
    Debug.print "send_exception";
    Array.iter (fun thread_stub ->
      hard_reply_to thread_stub (Exception exn);
      ) array;
    Debug.print "now exit";
    Thread.exit ()
  with
      e ->  sys_exit e  
;;


