(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)


open Jtypes;;
open Space;;
(*
1  On ne lance pas de nouveau thread dans ce handler. On suppose que, lorsque 
la minimisation du nombre de threads sera effectuee, on verifie bien avant
d'essayer d'utiliser la pile d'un thread, qu'il est bien local. Si ce n'est
pas le cas, on lance un nouveau thread pour effectuer le travail.

2 On forwarde toujours les appels asynchrones. On repond aux appels synchrones
non locaux par la nouvelles position de la location demandee. En cas de panne
de ce site, on se trouve donc dans deux cas:
- Aucun appel synchrone n'a ete forwarde. L'espace ne peu plus connaitre la
nouvelle position de la location.
- Un appel synchrone a ete forwarde. L'espace connait la nouvelle position de 
la location.

3. Si on migre vers une location qui a migre, il faut forwarde la migration.
Que contient le bloc transmis:
- les locations ont des space_id sur le site avant migration, mais sont 
demarshalee avec la version locale, ie pointant vers le site local, donc
pas de probleme.

*)
type lookup_element =
    { mutable reqs: thread_stub list;
      pointer: Obj.t;
      space: space;
      class_info: Mop.class_info
    }
(*			
let lookup_table = Hashtbl.create 10
*)
let handler (space : space) inc =
  Debug.print "New handler";
  while true do
    let (message:message) = input_value inc in
    Jprims.critical_section  (
    function () ->
      try
        Debug.print0 'h' "new message";
        let (ts,_,locators) = message in
        let message = 
          Debug.trap 'h' "unmarshal" (fun () ->
            Junmarshal.unmarshal space message) in
        space.threshold <- ts;
        Debug.print1 'h' "message unmarshaled : %d\n" ts;
        match message with
          GetService (name,vt, thread_stub) ->
            Debug.print0 'h' "GETSERVICE";
            Channel.reply_to thread_stub
              (try Reply (Services.getLocalService name (Obj.magic vt)) with
                e -> Exception e)
        | Synchrone (uid, ord, arg, thread_stub) ->
            Debug.print0 'h' "SYNCHRONE";
            Channel.send_asynchrone (Channel.code_t,Channel.env_t,(find_stub uid), ord) 
              (Obj.repr (thread_stub,arg))
        | Invoke (uid,methode,label,args,thread_stub) ->
            Debug.print0 'h' "INVOKE";
            Channel.invoke_no_reply (find_stub uid)
              methode label args thread_stub
        | Asynchrone (uid, ord, arg) ->
            Debug.print0 'h' "ASYNCHRONE";
            Channel.send_asynchrone (Channel.code_t,Channel.env_t,(find_stub uid), ord) arg
        | LocationMigration (modules, m, dest) ->
            Debug.print0 'h' "MIGRATION";
            let (_,dest_space,local,_,_) = Obj.magic dest in
            if local then
              begin
                Debug.print0 'h' "migration to local";
                Agent.unmarshal space ts modules m dest
              end
            else
              begin
                Agent.insert_location_scions space ts modules m;
                if not ((Obj.magic dest_space) == None) then 
                  let (ts,_,_) as message = 
                    (Jmarshal.marshal dest_space
                       (LocationMigration (modules, m, dest)));
                  in
                  Agent.insert_migration_stubs dest_space ts m;
                  Space.send dest_space message
              end
        | ReplyThread (uid, v) -> 
            Debug.print0 'h' "REPLY";
            Channel.reply_to (find_stub uid) v; ()
        | Life(threshold,life_stubs) ->
            Dgc.life_handler space threshold life_stubs
        | Failed list ->
            Dgc.fail_handler space list
      with
        e -> Debug.print0 'h' "Exception in handler\n"
            )
  done
;;
