(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

(*
  L'algo est bancal.
  
  Tout stub pouvant etre reference depuis l'exterieur possede au
  moins un gc_stub. Si l'objet est local, le gc_stub pointe vers
  l'espace courant.
*)

open Jtypes;;
open Reloc

let rec get_pointer_locator dest_space ts pointer = 
  let (locator, position, local, _, gc_stub) = Obj.magic pointer in
  match locator with
    None ->
      let tag = Obj.tag pointer in
      let locator = 
        {
        age = if (Obj.magic position) == None then failed_age else 0;
        distance = 0;
        uid = Space.new_uid ();
        otype =
        (if tag = Tag.location then
          Location ((Obj.magic position).space_id,None)
        else if tag = Tag.automata then
          Automata (ptr2loc dest_space ts position)
        else if tag = Tag.thread then
          Thread (ptr2loc dest_space ts position)
        else if tag = Tag.object then
          let class_info = Mop.get_class_info pointer in
          Debug.print0 'm' "Locator for object\n";
          Location ((Obj.magic position).space_id,
                    Some 
                      (class_info.Mop.class_type,
                       class_info.Mop.class_sig,
		       class_info.Mop.table.Mop.size)
		      )
        else failwith ("No locator for tag : " ^ string_of_int tag))
      } 
      in
      Debug.print1 'l' "New locator: %s\n" 
        (Space.string_of_uid locator.uid);
      let gc_stub = Some (pointer,Space.local_space,0) in
      Whashtbl.add Space.stubs_table locator.uid (Some pointer);
      Jprims.set_stub_stub pointer gc_stub;
      Jprims.set_stub_locator pointer (Some locator);
      locator
  | Some locator -> 
      if (Obj.magic position) == None then locator.age <- failed_age;
      locator


and ptr2loc dest_space ts pointer =
  Debug.print "<------------ ptr2loc";
  let locator = get_pointer_locator dest_space ts (pointer:Obj.t) in
  if locator.age = failed_age then locator else
  try
    let gc_scion = Hashtbl.find dest_space.scions_table
        (locator.uid,locator.age) 
    in
    if gc_scion.scion_ts < ts then gc_scion.scion_ts <- ts;
    Debug.print1 'l' "ptr2loc %s\n" (Space.string_of_uid locator.uid);
    locator
  with
    Not_found ->
      let scion =
        try
          Hashtbl.find Space.scions_table (locator.uid,locator.age)
        with
          Not_found ->
            let scion =
              { 
              object = Obj.repr pointer;
              cyclic_ts = 0;
              scion_age = locator.age;
              old_stub = None;
              gc_scion_counter = 0;
            }
            in
            Hashtbl.add Space.scions_table (locator.uid,locator.age) scion;
            scion
      in
      let gc_scion =
        {
        scion_ts = ts;
        source = dest_space;
        scion = (scion.gc_scion_counter<- scion.gc_scion_counter+1;scion)
      }
      in
      Hashtbl.add dest_space.scions_table (locator.uid,locator.age) gc_scion;
      Debug.print1 'l' "ptr2loc: %s\n" (Space.string_of_uid locator.uid);
      locator
 

and marshal (dest_space : space) message =
  Debug.print0 'l' "MARSHAL ...\n";
  let ts = dest_space.next_ts in
  dest_space.next_ts <- dest_space.next_ts + 1;
  let (buffer, pointers) = Jprims.marshal_message message in
  let locators =
    Array.map 
      (fun pointer ->     
        let tag = Obj.tag pointer in
        if tag = Tag.reloc then
          let module_data = (Obj.magic pointer : module_data) in
          Debug.print1 'x' "Marshal module %s\n" module_data.module_name;

          if module_data.migrate && 
            not(List.mem module_data.module_name dest_space.modules) then
            begin
              Debug.print0 'x' "Module migrate\n";
              dest_space.modules <- module_data.Reloc.module_name
                 :: dest_space.modules;
              RelocModule module_data
            end
          else
            DummyModule (module_data.module_name,Obj.size pointer)
        else
          if tag = Tag.exn then
            let exn = (Obj.magic pointer : exn) 
            in
            RelocException(exn.exn_name,exn.mod_name,exn.mod_num)
          else
            (StubLocator
               (ptr2loc dest_space ts pointer)))
      pointers
  in
  Debug.print2 'l' "Marshal %s (%d)\n" 
    (match message with
      GetService _ -> "GETSERVICE"
    | Synchrone _ -> "SYNCHRONE"
    | Invoke _ -> "INVOKE"
    | Asynchrone _ -> "ASYNCHRONE"
    | LocationMigration _ -> "LOCATIONMIGRATION"
    | ReplyThread _ -> "REPLYTHREAD"
    | Life _ -> "LIFE") ts;
  ts, buffer, locators
;;

let send dest_space message =
  Debug.print "Locator.send";
  Space.send dest_space (marshal dest_space message)
;;

let _ = Jprims.set_locator_send send

  
