(***********************************************************************)
(*                                                                     *)
(*                     the join-calculus language                      *)
(*                                                                     *)
(*   Cedric Fournet and Luc Maranget, projet PARA, INRIA Rocquencourt  *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Printf;;

type ('k, 'd) t = { key : 'd -> 'k; mutable bucket : 'd Weak.t array };;
  
let create f = {key = f; bucket = [| |]};;

let clear t = t.bucket <- [| |];;

let min_bucket_size = 16;;
let max_bucket_size mask = max (mask lsr 3) min_bucket_size;;

let flip mask b =
  let digit = ref (mask + b) in
  let result = ref (b land 1) in
  while !digit > 3 do
    digit := !digit lsr 1; result := !result lsl 1 + !digit land 1; ()
  done;
  !result
;;

let iter f t =
  let mask = Array.length t.bucket in
  let last_d = ref (Weak.create 0) in
  for i = 0 to mask - 1 do
    let b = flip mask i in
    let d = t.bucket.(b) in
    if d != !last_d then
      begin
        last_d := d;
        for u = 0 to Weak.length d - 1 do
          match Weak.get d u with
            None -> ()
          | Some v -> f v
        done
      end
  done
;;

(*  
let dump t = iter (fun i b u k v ->
  printf " bucket %d (flip %d), pos %d, t(%d)= %d\n" b i u k v) t
*)
  
external hash_param : int -> int -> 'a -> int = "hash_univ_param" "noalloc";;
  
let sharing t b =
  let mask = Array.length t.bucket in
  let d = t.bucket.(b) in
  let fb = flip mask b in
  let i = ref fb in
  let j = ref fb in
  while decr i; !i >= 0 && t.bucket.(flip mask !i) == d do () done;
  while incr j; !j < mask && t.bucket.(flip mask !j) == d do () done;
  !i + 1, !j - 1
;;
    
let double_index t =
  let n = Array.length t.bucket in
  let n' = 2 * n in
  let bucket' = Array.create n' (Weak.create 0) in
  Array.blit t.bucket 0 bucket' 0 n;
  Array.blit t.bucket 0 bucket' n n;
  t.bucket <- bucket'
;;
              
let rec resize t b =
  let mask = Array.length t.bucket in
  let l = Weak.length t.bucket.(b) in
  let lmax = max_bucket_size mask in
  let (i, j) = sharing t b in
  let d = t.bucket.(b) in
  if l < lmax then
    let d' = Weak.create lmax in
    Weak.blit d 0 d' 0 l; for u = i to j do t.bucket.(flip mask u) <- d' done
  else
    let (i, j) =
      if i = j then begin double_index t; sharing t b end else i, j
    in
    let new_size = lmax in
    let mask = Array.length t.bucket in
    let d2 = Weak.create new_size in
    let q = ref 0 in
    let pivot = (i + j) / 2 in
    for p = 0 to l - 2 do
      match Weak.get d p with
        None -> ()
      | Some v as x ->
          let b' = hash_param 10 100 (t.key v) mod mask in
          if flip mask b' > pivot then
            begin Weak.set d2 !q x; Weak.set d p None; incr q end
    done;
    for u = pivot + 1 to j do t.bucket.(flip mask u) <- d2; () done;
    if !q = 0 then resize t (flip mask i);
    if !q = lmax - 1 then resize t (flip mask j);
    ()
;;

let find t key =
  let mask = Array.length t.bucket in
  let b =
    match mask with
      0 -> raise Not_found
    | 1 -> 0
    | _ -> hash_param 10 100 key mod mask
  in
  let d = t.bucket.(b) in
  let rec loop i =
    if i < Weak.length d then
      match Weak.get d i with
        Some v when t.key v = key -> v
      | _ -> loop (i + 1)
    else raise Not_found
  in
  loop 0
;;
      
(* was much funnier:
  exception Found of int
  try
    for i = 0 to Weak.length d - 1 do
      match Weak.get d i with
      | Some v when t.key v = key -> raise (Found i)
      | _ -> ()
    done;
    raise Not_found
  with
    Found i -> (
      ( * may just have been garbaged-collected... * )
      match Weak.get d i with
      | Some v -> v
      | None -> raise Not_found ) *)

  
let remove t key =
  let mask = Array.length t.bucket in
  let b =
    match mask with
      0 -> raise Not_found
    | 1 -> 0
    | _ -> hash_param 10 100 key mod mask
  in
  let d = t.bucket.(b) in
  let l = Weak.length d in
  let p = ref 0 in
  while
    !p < l &&
    (match Weak.get d !p with
       Some v when t.key v = key -> Weak.set d !p None; false
     | _ -> true)
  do incr p
  done
;;
      
let rec add t v =
  let mask = Array.length t.bucket in
  if mask = 0 then
    let d = Weak.create min_bucket_size in
    Weak.set d 0 (Some v); t.bucket <- [| d |]
  else
    let b = if mask = 1 then 0 else hash_param 10 100 (t.key v) mod mask in
    let k = t.key v in
    let d = t.bucket.(b) in
    let l = Weak.length d in
    let p = ref 0 in
    while Weak.get d !p <> None do incr p done;
    if !p = l - 1 then begin resize t b; add t v end
    else Weak.set d !p (Some v)
;;

let length t = let x = ref 0 in iter (fun _ -> incr x) t; !x;;
  
let size t =
  let x = ref (Array.length t.bucket) in
  Array.iter (fun v -> x := !x + Weak.length v) t.bucket; !x
;;
