(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

open Jtypes;;
open Jprims;;   
open Unix;;
open ThreadUnix;;

                   (* GLOBAL DATA *)
let (spaces : (inet_addr * port, space) Hashtbl.t) = Hashtbl.create 11;;
let (scions_table : (uid*int, scion) Hashtbl.t) = Hashtbl.create 101
;;
let (stubs_table : (uid, Obj.t) Whashtbl.t) = set_local (Whashtbl.create 101)
;;
let (name_server : (string, Obj.t*Obj.t) Hashtbl.t) = Hashtbl.create 11;;
let (weak_array : gc_stub Weak.array ref) = ref (Weak.array_create 4100);;
                    (* FUNCTIONS *)



let find_stub uid =
  try 
    match Whashtbl.find stubs_table uid with
      None -> raise Not_found
    | Some stub -> Obj.magic stub
  with
    Not_found -> raise Not_found



let creation_time = gettimeofday ();;
let localname = gethostname ();;
let first_addr entry = entry.h_addr_list.(0);;
let localaddr = first_addr (gethostbyname localname);;
(*
let handler = ref (fun (s : file_descr) (space : space_id) -> ());;
*)
let join_handler = ref (fun space inc -> ());;
let join_initialised = ref false;;

let string_of_space_id ((addr,port),time) =
  (Unix.string_of_inet_addr addr)^"@"^(string_of_int port)
  ^":"^(string_of_float time)

let string_of_uid = function
    Uid (space_id,ts) ->
      "UID ("^(string_of_int ts)^" "^(string_of_space_id space_id)^")"

let create_port port =
  let s = socket PF_INET SOCK_STREAM 0 in
  let _ =
    setsockopt s SO_REUSEADDR true;
    bind s (ADDR_INET (localaddr, port));
    listen s 5;
    ()
  in
  let saddr = getsockname s in
  match saddr with
    ADDR_INET (_, port) -> port, s
  | _ -> failwith "not ADDR_INET socket"
;;


let first_link = create_port 0;;
(*print_string "Open port :"; print_int (fst first_link); print_newline ();;*)

let local_id = (localaddr, fst first_link),creation_time;;
let (links : (port * file_descr) list ref) = ref [];; (* ports en ecoute *)
                                             
let local_space = 
  { space_id = local_id;
    queue = Fifo.create ();
    sender_thread = Thread.self ();
    next_ts = 1;
    ext_buckets = Hashtbl.create 13;
    set = 0;
    last_ts = 0;
    threshold = 0;
    modules = [];
    scions_table = Hashtbl.create 1;
    link = None;
    stubs_table = Whashtbl.create 1
  }

let links = ref [first_link];;

let rec accept s =
  try
    ThreadUnix.accept s
  with
    Unix_error(EAGAIN,_,_) -> 
      Debug.print0 's' "accept: Try again\n";
      accept s
  | e -> 
      accept s

let threads = ref 0
let rec wait_on_socket s =
  incr threads;
  Thread.create
    (fun () ->
       while true do
         let (s, _) = handle_unix_error accept s in
         let inc = in_channel_of_descr s in
         let outc = out_channel_of_descr s in
         let space_id = input_value inc in
         let space = get_space space_id in
         space.modules <- input_value inc;
         Jprims.set_critical_section true;
         match space.link with
           None ->
             space.link <-
                Some
                  {conn_inc = inc;
                    conn_outc = outc; 
                    conn_thread = 
                    Thread.create
                      (fun () ->
                        output_value outc true;
                        output_value outc local_id;
                        output_value outc (Reloc.module_list ());
                        flush outc;
                        try !join_handler space inc with
                          e -> close s; raise e
                              ) ()
                  };
             Jprims.set_critical_section false;
             Debug.print1 's' "Link open to %s\n"
               (string_of_space_id  space.space_id)
         | _ ->
             Jprims.set_critical_section false;
             output_value outc false;
             flush outc;
             Debug.print "Remote Space already connected"
       done)
    ()
    
and get_space (space_id:space_id) =
  Debug.print1 's' "get space %s\n" (string_of_space_id space_id);
  try Hashtbl.find spaces (fst space_id) with
    Not_found ->
      let space =
        { space_id = space_id;
          queue = Fifo.create ();
          sender_thread = Thread.self ();
          last_ts = 0;
          ext_buckets = Hashtbl.create 13;
          set = 0;
          next_ts = 1;
          modules = [];
          threshold = 0;
          scions_table = Hashtbl.create 100;
          link = None;
          stubs_table = Whashtbl.create 100} in
      space.sender_thread <- Thread.create (fun () ->
        while true do
          try
            while true do
              set_critical_section true;
              let message = Fifo.take space.queue in
              set_critical_section false;
              let ts,buffer,ls = message in
              Debug.print "send'";
              send' space message
            done
          with
            Fifo.Empty -> Thread.sleep ()
        done) ();
      Hashtbl.add spaces (fst space_id) space;
      space
and
(* inside join_critical *)
    connect_to space =
  Debug.print "Connecting to space ......";
  try
    let (inc, outc) =
      try
        open_connection (ADDR_INET (fst (fst space.space_id), 
                                    snd (fst space.space_id)))
      with
        e -> printnl_string "error on connect";
          raise Exit
    in
    output_value outc local_id;
    output_value outc (Reloc.module_list ());
    flush outc;
    if (input_value inc) then
      begin
        input_value inc; (* local_id *)
        space.modules <- input_value inc;
        Jprims.critical_section
          (function () ->
            if (space.link = None) then
              space.link <-
                 Some
                   {conn_inc = inc; conn_outc = outc;
                     conn_thread = Thread.create (!join_handler space) inc};
            Debug.print1 's' "Link open to %s"
              (string_of_space_id space.space_id); 
              );
        
      end
    else
      printnl_string "connected2: not OK";
  with
    Unix.Unix_error _ as e -> Debug.print "Error in connect_to"; raise e
        
and  send' space message =
  let rec send () =
    match space.link with
        None -> 
          Debug.print "connect to ";
          connect_to space; send ()
      | Some conn ->
          Debug.print "Space.send : output value ";
          output_value conn.conn_outc message;
          Debug.print "Space.send : output value OK ";
          flush conn.conn_outc;
          Debug.print0 's' "flushed\n"
  in
    send ()

and first_connect_to space_addr =
  Debug.print "Connecting to space ......";
  try
    let space = Hashtbl.find spaces space_addr
    in
    if (space.link = None) then connect_to space;
    space.space_id
  with
    Not_found ->
      try
        let (inc, outc) =
          open_connection (ADDR_INET (fst space_addr,snd space_addr))
        in
        output_value outc local_id;
        output_value outc (Reloc.module_list ());
        flush outc;
        if (input_value inc) then
          begin
            let space = get_space (input_value inc) in
            space.modules <- input_value inc;
            let old = Jprims.set_critical_section  true
            in
            printnl_string "first connect";
            if (space.link = None) then
              space.link <-
                 Some
                   {conn_inc = inc; conn_outc = outc;
                     conn_thread = Thread.create (!join_handler space) inc};
            (Debug.print1 's' "Link open to %s\n"
               (string_of_space_id space.space_id)
               );
            Jprims.set_critical_section old;
            space.space_id
          end
        else 
          failwith "Connect to: connection refused"
      with
        Unix.Unix_error _ as e -> Debug.print "Error in connect_to"; raise e;;

let hard_send space message =
  Debug.print1 's' "hard send to %s\n" (string_of_space_id space.space_id);
  Fifo.put space.queue message;
  Debug.print "Wake up";
  Thread.wakeup space.sender_thread;
  Debug.print "OK"


let send space message =
  Jprims.critical_section (function () -> 
    hard_send space message)

let uid_counter = ref 0
let new_uid () = 
  incr uid_counter;
  Uid (local_id,!uid_counter)

let create_new_port port =
  let (port, s) = create_port port in 
  wait_on_socket s; port, s
;;

let new_port port =
  if List.mem_assoc port !links then
    failwith "Attempt to bind twice the same port"
  else links := create_new_port port :: !links
;;
    

let local_location = Jprims.create_first_location local_space () ();;
let _ =
  Jprims.set_location local_location;
  wait_on_socket (snd first_link);;
