/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: threads.c,v 1.5 1998/05/11 07:28:27 lefessan Exp $ */

#include "callback.h"
#include "config.h"
#include "misc.h"
#include "mlvalues.h"
#include "stacks.h"
#include "fail.h"
#include "io.h"
#include "roots.h"
#include "alloc.h"
#include "memory.h"
#include "signals.h"
#include "gc_ctrl.h"
#include "join.h"
#include "intext.h"
#include "threads.h"
#include <stdio.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#ifdef HAS_UNISTD
#include <unistd.h>
#endif
#ifdef HAS_SYS_SELECT_H
#include <sys/select.h>
#endif

value thread_set_location(value location)    /* ML */
{
  thread_t th = curr_thread;

  do {
    Assign(th->location,location);
    th->stub = Val_unit;
    th = th->next;
  }
  while(th!=curr_thread);
  return Val_unit;
}

value thread_self_location(value unit)    /* ML */
{
  return (value)curr_thread->location;
}

#define th_offset(field)  Val_int(th->field - th->stack_low)
#define offset(field)     th->stack_low + Int_val(Field(intern,field))

#define Stack_tag 109
/* Attention aux trap_sp: chaque trap contient un pointeur direct vers le trap
suivant ... */


value marshal_thread(value thread_stub)    /* ML */
{
   thread_t th = (thread_t) Field(thread_stub,STUB_LOCAL);
   int stacksize,i;
   value intern,*sp,*trap_sp;

   Assign(Field(thread_stub,STUB_LOCAL),Val_unit);
  /* on verifie que le thread est bien marshallable */
   if(th == curr_thread ||
     (th->status != RUNNABLE && th->status != SUSPENDED && 
      th->status != BLOCKED_FAIL))
     { failwith("Unable to marshall thread"); }

   /* on alloue un bloc pour la pile et les infos */
   Begin_root(th);
   stacksize = th->stack_high - th->sp;
   DPRINT1("THREAD : %d\n",Int_val(th->ident));
   DPRINT1("MARSHAL THREAD STACK : size %d\n",stacksize);
   intern = alloc_shr(THREAD_MARSH_SIZE + stacksize,Stack_tag);
   End_roots();

   /* on affecte l'etat du thread */
   Init(Field(intern,THREAD_LOCATION),(value)(th->location));
   Init(Field(intern,THREAD_TRAP_OFFSET),th_offset(trapsp));
   Init(Field(intern,THREAD_STACKTHRESHOLD),th_offset(stack_threshold));
   Init(Field(intern,THREAD_STACKSIZE),th_offset(stack_high));
   Init(Field(intern,THREAD_SYNC),th->sync);
   if(th->status == BLOCKED_FAIL)
     {Init(Field(intern,THREAD_JOINING),th->joining);}
   else
     { Field(intern,THREAD_JOINING) = Val_unit; }
   Field(intern,THREAD_STATUS) = th->status;

   /* on copie la pile */
   sp = th->sp;
   trap_sp = th->trapsp;
   for(i=THREAD_MARSH_SIZE;sp < th->stack_high;) 
     {
       if(sp == trap_sp){
	 trap_sp = Trap_link(sp);
	 Trap_link(sp) = (value*)Val_int(Trap_link(sp) - th->stack_low);
       }
       DPRINT1("Stack : %lx\n",*sp);
       if(Is_block(*sp) && (Is_in_heap(*sp) || Is_young(*sp)))
	  DPRINT1("Tag ---> %d\n",Tag_val(*sp));
       Init(Field(intern,i++),(*sp++));
     }

   /* on supprime le thread du runtime courant */
   
   /* Remove thread from the doubly-linked list */
   Assign(th->prev->next, th->next);
   Assign(th->next->prev, th->prev);
   
   free_stack(th);

   th->stack_low = NULL;
   th->stack_high = NULL;
   th->stack_threshold = NULL;
   th->sp = NULL;
   th->trapsp = NULL;
   
   return intern;
}

value unmarshal_thread(value paire)
{
  thread_t th;
  int stacksize,i;
  value *sp,*trap_sp,*p;
  value intern;

  /* Allocate the thread and its stack */
  Begin_root(paire);
  th = (thread_t) alloc_shr(sizeof(struct thread_struct) / sizeof(value),
			    Local_tag);
  End_roots();
  th->ident = next_ident;
  next_ident = Val_int(Int_val(next_ident) + 1);

  intern = Field(paire,1);
  stacksize = Int_val(Field(intern,THREAD_STACKSIZE));


  if(stacksize == Thread_stack_size / sizeof(value))
    th->stack_low = get_stack();
  else
    th->stack_low =  (value *) stat_alloc(stacksize*sizeof(value));

  th->stack_high = th->stack_low + stacksize;
  DPRINT2("UNMARSHAL THREAD: stack %lx-%lx\n",th->stack_low ,th->stack_high);
  th->stack_threshold = th->stack_low + Int_val(Field(intern,
						      THREAD_STACKTHRESHOLD));
  th->sp = th->stack_low + (stacksize + THREAD_MARSH_SIZE
			    - Wosize_val(intern));
  th->trapsp = th->stack_low + Int_val(Field(intern,THREAD_TRAP_OFFSET));
  th->sync = Field(intern,THREAD_SYNC);

  /* On recopie la pile */
  sp = th->sp;
  trap_sp = th->trapsp;
  for(i=THREAD_MARSH_SIZE;sp < th->stack_high;) *sp++ = Field(intern,i++);
  for(p=trap_sp;p<th->stack_high;p=Trap_link(p))
    Trap_link(((value)p))
      = th->stack_low +
      Int_val(
	      (value)Trap_link((value)p));

  Init(th->joining,Field(intern,THREAD_JOINING));
  th->status = Field(intern,THREAD_STATUS);;
  th->reused = Val_unit; /* to be reused */
  
  th->readfds = NO_FDS;
  th->writefds = NO_FDS;
  th->exceptfds = NO_FDS;
  th->delay = NO_DELAY;
  th->waitpid = NO_WAITPID;
  th->retval = Val_unit;
  Init(th->location,Field(intern,THREAD_LOCATION));
  Init(th->stub,Field(paire,0));
  Assign(Field(th->stub,STUB_LOCAL),th);
  
  /* Insert thread in doubly linked list of threads */
  Init(th->prev,(value) curr_thread->prev);
  Init(th->next,(value) curr_thread);
  Assign(curr_thread->prev->next, th);
  Assign(curr_thread->prev, th);

  return Val_unit;
}

value go_thread;
value set_go_thread(value thread) /* ML */
{
  DPRINT1("Set go thread %lx\n",thread);
  go_thread = thread;
  register_global_root(&go_thread);

  return Val_unit;
}

/* Quand un thread demande un go (), on doit immediatement scheduler 
le thread_go avant qu'un autre thread risque de faire un autre go() ou
que ce thread effectue l'instruction suivant le go ().
*/
value schedule_go_thread(value retval) /* ML */
{
  thread_t run_thread;
  
  /* Save the status of the current thread */
  curr_thread->stack_low = stack_low;
  curr_thread->stack_high = stack_high;
  curr_thread->stack_threshold = stack_threshold;
  curr_thread->sp = extern_sp;
  curr_thread->trapsp = trapsp;

  check_callback("schedule_go_thread");

  run_thread = (thread_t) go_thread;

  /* Free everything the thread was waiting on */
  Assign(run_thread->status, RUNNABLE);
  Assign(run_thread->readfds, NO_FDS);
  Assign(run_thread->writefds, NO_FDS);
  Assign(run_thread->exceptfds, NO_FDS);
  Assign(run_thread->delay, NO_DELAY);
  Assign(run_thread->joining, NO_JOINING);
  run_thread->waitpid = NO_WAITPID;

  /* Activate the thread */
  curr_thread = run_thread;
  stack_low = curr_thread->stack_low;
  stack_high = curr_thread->stack_high;
  stack_threshold = curr_thread->stack_threshold;
  extern_sp = curr_thread->sp;
  trapsp = curr_thread->trapsp;

  return retval;
}

value extract_location_threads (value location) /* ML */
{
  thread_t th;
  int n;
  value array;

  th = curr_thread->next;
  if(curr_thread->location == location) n=1; else n=0;
  while(th!=curr_thread){
    if(th->location == location){ n++;}
    th = th->next;
  }
  if (n==0) { return Atom(0); }
  
  Begin_roots3(location,array,th);
  array = obj_block(0,Val_int(n));
  DPRINT1("n =  %d\n",n);
  DPRINT1("Nbr threads: %d\n",Wosize_val(array));
  th = curr_thread->next;
  n=0;
  if(curr_thread->location == location) {
    Assign(Field(array,n++),get_thread_stub(curr_thread));
  }
  while(th!= (thread_t)curr_thread){
    if(th->location == location){ 
      DPRINT1("Extract one thread %d\n",Int_val(th->ident));
      Assign(Field(array,n++),get_thread_stub(th));
    }
    th = th->next;
  }
  End_roots();
  DPRINT1("Nbr threads %d\n",Wosize_val(array));
 return array;
}

/*

Channel.ml:
external get_thread_stub: unit -> thread_stub = "stub_thread_get"
external wait_on_stub: unit -> 'a = "stub_thread_wait"
external reply_stub: thread -> 'a -> unit = "stub_thread_reply"

*/

/* schedule a thread: Should have a CAML Wrapper */

value stub_thread_wait(value unit)
{
  /*
  Assert(curr_thread != NULL);
  DPRINT1("Thread %d waiting for reply\n",curr_thread->ident);
  */
  set_critical_section(Val_false);
  if(callback_depth>0)
    failwith("stub_thread_wait: can not schedule");
  curr_thread->status = BLOCKED_REPLY;
  return schedule_thread();
}

/* WARNING: ici, on reveille le thread. Il ne peut donc pas s'agir du
thread principal, a moins de se trouver en section critique ! (sinon
la reponse est perdue si ce thread est reschedule avant de mourir)
*/
value stub_thread_reply(value thread_stub,value reply)
{
  thread_t th = (thread_t)Field(thread_stub,STUB_LOCAL);
  /*
  DPRINT1("Stub reply to %d\n",th->ident);
  if(th->status != BLOCKED_REPLY){
    DPRINT1("Reply to thread %d but not waiting for reply ...\n",th->ident);
  };
  */
  th->status = RUNNABLE;
  Assign(th->retval,reply);
  DPRINT1("Set reply: %ld",th->retval);

  return Val_unit;
}


/* set_local: 'a -> unit */

value set_local(value v)
{
  Tag_val(v) = Local_tag;

  return v;
}

value print_stack(value unit)
{
  /*  DPRINT1("CURRENT STACK : %d\n",extern_sp); */
  return Val_unit;
}

value print_stats(value unit) /* ML */
{
  printf("Nbre de threads crees : %ld\n",next_ident);
  printf("Nbre de piles allouees : %ld\n",stack_allouees);
  return Val_unit;
}

value local_thread_reuse(value location,value clos,value args
			 ,thread_t thread) /* ML */
{
  thread_t th=(thread_t) thread;

  DPRINT("Reuse thread ...\n");

  /* on va empiler sur la pile les anciens parametres du thread,
     placer dans thread->reused l'offset de ces parametres, puis lancer
     la nouvelle closure ... */

  if(th == curr_thread){
    DPRINT("Reusing current thread");
    DPRINT1("Reuse stacklow %lx\n",stack_low);
    curr_thread->stack_low = stack_low;
    curr_thread->stack_high = stack_high;
    curr_thread->stack_threshold = stack_threshold;
    curr_thread->sp = extern_sp;
    curr_thread->trapsp = trapsp;
  }
  if((value)th->location == location && th->sp-6> th->stack_threshold){
    /* in the same location, we don't need to make a big frame in the stack
     */
    /*
    if(th->status != BLOCKED_REPLY)
      failwith("Reuse unblocked thread !!!");
      */
    
    DPRINT("1");
    DPRINT1("Stack sp : %lx\n",th->sp);
    DPRINT1("Stack diff : %lx\n",((value)th->stack_high)-((value)th->sp));
    (th->sp)[-1] = th->sync;
    th->sync = Val_int(((value*)(th->stack_high))-((value*)(th->sp)));
    (th->sp)[-2] = Val_int(((value*)(th->stack_high))-((value*)(th->trapsp)));;
    DPRINT1("OLD TRAP : %d\n", Int_val((th->sp)[-2]));
    th->sp -=2;

    DPRINT("2");
    /* now, we fake the same call as in thread_new */
    th->sp -= 5;
    th->sp[4] = args;
    th->sp[0] = Val_unit;
    th->sp[1] = (value) Code_val(clos);
    th->sp[2] = clos;
    th->sp[3] = Val_long(0);      /* no extra args */
    /* Fake a C call frame */
    th->sp--;
    th->sp[0] = Val_unit;         /* a dummy environment */
    /*
    if(th->sp <= th->stack_threshold){
      DPRINT("STack too small");
      if(th == curr_thread){
	DPRINT("Reallocating stack for curr_thread\n");
	extern_sp = th->sp;
	realloc_stack();
	th->sp = extern_sp;
	th->trapsp = trapsp;
      }
    }
    */
    if(th == curr_thread) extern_sp = th->sp;
    th->status = RUNNABLE;
    /*    
	  Necessaire en cas de reallocation de la pile, on ne peut garder:
	  th->trapsp = th->stack_high; */
    th->retval = Val_unit;
    DPRINT("3");
    /* exceptions should be caught by other ways */
  } else {
    /* not implemented */
    gen_thread_new(location,clos,Val_true,args);
  }
  
  DPRINT("4");
  return Val_unit;
}

/* 
   This function is used when a thread makes a local synchronous call.
   As the thread might be blocked, it must save its modifiable data and
   reread them after the call.
 */

value trigger_def_and_wait(value params) /* ML */
{
  DPRINT("trigger_def_and_wait\n");
  set_critical_section(Val_false);
  curr_thread->sp = extern_sp;
  curr_thread->trapsp = trapsp;
  curr_thread->status = BLOCKED_REPLY;
  curr_thread->retval = Val_unit;
  trigger_def(params);
  extern_sp = curr_thread->sp;
  trapsp = curr_thread->trapsp;
  
  if(curr_thread->status == BLOCKED_REPLY){
    DPRINT("Schedule another thread");
    return schedule_thread();
  }else{
    DPRINT("sur le meme thread");
    return curr_thread->retval;
  }
}

value exit_trigger_def_and_wait(value params) /* ML */
{
  DPRINT("exit trigger def and wait\n");
  if(curr_thread->sync != Val_unit){ 
    /* le thread courant est le thread ayant fait
       l'appel synchrone, et qui doit donc recevoir
       la reponse */
    Begin_root(params);
    thread_kill((value)curr_thread);            /* on revient en arriere */
    End_roots();
    return trigger_def_and_wait(params); /* on relance l'appel et on attend */
  }
  /* le thread courant est termine, la reponse se fera directement au thread
     responsable de l'appel synchrone. */
  trigger_def(params);

  return thread_kill((value) curr_thread); /* et si il n'y a pas d'autre thread a schedule */
}


value exit_wait_on_stub(value unit) /* ML */
{
  if(curr_thread->sync != Val_unit){ 
    /* le thread courant est le thread ayant fait
       l'appel synchrone, et qui doit donc recevoir
       la reponse */
    thread_kill( (value) curr_thread);            /* on revient en arriere */
    curr_thread->status = BLOCKED_REPLY; /* on attend la reponse */
    return schedule_thread();
  }
  /* le thread courant est termine, la reponse se fera directement au thread
     responsable de l'appel synchrone. */
  return thread_kill( (value) curr_thread);
}

value set_server(value unit)
{
  nbre_threads+=3; /* these threads will never die :))) */
  return Val_unit;
}
