/***********************************************************************/
/*                                                                     */
/*                           Objective Caml                            */
/*                                                                     */
/*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         */
/*                                                                     */
/*  Copyright 1996 Institut National de Recherche en Informatique et   */
/*  Automatique.  Distributed only by permission.                      */
/*                                                                     */
/***********************************************************************/

/* $Id: scheduler.c,v 1.9 1998/05/11 07:28:24 lefessan Exp $ */

/* The thread scheduler */

/*

1 - Inutile de tester tous les threads a chaque appel du scheduler. On
  peut le faire une seule fois par cycle:
   - on teste tous les threads en placant ce qui peuvent tourner dans
      une runqueue.
   - on fait tourner tour a tour tous les threads de la runqueue
   - on place immediatement les threads preemptes dans le prochaine 
      runqueue
   - on place les threads en attente de terminaison (JOIN) et ceux
      suspendus (SUSPENDED) dans deux queues speciales dont ils
      sont extraits sur demande (kill ou wake_up) .

 A chaque cycle, on teste donc uniquement les threads bloques (IO, DELAY 
et GC)

2. Securiser les acces aux descripteurs de threads pour empecher la
reutilisation des threads extruder par les fonctions du module Thread
(modification de th->reused) et le transmettre lors du marshaller.

3. essayer de mixer threads Systeme et threads Bytecode: les callbacks
entre les deux se font a travers une interface speciale qui cree un
thread du bon cote et fait attendre l'autre.  Quand on fait un
callback mixte, il faut lancer un thread pour executer ce callback et
suspendre le thread courant (BLOCKED_CALLBACK). Tout le probleme
reside alors a savoir si on se trouve dans le thread POSIX executant
le bytecode ou ceux executant le code natif.


*/ 

/* il faut verifier ce fichier pas a pas pour le rendre compatible avec
le compacteur qui deplace les objets */

#include "callback.h"
#include "config.h"
#include "misc.h"
#include "mlvalues.h"
#include "stacks.h"
#include "fail.h"
#include "io.h"
#include "roots.h"
#include "alloc.h"
#include "memory.h"
#include "signals.h"
#include "gc_ctrl.h"
#include "intext.h"
#include "threads.h"
#include <stdio.h>

#if ! (defined(HAS_SELECT) && \
       defined(HAS_SETITIMER) && \
       defined(HAS_GETTIMEOFDAY) && \
       (defined(HAS_WAITPID) || defined(HAS_WAIT4)))
#include "Cannot compile libthreads, system calls missing"
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#ifdef HAS_UNISTD
#include <unistd.h>
#endif
#ifdef HAS_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifndef FD_ISSET
/* Assume old-style BSD 4.2 fd sets */
typedef int fd_set;
#define FD_SETSIZE (sizeof(int) * 8)
#define FD_SET(fd,fds) (*(fds) |= 1 << (fd))
#define FD_CLR(fd,fds) (*(fds) &= ~(1 << (fd)))
#define FD_ISSET(fd,fds) (*(fds) & (1 << (fd)))
#define FD_ZERO(fds) (*(fds) = 0)
#endif

#ifndef HAS_WAITPID
#define waitpid(pid,status,opts) wait4(pid,status,opts,NULL)
#endif

/* The thread currently active */
thread_t curr_thread = NULL;
/* Identifier for next thread creation */
value next_ident = Val_int(0);
value critical_section = Val_false;

/*-------------------- Recyclage de threads --------------*/

value **free_stacks = NULL;
int free_stacks_size = 0;
int next_free_stack = 0;

value free_threads = Val_unit;
int free_threads_size = 0;
int next_free_thread = 0;

/*
value *stacks_buffer[STACKS_BUFFER_SIZE];
int next_stack_free=0;
int first_stack_to_use=0;
*/

value stack_allouees = 0;
int nbre_threads = 0;
int terminaison = 1; /* we detect terminaison */
static int nthreads = 0;
static int  thread_alloues = 0;
static int nbre_passes = Val_int(0);

/* Scan the stacks of the other threads */

static void (*prev_scan_roots_hook) (scanning_action);
extern void byte_do_local_roots(scanning_action,value*,value*,value);

static void thread_scan_roots(scanning_action action)
{
  thread_t th,start;

  check_runqueue();
  
  /* Scan all active descriptors */
  start = curr_thread;
  (*action)((value) curr_thread, (value *) &curr_thread);
  /* Don't scan curr_thread->sp, this has already been done.
     Don't scan local roots either, for the same reason. */
  for (th = start->next; th != start; th = th->next) {
    byte_do_local_roots(action, th->sp, th->stack_high, (value)NULL);
  }
  /* Hook */
  if (prev_scan_roots_hook != NULL) (*prev_scan_roots_hook)(action);
}

/* Initialize the thread machinery */
static long last_gc;

value thread_initialize(value unit)       /* ML */
{
  struct itimerval timer;
  /* Create a descriptor for the current thread */
  last_gc = stat_major_collections;
  register_global_root(&free_threads);

  curr_thread =
    (thread_t) alloc_shr(sizeof(struct thread_struct) / sizeof(value),
			 Local_tag);
  thread_alloues++;
  
  curr_thread->ident = next_ident;
  next_ident = Val_int(Int_val(next_ident) + 1);
  curr_thread->next = curr_thread;
  curr_thread->prev = curr_thread;
  curr_thread->stack_low = stack_low;
  curr_thread->stack_high = stack_high;
  curr_thread->stack_threshold = stack_threshold;
  curr_thread->sp = extern_sp;
  curr_thread->trapsp = trapsp;
  curr_thread->status = RUNNABLE;
  curr_thread->readfds = NO_FDS;
  curr_thread->writefds = NO_FDS;
  curr_thread->exceptfds = NO_FDS;
  curr_thread->delay = NO_DELAY;
  curr_thread->joining = NO_JOINING;
  curr_thread->waitpid = NO_WAITPID;
  curr_thread->retval = Val_unit;

  curr_thread->stub = Val_unit;
  curr_thread->location= Val_unit;
  curr_thread->reused = Val_unit;
  curr_thread->sync = Val_unit;
  curr_thread->passe = Val_int(0);
  /* Initialize GC */
  prev_scan_roots_hook = scan_roots_hook;
  scan_roots_hook = thread_scan_roots;
  /* Initialize interval timer */
  timer.it_interval.tv_sec = 0;
  timer.it_interval.tv_usec = Thread_timeout;
  timer.it_value = timer.it_interval;
  setitimer(ITIMER_VIRTUAL, &timer, NULL);

  check_runqueue();
  
  return Val_unit;
}

/* Create a thread */


value *get_stack()
{
  if(next_free_stack>0){
    return  free_stacks[--next_free_stack];
  } else { 
    stack_allouees++;
    return (value *) stat_alloc(Thread_stack_size);
  }
}
value gen_thread_new(value location,value clos,
		     value param,value clos_arg) /* ML */
{
  thread_t th;
  /* Allocate the thread and its stack */

  nbre_threads++;

  /* get a new thread */
  if(next_free_thread>0){
    th = (thread_t) Field(free_threads,--next_free_thread);
  } else {
    Begin_roots3(location,clos,clos_arg);
    th = (thread_t) alloc_shr(sizeof(struct thread_struct) / sizeof(value),
			      Local_tag);
    End_roots();
  }
  thread_alloues++;
  th->stub = Val_unit;
  Init(th->location,location);
  th->reused = Val_unit;
  th->sync = Val_unit;
  th->ident = next_ident;
  next_ident = next_ident + 2;
  DPRINT2("NEW THREAD %d with %x\n",Int_val(th->ident),location);

  if(next_free_stack>0){
    th->stack_low = free_stacks[--next_free_stack];
  } else { 
    stack_allouees++;
    th->stack_low = (value *) stat_alloc(Thread_stack_size);
  }

  DPRINT1("new thread : stacklow %lx\n", th-> stack_low);

  th->stack_high = th->stack_low + Thread_stack_size / sizeof(value);
  th->stack_threshold = th->stack_low + Stack_threshold / sizeof(value);
  th->sp = th->stack_high;
  /*  DPRINT1("NEW THREAD stack high : %d\n",th->sp);*/
  th->trapsp = th->stack_high;
  /* Set up a return frame that pretends we're applying clos to ().
     This way, when this thread is activated, the RETURN will take us
     to the entry point of the closure. */
  if(param == Val_false)
    th->sp -= 4;
  else {
    th->sp -= 5;
    Assign(th->sp[4],clos_arg);
  }
  th->sp[0] = Val_unit;
  th->sp[1] = (value) Code_val(clos);
  th->sp[2] = clos;
  th->sp[3] = Val_long(0);      /* no extra args */
  /* Fake a C call frame */
  th->sp--;
  th->sp[0] = Val_unit;         /* a dummy environment */
  /* The thread is initially runnable */
  th->status = RUNNABLE;
  th->readfds = NO_FDS;
  th->writefds = NO_FDS;
  th->exceptfds = NO_FDS;
  th->delay = NO_DELAY;
  th->joining = NO_JOINING;
  th->waitpid = NO_WAITPID;
  th->retval=Val_unit;
  th->passe = Val_int(0);

  /* Insert thread in doubly linked list of threads */
  Init(th->prev,(value) curr_thread->prev);
  Init(th->next ,(value) curr_thread);
  Assign(curr_thread->prev->next, th);
  Assign(curr_thread->prev, th);

  /*  DPRINT("NEW THREAD inserted\n"); */

  check_runqueue();
  
  /* Return thread */
  return (value) th;
}

value thread_new(value clos)          /* ML */
{
  return gen_thread_new((value)curr_thread->location,clos,Val_false,Val_unit);
}

value gen_thread_in(value location,value clos)          /* ML */
{
  return gen_thread_new(location,clos,Val_false,Val_unit);
}

/* Return the thread identifier */

value thread_id(value th)             /* ML */
{
  return ((struct thread_struct *)th)->ident;
}

/* Return the current time as a floating-point number */

static double timeofday(void)
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return (double) tv.tv_sec + (double) tv.tv_usec * 1e-6;
}


/* Find a runnable thread and activate it */

#define FOREACH_THREAD(x) x = curr_thread; do { x = x->next;
#define END_FOREACH(x) } while (x != curr_thread)

static value alloc_process_status(int pid, int status);
static void add_fdlist_to_set(value fdl, fd_set *set);
static value inter_fdlist_set(value fdl, fd_set *set);


value schedule_thread(void)
{
  thread_t run_thread, th;
  fd_set readfds, writefds, exceptfds;
  double delay, now;
  int need_select, need_wait;

  /* Don't allow preemption during a callback */
  check_runqueue();
  
  if (callback_depth > 0){  
    DPRINT1( "thread_schedule exited: callback depth : %d \n",callback_depth);
    return curr_thread->retval;
  }
  if(critical_section==Val_true) {
    DPRINT("Scheduler exited in critical section\n"); 
    /*
    if(curr_thread->status & BLOCKED_IO){
      DPRINT("Select in critical section\n");
    }
    */
    fatal_error("Enter scheduler in critical section");
    
    return curr_thread->retval;
  }

  /* faire cette loop a chaque passage dans le scheduler est inacceptable.
     On utilise donc une information "passe" stockee dans
     chaque thread et un compteur de passe. Lorsque le meme "passe" est 
     retrouve sur le thread courant et sur le compteur, il faut faire une
     nouvelle passe... */

  /* Save the status of the current thread */
  curr_thread->stack_low = stack_low;
  curr_thread->stack_high = stack_high;
  curr_thread->stack_threshold = stack_threshold;
  curr_thread->sp = extern_sp;
  curr_thread->trapsp = trapsp;

  DPRINT1("sched stacklow %lx\n",stack_low);

  if(curr_thread->passe == nbre_passes){

try_again:
  nbre_passes++;
  /* Build fdsets and delay for select.
     See if some join or wait operations succeeded. */
  FD_ZERO(&readfds);
  FD_ZERO(&writefds);
  FD_ZERO(&exceptfds);
  delay = DELAY_INFTY;
  now = -1.0;
  need_select = 0;
  need_wait = 0;

  FOREACH_THREAD(th)
    if(th->status & BLOCKED_ANY){
      if (th->status & (BLOCKED_IO - 1)) {
        add_fdlist_to_set(th->readfds, &readfds);
        add_fdlist_to_set(th->writefds, &writefds);
        add_fdlist_to_set(th->exceptfds, &exceptfds);
        need_select = 1;
      }
      if (th->status & (BLOCKED_DELAY - 1)) {
        double th_delay;
        if (now < 0.0) now = timeofday();
        th_delay = Double_val(th->delay) - now;
        if (th_delay <= 0) {

          th->status = RUNNABLE;
          Assign(th->retval, RESUMED_DELAY);
	  Assign(th->readfds, NO_FDS);
	  Assign(th->writefds, NO_FDS);
	  Assign(th->exceptfds, NO_FDS);
	  Assign(th->delay, NO_DELAY);        
	} else {
          if (th_delay < delay) delay = th_delay;
        }
      }
      if (th->status & (BLOCKED_JOIN - 1)) {
        if (((thread_t)(th->joining))->status == KILLED) {
          th->status = RUNNABLE;
	  Assign(th->joining, NO_JOINING);
	  Assign(th->retval, RESUMED_JOIN);
        }
      }
      else if (th->status & (BLOCKED_WAIT - 1)) {
        int status, pid;
        pid = waitpid(Int_val(th->waitpid), &status, WNOHANG);
        if (pid > 0) {
          th->status = RUNNABLE;
          th->waitpid = NO_WAITPID;
	  Begin_root(th); /*++*/
          Assign(th->retval, alloc_process_status(pid, status));
          End_roots();/*++*/
        } else {
          need_wait = 1;
        }
      } 
      else if (th->status & (BLOCKED_GC - 1)){
        if(stat_major_collections>last_gc){
          th->status = RUNNABLE;
          Assign(th->retval,Val_unit);
        } 
      }
      else if (th->status & (BLOCKED_FAIL - 1)){
        if( Field(th->joining,STUB_POS) == Val_unit){
          th->status = RUNNABLE;
          Assign(th->retval,Val_unit);
          Assign(th->joining, NO_JOINING);
        }
      }
    }
  END_FOREACH(th);

  last_gc = stat_major_collections;

  /* Find if a thread is runnable. */
  run_thread = NULL;
  FOREACH_THREAD(th)
    if (th->status == RUNNABLE) { run_thread = th; break; }
  END_FOREACH(th);

  /* Do the select if needed */
  if (need_select || run_thread == NULL) {
    struct timeval delay_tv, * delay_ptr;
    int retcode;
    /* If a thread is blocked on wait, don't block forever */
    if (need_wait && delay > Thread_timeout * 1e-6) {
      delay = Thread_timeout * 1e-6;
    }
    /* Convert delay to a timeval */
    /* If a thread is runnable, just poll */
    if (run_thread != NULL) {
      delay_tv.tv_sec = 0;
      delay_tv.tv_usec = 0;
      delay_ptr = &delay_tv;
    }
    else if (delay != DELAY_INFTY) {
      delay_tv.tv_sec = (unsigned int) delay;
      delay_tv.tv_usec = (delay - (double) delay_tv.tv_sec) * 1E6;
      delay_ptr = &delay_tv;
    }
    else {
      delay_ptr = NULL;
    }
    enter_blocking_section(); /* alloc */
    retcode = select(FD_SETSIZE, &readfds, &writefds, &exceptfds, delay_ptr);
    leave_blocking_section(); /* alloc */
    if (retcode > 0) {
      /* Some descriptors are ready. 
         Mark the corresponding threads runnable. */
      FOREACH_THREAD(th)
        if (th->status & (BLOCKED_IO - 1)) {
	  value r = Val_unit, w = Val_unit, e = Val_unit;
          Begin_roots4(th,r,w,e);
            r = inter_fdlist_set(th->readfds, &readfds);
	    w = inter_fdlist_set(th->writefds, &writefds);
	    e = inter_fdlist_set(th->exceptfds, &exceptfds);
	    if (r != NO_FDS || w != NO_FDS || e != NO_FDS) {
	      value retval = alloc(3, TAG_RESUMED_IO);
	      Field(retval, 0) = r;
	      Field(retval, 1) = w;
	      Field(retval, 2) = e;
	      Assign(th->retval, retval);
	      Assign(th->readfds, NO_FDS);
	      Assign(th->writefds, NO_FDS);
	      Assign(th->exceptfds, NO_FDS);
	      Assign(th->delay, NO_DELAY);	      
	      th->status = RUNNABLE;
	      if (run_thread == NULL) run_thread = th; /* Found one. */
	    }
          End_roots();
        }
      END_FOREACH(th);
    }
    /* If we get here with run_thread still NULL, one of the following
       may have happened:
       - a delay has expired
       - a wait() needs to be polled again
       - the select() failed (e.g. was interrupted)
       In these cases, we go through the loop once more to make the
       corresponding threads runnable. */
    if (run_thread == NULL &&
        (delay != DELAY_INFTY || need_wait || retcode == -1))
      goto try_again;
  }
  } else {
    /* on recherche un thread RUNNABLE avec th->passe <> nbre_passes */

    curr_thread->passe = nbre_passes;
    run_thread = curr_thread->next;
    while(run_thread->passe != nbre_passes){
      if(run_thread->status == RUNNABLE) break;
      run_thread->passe = nbre_passes;
      run_thread = run_thread->next;
    }
    if(run_thread->passe == nbre_passes) goto try_again;
  }
  /* If we haven't something to run at that point, we're in big trouble. */
  if (run_thread == NULL) invalid_argument("Thread: deadlock");

  /* Free everything the thread was waiting on */

  /* Activate the thread */
  curr_thread = run_thread;
  DPRINT1("Scheduling to thread : %d\n",Int_val(curr_thread->ident)); 

  stack_low = curr_thread->stack_low;
  stack_high = curr_thread->stack_high;
  stack_threshold = curr_thread->stack_threshold;
  extern_sp = curr_thread->sp;
  trapsp = curr_thread->trapsp;

  if(extern_sp<stack_threshold) {
    DPRINT("Reallocating stack for new scheduled thread\n");
    realloc_stack();
  }
  check_runqueue();
  
  return curr_thread->retval;
}

/* Reschedule without suspending the current thread */

value set_critical_section(value section) /* ML */
{
  value old = critical_section;
  
  critical_section = section;
  /*
  if(critical_section==Val_true)
    DPRINT("--------------  ENTERING CRITICAL SECTION\n");
  else
    DPRINT("--------------  LEAVING CRITICAL SECTION\n");
  */
  check_runqueue();
  
  return old;
}

value thread_yield(value unit)        /* ML */
{
  Assert(curr_thread != NULL);
  if(critical_section == Val_true)
    return Val_unit;
  curr_thread->retval = Val_unit;
  return schedule_thread();
}

/* Suspend the current thread */

value thread_sleep(value unit)        /* ML */
{
  Assert(curr_thread != NULL);
  check_callback("thread_sleep");
  curr_thread->status = SUSPENDED;
  return schedule_thread();
}

/* Suspend the current thread on a select() request */

value thread_select(value arg)        /* ML */
{
  double date;
  /* Don't do an error if we're not initialized yet
     (we can be called from thread-safe Pervasives before initialization),
     just return immediately. */
  if (curr_thread == NULL) return RESUMED_WAKEUP;
  /* si on est dans un callback, on doit bloquer jusqu'a la fin du select */
  Assign(curr_thread->readfds, Field(arg, 0));
  Assign(curr_thread->writefds, Field(arg, 1));
  Assign(curr_thread->exceptfds, Field(arg, 2));
  date = Double_val(Field(arg, 3));
  if (date >= 0.0) {
    date += timeofday();
    Assign(curr_thread->delay, copy_double(date));
    curr_thread->status = BLOCKED_IO | BLOCKED_DELAY;
  } else {
    curr_thread->status = BLOCKED_IO;
  }
  if(callback_depth==0 && critical_section==Val_false)
    return schedule_thread();
  else {
    fd_set readfds, writefds, exceptfds;
    double delay, now;
    int need_select, need_wait;

    fprintf(stderr,"Warning: blocked in callback\n");

  select_try_again:
    /* Build fdsets and delay for select.
    See if some join or wait operations succeeded. */
    FD_ZERO(&readfds);
    FD_ZERO(&writefds);
    FD_ZERO(&exceptfds);
    delay = DELAY_INFTY;
    now = -1.0;
    need_select = 0;
    need_wait = 0;

    add_fdlist_to_set(curr_thread->readfds, &readfds);
    add_fdlist_to_set(curr_thread->writefds, &writefds);
    add_fdlist_to_set(curr_thread->exceptfds, &exceptfds);
    need_select = 1;

    if (curr_thread->status & (BLOCKED_DELAY - 1)) {
      double th_delay;
      if (now < 0.0) now = timeofday();
      th_delay = Double_val(curr_thread->delay) - now;
      if (th_delay <= 0) {
        curr_thread->status = RUNNABLE;
        curr_thread->retval = RESUMED_DELAY;
      } else {
        if (th_delay < delay) delay = th_delay;
      }
    }
    if (curr_thread->status != RUNNABLE){
      struct timeval delay_tv, * delay_ptr;
      int retcode;

      /* Convert delay to a timeval */
      if (delay != DELAY_INFTY) {
        delay_tv.tv_sec = (unsigned int) delay;
        delay_tv.tv_usec = (delay - (double) delay_tv.tv_sec) * 1E6;
        delay_ptr = &delay_tv;
      }
      else {
        delay_ptr = NULL;
      }

      enter_blocking_section();
      retcode = select(FD_SETSIZE, &readfds, &writefds, &exceptfds, delay_ptr);
      leave_blocking_section();

      if (retcode > 0) {
        /* Some descriptors are ready. 
        Mark the corresponding threads runnable. */
        value r = Val_unit, w = Val_unit, e = Val_unit;
        Begin_roots3(r,w,e);
        r = inter_fdlist_set(curr_thread->readfds, &readfds);
        w = inter_fdlist_set(curr_thread->writefds, &writefds);
        e = inter_fdlist_set(curr_thread->exceptfds, &exceptfds);
        if (r != NO_FDS || w != NO_FDS || e != NO_FDS) {
          value retval = alloc(3, TAG_RESUMED_IO);
          Field(retval, 0) = r;
          Field(retval, 1) = w;
          Field(retval, 2) = e;
          Assign(curr_thread->retval, retval);
          curr_thread->status = RUNNABLE;
        }
        End_roots();
      }
      /* If we get here with run_thread still NULL, one of the following
      may have happened:
      - a delay has expired
       - a wait() needs to be polled again
       - the select() failed (e.g. was interrupted)
       In these cases, we go through the loop once more to make the
       corresponding threads runnable. */
      if (curr_thread->status!= RUNNABLE &&
          (delay != DELAY_INFTY || retcode == -1))
        goto select_try_again;
    }

    /* If we haven't something to run at that point, we're in big trouble. */
    if (curr_thread->status != RUNNABLE) invalid_argument("Thread: deadlock");

    /* Free everything the thread was waiting on */
    Assign(curr_thread->readfds, NO_FDS);
    Assign(curr_thread->writefds, NO_FDS);
    Assign(curr_thread->exceptfds, NO_FDS);
    Assign(curr_thread->delay, NO_DELAY);
    Assign(curr_thread->joining, NO_JOINING);
    curr_thread->waitpid = NO_WAITPID;
    
    return curr_thread->retval;
  }
}

/* Primitives to implement suspension on buffered channels */

value thread_inchan_ready(value vchan) /* ML */
{
  struct channel * chan = Channel(vchan);
  return Val_bool(chan->curr < chan->max);
}

value thread_outchan_ready(value vchan, value vsize) /* ML */
{
  struct channel * chan = Channel(vchan);
  long size = Long_val(vsize);
  /* Negative size means we want to flush the buffer entirely */
  if (size < 0) {
    return Val_bool(chan->curr == chan->buff);
  } else {
    int free = chan->end - chan->curr;
    if (chan->curr == chan->buff)
      return Val_bool(size < free);
    else
      return Val_bool(size <= free);
  }
}

/* Suspend the current thread for some time */

value thread_delay(value time)          /* ML */
{
  double date = timeofday() + Double_val(time);
  Assert(curr_thread != NULL);
  check_callback("thread_delay");
  curr_thread->status = BLOCKED_DELAY;
  Assign(curr_thread->delay, copy_double(date));
  return schedule_thread();
}

/* Suspend the current thread until another thread terminates */

value thread_join(value th)          /* ML */
{
  check_callback("thread_join");
  Assert(curr_thread != NULL);
  if (((thread_t)th)->status == KILLED) return Val_unit;
  curr_thread->status = BLOCKED_JOIN;
  Assign(curr_thread->joining, th);
  return schedule_thread();
}

/* Suspend the current thread until a Unix process exits */

value thread_wait_pid(value pid)          /* ML */
{
  Assert(curr_thread != NULL);
  check_callback("thread_wait_pid");
  curr_thread->status = BLOCKED_WAIT;
  curr_thread->waitpid = pid;
  return schedule_thread();
}

/* Suspend the current thread until a major GC finishes */

value thread_wait_gc(value unit)          /* ML */
{
  Assert(curr_thread != NULL);
  check_callback("thread_wait_gc");
  curr_thread->status = BLOCKED_GC;
  return schedule_thread();
}

/* Reactivate another thread */

value thread_wakeup(value thread)     /* ML */
{
  thread_t th = (thread_t) thread;
  switch (th->status) {
  case SUSPENDED:
    th->status = RUNNABLE;
    th->retval = RESUMED_WAKEUP;
    break;
  case KILLED:
    DPRINT("Thread.wakeup: killed thread");
  case RUNNABLE:
  case BLOCKED_IO:
    break;
  default:
  }
  return Val_unit;
}

/* Return the current thread */

value thread_self(value unit)         /* ML */
{
  Assert(curr_thread != NULL);
  return (value) curr_thread;
}

/* recyclage des piles de threads */
void free_stack(thread_t th)
{
  if(th->stack_high == th->stack_low + Thread_stack_size / sizeof(value)){
    /* is there a free slot in the buffer for this stack ? */
    if(next_free_stack>=free_stacks_size){
      /* we must realloc the stack buffer */
      value **old_stacks = free_stacks;
      int old_size = free_stacks_size;
      int i;

      free_stacks_size = 50 + free_stacks_size;
      free_stacks = (value**)stat_alloc(sizeof(value*)*free_stacks_size);
      
      if(free_stacks == NULL) exit(3);

      for(i=0;i<old_size;i++) free_stacks[i] = old_stacks[i];
      stat_free((char*)old_stacks);
    }
    free_stacks[next_free_stack++] = th->stack_low;
  } else
    stat_free((char *) th->stack_low);
}

/* recyclage des structures de threads */
void free_thread(thread_t th)
{
  if(th->reused == Val_unit){
    /* is there a free slot in the buffer for this thread ? */
    if(next_free_thread>=free_threads_size){
      /* we must realloc the thread buffer */
      value old_threads = free_threads;
      int old_size = free_threads_size;
      int i;

      free_threads_size = 50+free_threads_size;

      Begin_roots2(old_threads,th);
      Assign(free_threads, alloc_shr(free_threads_size,0));
      End_roots();

      for(i=0;i<old_size;i++) {
	Init(Field(free_threads,i),Field(old_threads,i));
      }
      for(;i<free_threads_size;i++)
	Field(free_threads,i) = Val_unit;
    }
    Assign(Field(free_threads,next_free_thread),th);
    next_free_thread++;
  }
}

/* recyclage des stubs de threads (pour le GC dist) */
void free_thread_stub(thread_t th)
{
  if(th->stub != Val_unit){
    if(Field(th->stub,STUB_LOCATOR) != Val_unit){
      value location = th->location;
      local_location_t lloc = (local_location_t)Field(location,STUB_LOCAL);
      
      if(lloc->next_free_stub >= lloc->free_stubs_size){
	/* we must realloc the stubs buffer */
	value old_stubs = lloc->free_stubs;
	int old_size = Int_val(lloc->free_stubs_size);
	int i;
	
	lloc->free_stubs_size = 20+lloc->free_stubs_size;
	
	Begin_roots3(old_stubs,th,lloc);
	Assign(lloc->free_stubs,alloc(Int_val(lloc->free_stubs_size),0));
	End_roots();
	
	for(i=0;i<old_size;i++) {
	  Init(Field(lloc->free_stubs,i),Field(old_stubs,i));
	}
	for(;i<Int_val(lloc->free_stubs_size);i++)
	  Field(lloc->free_stubs,i) = Val_unit;
      }
      Assign(Field(lloc->free_stubs,Int_val(lloc->next_free_stub)),th->stub);
      lloc->next_free_stub+=2;
    }
    Assign(Field(th->stub,STUB_LOCAL),Val_unit);
    Assign(th->stub,Val_unit);
  }
  Assign(th->location,Val_unit);
}

/* remove a thread from the scheduler */ 
void remove_thread(thread_t th)
{
  if(curr_thread == th) {
    Begin_root(th);
    schedule_thread();
    End_roots();
  }
  
  /* Remove thread from the doubly-linked list */
  Assign(th->prev->next, th->next);
  Assign(th->next->prev, th->prev);
  
  free_stack(th);
  free_thread(th); /* alloc */
}

/* Kill a thread */
value thread_kill(value thread)       /* ML */
{
  value retval = Val_unit;
  thread_t th = (thread_t) thread;
  /* Don't paint ourselves in a corner */

  check_runqueue();
  
  if(critical_section==Val_true){
    Begin_root(th);
    set_critical_section(Val_false);
    End_roots();
  }
  
  if(th->sync!=Val_unit){
    /* we are exiting from a synchronous call ... */
    DPRINT1("exiting from a synchronous call ...%ld",Int_val(th->sync));
    if(th != curr_thread) failwith("Exit sync: th<>curr_thread");
    th->stack_high = stack_high;
    extern_sp = ((value*)(th->stack_high)) - Int_val(th->sync);
    DPRINT1("NEW TRAP : %d\n", Int_val(extern_sp[-2]));
    trapsp =((value*)(th->stack_high)) - Int_val(extern_sp[-2]);
    th->sync = extern_sp[-1];
    DPRINT1("Return %ld",th->retval);
    return th->retval;
  }
  DPRINT("killing the thread ...");
  if (th == th->next) failwith("Thread.kill: cannot kill the last thread");
  /* This thread is no longer waiting on anything */
  if (th->status == KILLED) failwith("Thread.kill: killed thread");
  th->status = KILLED;

  if(terminaison){
    /* at the end, there are three threads: the main one, one waiting
    on the port, and one for migrations */
    nbre_threads--;
    if(nbre_threads<nthreads+1){
      if(th->next->status & (BLOCKED_END-1)){
	th->next->status = RUNNABLE;
      } else
	if(th->next->next->status & (BLOCKED_END-1)){
	  th->next->next->status = RUNNABLE;
	} else
	  fprintf(stderr,"No threads waiting\n");
    }
  }

  /* If this is the current thread, activate another one */
  if (th == curr_thread){
    Begin_root(th);
    retval = schedule_thread();
    End_roots();
  }
  /* Remove thread from the doubly-linked list */
  Assign(th->prev->next, th->next);
  Assign(th->next->prev, th->prev);
  /* Free its resources */
  free_stack(th);
  th->stack_low = NULL;
  th->stack_high = NULL;
  th->stack_threshold = NULL;
  th->sp = NULL;
  th->trapsp = NULL;

  check_runqueue();
  free_thread_stub(th);
  free_thread(th);

  return retval;
}

/* Set a list of file descriptors in a fdset */

static void add_fdlist_to_set(value fdl, fd_set *set)
{
  for (/*nothing*/; fdl != NO_FDS; fdl = Field(fdl, 1)) {
    FD_SET(Int_val(Field(fdl, 0)), set);
  }
}

/* Build the intersection of a list and a fdset (the list of file descriptors
   which are both in the list and in the fdset). */

static value inter_fdlist_set(value fdl, fd_set *set)
{
  value res = Val_unit;
  value cons;

  Begin_roots2(fdl, res);
    for (res = NO_FDS; fdl != NO_FDS; fdl = Field(fdl, 1)) {
      int fd = Int_val(Field(fdl, 0));
      if (FD_ISSET(fd, set)) {
        cons = alloc(2, 0);
	Field(cons, 0) = Val_int(fd);
	Field(cons, 1) = res;
	res = cons;
	FD_CLR(fd, set); /* wake up only one thread per fd ready */
      }
    }
  End_roots();
  return res;
}

/* Auxiliary function for allocating the result of a waitpid() call */

#if !(defined(WIFEXITED) && defined(WEXITSTATUS) && defined(WIFSTOPPED) && \
      defined(WSTOPSIG) && defined(WTERMSIG))
/* Assume old-style V7 status word */
#define WIFEXITED(status) (((status) & 0xFF) == 0)
#define WEXITSTATUS(status) (((status) >> 8) & 0xFF)
#define WIFSTOPPED(status) (((status) & 0xFF) == 0xFF)
#define WSTOPSIG(status) (((status) >> 8) & 0xFF)
#define WTERMSIG(status) ((status) & 0x3F)
#endif

#define TAG_WEXITED 0
#define TAG_WSIGNALED 1
#define TAG_WSTOPPED 2

static value alloc_process_status(int pid, int status)
{
  value st, res;

  if (WIFEXITED(status)) {
    st = alloc(1, TAG_WEXITED);
    Field(st, 0) = Val_int(WEXITSTATUS(status));
  }
  else if (WIFSTOPPED(status)) {
    st = alloc(1, TAG_WSTOPPED);
    Field(st, 0) = Val_int(WSTOPSIG(status));
  }
  else {
    st = alloc(1, TAG_WSIGNALED);
    Field(st, 0) = Val_int(WTERMSIG(status));
  }
  Begin_root(st);
    res = alloc(2, TAG_RESUMED_WAIT);
    Field(res, 0) = Val_int(pid);
    Field(res, 1) = st;
  End_roots();
  return res;
}

value suspend_on_terminaison(value nth)  /* ML */
{
  nthreads = Int_val(nth);
  if(terminaison){
    if(nbre_threads<nthreads+1) return Val_unit;
    check_callback("suspend_on_terminaison");
    curr_thread->status = BLOCKED_END;
    return schedule_thread();
  }
  return Val_unit;
}

#include <errno.h>
#ifndef EAGAIN
#define EAGAIN (-1)
#endif
#ifndef EINTR
#define EINTR (-1)
#endif
#ifndef EWOULDBLOCK
#define EWOULDBLOCK (-1)
#endif

value try_again(value err)
{
  if(err == Val_int(EAGAIN) || err == Val_int(EINTR) || 
     err == Val_int(EWOULDBLOCK)) {
    return Val_true;
  }
  
  return Val_false;
}

value suspend_on_fail(value location)
{
  if( Field(location,STUB_POS) != Val_unit) {
    Assign(curr_thread->joining,location);
    curr_thread->status = BLOCKED_FAIL;
    return schedule_thread();
  }
  return Val_unit;
}

/* the argument thread should not be recycled */ 
value thread_dirty(value thread)
{
  thread_t th = (thread_t) thread;
  th->reused = Val_int(1);
  return Val_unit;
}
