(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*         Jerome Vouillon, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: mop.ml,v 1.2 1998/04/27 14:17:04 lefessan Exp $ *)

(*
open Jtypes
*)
open Obj

type class_id = Class_id of string * int
type 'a option = 'a Pervasives.option

(**** Object representation ****)

(*
  Chaque objet est une location qui contient ses threads et ses definitions.
Chaque invocation de methode provoque la creation d'un nouveau thread dans
cette location. L'ordre go loc fait migrer la location a l'interieur d'une
autre location.

Deux GetMethodes:
  L'une va chercher la methode dans obj.(5), l'autre dans obj.(6).
Pour un objet local, ces deux champs pointent sur la meme TMV. Pour un 
objet ayant migrer, obj.(6) est l'ancienne table des methodes virtuelles,
alors que obj.(5) est la nouvelle locale.

- Appels a une methode de *self* ou d'un ancetre: acces direct a la TMV 
  d'origine. GETANCMETHOD
   Lsend(Ancestor(Lvar selfid),...)

- Appels a une methode d'un objet inconnu dans un module non migrable: 
 on utilise donc la valeur retournee par "Oo.new_method" pour acceder la TMV
 locale. On verifie que celle-ci contient bien une methode, sinon on resoud
 en utilisant "Oo.resolve_local_method obj label". GETMETHOD
   Lsend(Unknown(Lvar new_id),...)

- Appels a une methode d'un objet inconnu dans un module migrable:
 Les acces sont faits par (GETJOINMETHOD_UNRESOLVED(module,string))
 qui est remplace par (CONSTINT(label);GETMETHOD)
   Lsend(Join(string),...)

*)

(**** Object copy ****)

let copy (o : < .. >) : 'a =
  let o = Obj.repr (o : 'a) in
  let s = Obj.size o in
  let r = Obj.new_block Tag.object s in
  for i = 0 to s - 1 do Obj.set_field r i (Obj.field o i) done;
  Obj.set_field r 0 (Obj.repr 0); (* no locator *)
  Obj.set_field r 2 r;
  Obj.obj r

(**** Compression options ****)
(* Parameters *)
type params = {
    mutable compact_table : bool;
    mutable copy_parent : bool;
    mutable clean_when_copying : bool;
    mutable retry_count : int;
    mutable bucket_small_size : int
  } 

let params = {
  compact_table = true;
  copy_parent = true;
  clean_when_copying = true;
  retry_count = 3;
  bucket_small_size = 16
} 

(**** Parameters ****)

let step = Sys.word_size / 16
let first_bucket = 1
let bucket_size = 32                    (* Must be 256 or less *)
let initial_object_size = 7

(**** Index ****)

type label = int

let label_count = ref 0

let next label =
  incr label_count;
  let label = label + step in
  if label mod (step * bucket_size) = 0 then
    label + step * (65536 - bucket_size)
  else
    label

let decode label =
  (label / 65536 / step, (label mod (step * bucket_size)) / step)

(**** Items ****)

type item

let dummy_item = (magic () : item)

(**** Buckets ****)

type bucket = item array

let version = ref 0

let set_bucket_version (bucket : bucket) =
  bucket.(bucket_size) <- (magic !version : item)

let bucket_version bucket =
  (magic bucket.(bucket_size) : int)

let bucket_list = ref []

let empty_bucket = [| |]

let new_bucket () =
  let bucket = Array.create (bucket_size + 1) dummy_item in
  set_bucket_version bucket;
  bucket_list := bucket :: !bucket_list;
  bucket

let copy_bucket bucket =
  let bucket = Array.copy bucket in
  set_bucket_version bucket;
  bucket.(bucket_size) <- (magic !version : item);
  bucket_list := bucket :: !bucket_list;
  bucket

(**** Make a clean bucket ****)

let new_filled_bucket pos methods =
  let bucket = new_bucket () in
  List.iter
    (fun (lab, met) ->
       let (buck, elem) = decode lab in
       if buck = pos then
         bucket.(elem) <- (magic met : item))
    (List.rev methods);
  bucket

(**** Bucket merging ****)

let small_buckets = ref (Array.create 10 [| |])
let small_bucket_count = ref 0

let insert_bucket bucket =
  let length = Array.length !small_buckets in
  if !small_bucket_count >= length then begin
    let new_array = Array.create (2 * length) [| |] in
    Array.blit !small_buckets 0 new_array 0 length;
    small_buckets := new_array
  end;
  !small_buckets.(!small_bucket_count) <- bucket;
  incr small_bucket_count

let remove_bucket n =
  !small_buckets.(n) <- !small_buckets.(!small_bucket_count - 1);
  decr small_bucket_count

let bucket_used b =
  let n = ref 0 in
  for i = 0 to bucket_size - 1 do
    if b.(i) != dummy_item then incr n
  done;
  !n

let small_bucket b = bucket_used b <= params.bucket_small_size

exception Failed

let rec except e =
  function
    [] -> []
  | e'::l -> if e == e' then l else e'::(except e l)

let merge_buckets b1 b2 =
  for i = 0 to bucket_size - 1 do
    if
      (b2.(i) != dummy_item) & (b1.(i) != dummy_item) & (b2.(i) != b1.(i))
    then
      raise Failed
  done;
  for i = 0 to bucket_size - 1 do
    if b2.(i) != dummy_item then
      b1.(i) <- b2.(i)
  done;
  bucket_list := except b2 !bucket_list;
  b1

let rec choose bucket i =
  if (i > 0) & (!small_bucket_count > 0) then begin
    let n = Random.int !small_bucket_count in
    if not (small_bucket !small_buckets.(n)) then begin
      remove_bucket n; choose bucket i
    end else
      try
        merge_buckets !small_buckets.(n) bucket
      with Failed ->
        choose bucket (i - 1)
  end else begin
    insert_bucket bucket;
    bucket
  end

let compact b =
  if
    (b != empty_bucket) & (bucket_version b = !version) & (small_bucket b)
  then
    choose b params.retry_count
  else
    b

let compact_buckets buckets =
  for i = first_bucket to Array.length buckets - 1 do
    buckets.(i) <- compact buckets.(i)
  done

(**** Labels ****)

let first_label = first_bucket * 65536 * step

let last_label = ref first_label
let methods = Hashtbl.create 101

let new_label () =
  let label = !last_label in
  last_label := next !last_label;
  label

let new_method met =
  try
    let label = Hashtbl.find methods met in
    Debug.print2 'o' "New method : %s label : %d\n" met label;
    label
  with Not_found ->
    let label = new_label () in
    Hashtbl.add methods met label;
    label

let new_anonymous_method =
  new_label

(**** Types ****)

type object = t array

(**** Sparse array ****)

module S = struct type t = string let compare = compare end

module Vars = Map.Make(S)
type vars = int Vars.t

module Meths = Map.Make(S)
type meths = label list Meths.t

type obj_init
type table =
 {
    mutable messages: (string*int*int) list;
   mutable buckets: bucket array;
   mutable ext_buckets: bucket array;
   mutable methods: meths;
   mutable saved_methods: meths list;
   mutable meth_defs: (label * item) list;
   mutable vars: vars;
   mutable saved_vars: vars list;
   mutable saved_var_lst: int option list;
   mutable size: int;
   mutable init: obj_init list list }

let table_count = ref 0

let new_table () =
  incr table_count;
  { messages = [];
    buckets = [| new_bucket () |];
    ext_buckets = [| |];
    methods = Meths.empty;
    saved_methods = [];
    meth_defs = [];
    vars = Vars.empty;
    saved_vars = [];
    saved_var_lst = [];
    size = initial_object_size;
    init = [[]; []]}

let copy_table array1 array2 =
  incr version;
  array1.messages <- array2.messages;
  array1.buckets <- Array.copy array2.buckets;
  array1.methods <- array2.methods;
  array1.saved_methods <- [];
  array1.meth_defs <- array2.meth_defs;
  array1.vars <- array2.vars;
  array1.saved_vars <- [];
  array1.saved_var_lst <- array1.saved_var_lst;
  array1.size <- array2.size;
  array1.init <- array2.init

let resize array new_size =
  let old_size = Array.length array.buckets in
  if new_size > old_size then begin
    let new_buck = Array.create new_size empty_bucket in
    Array.blit array.buckets 0 new_buck 0 old_size;
    array.buckets <- new_buck
 end

let put array label element =
  let (buck, elem) = decode label in
  resize array (buck + 1);
  let bucket = ref (array.buckets.(buck)) in
  if !bucket == empty_bucket then begin
    bucket := new_bucket ();
    array.buckets.(buck) <- !bucket
  end;
  if !bucket.(elem) != element then begin
    if
      (bucket_version !bucket < !version)
          &
      (!bucket.(elem) != dummy_item)
    then begin
      if params.clean_when_copying then
        bucket := new_filled_bucket buck array.meth_defs
      else
        bucket := copy_bucket !bucket;
      array.buckets.(buck) <- !bucket
    end;
    !bucket.(elem) <- element
  end

(**** Classes ****)

type t
type class_info = (* taille allouee : 6 dans bytecomp/translclass *)
  {mutable obj_init: t;
   mutable class_init: table -> unit;
   mutable table: table;
   mutable class_sig: (string*int*int) list;
   mutable class_type: class_id;
   mutable methodes: (string,label) Hashtbl.t
  }

let set_initializer table init =
  match table.init with
    l::l'::l'' ->
      let i =
        List.fold_right
          (fun init2 init1 -> (magic init1 : obj_init -> obj_init) init2)
          l init
      in
        table.init <- (i::l')::l''
  | _ ->
      invalid_arg "Fatal error in Oo.set_initializer."

let inheritance table cl vars meths =
  if
    params.copy_parent
    && (table.methods = Meths.empty) && (table.size = initial_object_size)
    && (table.init = [[]; []])
  then begin
    copy_table table cl.table;
    table.init <- table.init@[[]];
    table.vars <- 
      List.fold_left
        (fun s v ->
           try Vars.add v (Vars.find v table.vars) s with Not_found -> s)
        Vars.empty
        vars;
    table.methods <-
      List.fold_left
        (fun s m ->
           try Meths.add m (Meths.find m table.methods) s with Not_found -> s)
        Meths.empty
        meths
  end else begin
    table.init <- []::table.init;
    table.saved_vars <- table.vars::table.saved_vars;
    table.vars <-
      List.fold_left
        (fun s v ->
           try Vars.add v (Vars.find v table.vars) s with Not_found -> s)
        Vars.empty
        vars;
    table.saved_methods <- table.methods::table.saved_methods;
    table.methods <-
      List.fold_left
        (fun s m ->
           try Meths.add m (Meths.find m table.methods) s with Not_found -> s)
        Meths.empty
        meths;
    cl.class_init table;
    table.vars <-
      List.fold_left
        (fun s v ->
           try Vars.add v (Vars.find v table.vars) s with Not_found -> s)
        (List.hd table.saved_vars)
        vars;
    table.saved_vars <- List.tl table.saved_vars;
    table.methods <-
      List.fold_left
        (fun s m ->
           try Meths.add m (Meths.find m table.methods) s with Not_found -> s)
        (List.hd table.saved_methods)
        meths;
    table.saved_methods <- List.tl table.saved_methods
  end

let get_method_labels table name =
  try
    Meths.find name table.methods
  with Not_found ->
    let label = new_anonymous_method () in
    table.methods <- Meths.add name [label] table.methods;
    [label]

let get_method_label table name =
  List.hd (get_method_labels table name)

let lookup obj lab =
  Debug.print "send";
  let (buck, elem) = decode lab in
  (magic obj : item array array array).(6).(buck).(elem)

let set_method table name nparams element =
  let labels = get_method_labels table name
  in
  table.messages <- (name,List.hd labels,nparams)::table.messages; 
  List.iter
    (function label ->
      Debug.print2 'o' "set methode :%s label : %d\n" name label;
      table.meth_defs <- (label, element) :: table.meth_defs;
      put table label element) labels

let get_method table label =
  let (buck, elem) = decode label in
  table.buckets.(buck).(elem)

let new_slot table =
  let index = table.size in
  table.size <- index + 1;
  index

let get_variable table name =
  try
    Vars.find name table.vars
  with Not_found ->
    let index = new_slot table in
    table.vars <- Vars.add name index table.vars;
    index

let hide_variable table name =
  try
    let i = Vars.find name table.vars in
    table.vars <- Vars.remove name table.vars;
    table.saved_var_lst <- Some i::table.saved_var_lst
  with Not_found ->
    table.saved_var_lst <- None::table.saved_var_lst

let rec list_remove name =
  function
    [] ->
       invalid_arg "Fatal error in Oo.get_private_variable"
  | (n, _) as a::l ->
       if name = n then l
       else a::list_remove name l

let get_private_variable table name =
  let index =
    try Vars.find name table.vars with Not_found -> new_slot table
  in
  table.vars <-
    begin match List.hd table.saved_var_lst with
      None ->
        Vars.remove name table.vars
    | Some i ->
        Vars.add name i table.vars
    end;
  table.saved_var_lst <- List.tl table.saved_var_lst;
  index

let method_count = ref 0
let inst_var_count = ref 0



let create_ext_buckets messages wrappe =
  let ext_buckets = ref [||] in
  List.iter (fun (meth_name,label,nparams) ->
    let ext_label = new_method meth_name in
    let (buck,elem) = decode ext_label in
    let bucket_size = Array.length !ext_buckets in
    if bucket_size <= buck then
      begin
        let new_buckets = Array.create (buck+1) empty_bucket in
        Array.blit !ext_buckets 0 new_buckets 0 bucket_size;
        ext_buckets := new_buckets
      end;
    Debug.print0 'o' "bucket ...\n";
    let bucket = (!ext_buckets).(buck) in
    Debug.print0 'o' "...bucket \n";
    let bucket =
      if bucket = empty_bucket then
        let bucket = new_bucket () in
        (!ext_buckets).(buck) <- bucket;
        bucket
      else 
        bucket
    in
    bucket.(elem) <- wrappe nparams meth_name label
         ) messages;
  !ext_buckets
  

let oo_create_class class_info public_methods creator class_init class_type =
  let table = new_table () in
  List.iter
    (function met ->
       table.methods <- Meths.add met [new_method met] table.methods)
    public_methods;
  class_init table;
  method_count := !method_count + List.length table.meth_defs;
  if params.compact_table then
    compact_buckets table.buckets;
  inst_var_count := !inst_var_count + table.size - 1;
  Debug.print "Class info ";
  class_info.class_init <- class_init;
  class_info.table <- table;
  class_info.obj_init <- creator table;
  class_info.class_sig <- table.messages; 
  class_info.methodes <- methods;
   class_info.class_type <- class_type;
  Debug.print "Class info 1";
  table.buckets.(0) <- Obj.magic class_info;
  Debug.print "Class info 2";
  class_info

let create_class class_info public_methods creator class_init module_data =
  let class_id = Class_id (module_data.Reloc.module_name, 
                           module_data.Reloc.classes)
  in
    module_data.Reloc.classes <- module_data.Reloc.classes + 1;
  oo_create_class class_info public_methods creator class_init class_id
    
(**** Objects ****)

let create_object table =
  Debug.print "create object";
  let obj = Obj.new_block Tag.object table.size in
  Obj.set_field obj 2 obj;                          (* local *)
                                                    (* stub *)
  Obj.set_field obj 5 (Obj.repr table.ext_buckets);     (* TMV *)
  Obj.set_field obj 6 (Obj.repr table.buckets);     (* TMV *)
  let initialization = Obj.magic List.hd (List.hd table.init) in
  initialization (Obj.obj obj)

(**** Statistics ****)

type stats =
  { classes: int; labels: int; methods: int; inst_vars: int; buckets: int;
    distrib : int array; small_bucket_count: int; small_bucket_max: int }

let distrib () =
  let d = Array.create 32 0 in
  List.iter
    (function b ->
       let n = bucket_used b in
       d.(n - 1) <- d.(n - 1) + 1)
    !bucket_list;
  d

let stats () =
  { classes = !table_count; labels = !label_count;
    methods = !method_count; inst_vars = !inst_var_count;
    buckets = List.length !bucket_list; distrib = distrib ();
    small_bucket_count = !small_bucket_count;
    small_bucket_max = Array.length !small_buckets  }

let sort_buck lst =
  List.map snd
    (Sort.list (fun (n, _) (n', _) -> n <= n')
       (List.map (function b -> (bucket_used b, b)) lst))

let show_buckets () =
  List.iter
    (function b ->
       for i = 0 to bucket_size - 1 do
         print_char (if b.(i) == dummy_item then '.' else '*')
       done;
       print_newline ())
    (sort_buck !bucket_list)
;;


let get_class_info obj = (magic obj : class_info array array).(6).(0)

external set_find_method : ( string -> label ) -> unit = "set_find_method"

let _ =   set_find_method new_method

let invoke_methode methode target args =
  let m = ref ((Obj.magic methode) target)
  in
  List.iter
    (function arg ->
      m:= (Obj.magic (* sur *) (!m)) arg) args;
  (Obj.repr (!m))
    
let wrappe_methode true_invoke nparams =
  function this_stub -> 
    let rec iter nparams list =
      if nparams=0 then
        true_invoke this_stub (List.rev list)
      else
        Obj.magic
          (function param ->
	    iter (nparams-1) (param::list)
              )
    in
    iter nparams []
