(***********************************************************************)
(*                                                                     *)
(*                    ______________                                   *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(***********************************************************************)

(*         File written in the Objective-CAML language                 *)

type exn = {
    mutable exn_name : string;
    mutable mod_name : string;
    mutable mod_num : int
  }

type module_data = {
  module_name : string;
  mutable module_slot : int; (* slot in the module table for this module *)
  offset : int;
  longueur : int;
  mutable begin_code : Obj.t;
  mutable end_code : Obj.t;
  reloc : int array;
  bytecode : string;
  mutable interface : int; (* slot in the global table for this module *)
  mutable classes : int;
  mutable migrate: bool;
  mutable exceptions: exn list;
  }
;;
type module_table = module_data array;;

external set_add_module : (module_data -> module_table -> unit) -> 
  unit =  "set_add_module"
external set_find_module : (string -> int) -> unit = "set_find_module"
external set_find_prim : (string -> int) -> unit = "set_find_prim"
external available_primitives : unit -> string array = "available_primitives"

let exn_htable = Hashtbl.create 53
let module_htable = Hashtbl.create 53
let modules_list = ref []
let init_module_htable = ref true

exception ModuleNotAvailable of string
exception PrimitiveNotAvailable of string

let add_exn exn =
  try
    Hashtbl.find exn_htable (exn.mod_name,exn.mod_num)
  with
    Not_found ->
      Hashtbl.add exn_htable (exn.mod_name,exn.mod_num) exn;
      exn

let add_module module_data =
  modules_list := module_data.module_name :: !modules_list;
  Hashtbl.add module_htable module_data.module_name module_data;
  List.iter (fun exn -> 
    Hashtbl.add exn_htable (exn.mod_name,exn.mod_num) exn
      ) module_data.exceptions;
  module_data.exceptions <- []

let add_module_table module_data module_table =
  if !init_module_htable then
    (* first call to add_module: we must put several modules at once in
      the module_table *)
    begin
      for i=0 to module_data.module_slot do
        add_module module_table.(i)
      done;
      init_module_htable := false
    end
  else add_module module_data

let module_list () = !modules_list
    
let find_module module_name =
  try
    let m = Hashtbl.find module_htable module_name in
    if m.interface = -1 then raise Not_found; (* dummy module *)
    m.module_slot
  with
   Not_found -> raise (ModuleNotAvailable module_name)

let find_prim = 
  let prim_htable = Hashtbl.create 50 in
  let prims = available_primitives () in
    for i=0 to (Array.length prims)-1 do
       Hashtbl.add prim_htable prims.(i) i
    done;
    function prim_name ->
      try
	let p = Hashtbl.find prim_htable prim_name in
          p
      with
          Not_found -> 
            raise (PrimitiveNotAvailable prim_name)
              
external install_module : module_data -> module_data -> unit = "install_module"

external install_dummy_module: string -> int -> module_data
    = "install_dummy_module"

let is_dummy (m: module_data) = (m.longueur = 0)

let find_module_data name size =
  try
    Hashtbl.find module_htable name
  with
      Not_found -> 
        let m = install_dummy_module name size in
        add_module m;
        m

let get_module_data name = Hashtbl.find module_htable name

(*
external prepare_install: int -> unit = "module_prepare_install"
*)
external global_data : unit -> Obj.t array = "get_global_data"
external realloc_global_data : int -> unit = "realloc_global"
external static_alloc : int -> string = "static_alloc"
external static_free : string -> unit = "static_free"
external static_resize : string -> int -> string = "static_resize"
type closure = unit -> Obj.t
external reify_bytecode : string -> int -> Obj.t array -> 
  closure = "reify_bytecode"

external set_add_exn: (exn -> exn) -> unit = "set_add_exn" 

let find_exn exn_name mod_name mod_num = 
  try
    Hashtbl.find exn_htable (mod_name,mod_num)
  with
    Not_found ->
      let exn = (Obj.magic (Obj.new_block Tag.exn 3) : exn)
      in
      exn.exn_name <- exn_name;
      exn.mod_name <- exn.mod_name;
      exn.mod_num <- exn.mod_num;
      exn

let _ =
  set_add_module add_module_table;
  set_find_module find_module;
  set_find_prim find_prim;
  set_add_exn add_exn

