(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Damien Doligez, projet Para, INRIA Rocquencourt          *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: weak.ml,v 1.1.1.1 1998/04/20 11:32:18 lefessan Exp $ *)

(* Weak array operations *)

type 'a t;;

external create: int -> 'a t = "weak_create";;

let length x = Obj.size(Obj.repr x) - 2;;

external set : 'a t -> int -> 'a option -> unit = "weak_set";;

external get: 'a t -> int -> 'a option = "weak_get";;

let fill ar ofs len x =
  if ofs < 0 || ofs + len > length ar
  then raise (Invalid_argument "Weak.fill")
  else begin
    for i = ofs to (ofs + len - 1) do
      set ar i x
    done
  end
;;

let blit ar1 of1 ar2 of2 len =
  if of1 < 0 || of1 + len > length ar1 || of2 < 0 || of2 + len > length ar2
  then raise (Invalid_argument "Weak.blit")
  else begin
    if of2 > of1 then begin
      for i = 0 to len - 1 do
        set ar2 (of2 + i) (get ar1 (of1 + i))
      done
    end else begin
      for i = len - 1 downto 0 do
        set ar2 (of2 + i) (get ar1 (of1 + i))
      done
    end
  end
;;

type 'a set;;

external set_create: int -> 'a set = "weakset_create";;
external set_add: 'a set -> 'a -> 'a set = "weakset_add";;
external set_array: 'a set -> 'a array = "weakset_array";;
external set_remove: 'a set -> 'a -> bool = "weakset_remove";;
external set_info: 'a set -> int * int = "weakset_info";;

(*

 *** array_add weak_array ele *** retourne l'index de la position a
  laquelle a ete stocke l'element *ele* dans *weak_t*, ou leve
  l'exception * Failure _ * si aucune place n'etait disponible. (on 
  utilisera alors *array_copy* pour aggrandir le tableau).
 *** array_copy weak_array size *** copie le contenu de *weak_array* dans
  un autre *'a array* de taille *size* nouvellement alloue.
 *** array_free weak_array n *** libere l'emplacement *n* de *weak_array* 
  de sorte  qu'il puisse etre utilise par un *array_add*. Tant que
  cette operation n'est pas effectuee, un slot occupe ne peut pas
  etre reutilise (meme par la liberation du GC).
 *** array_get weak_array n *** retourne *None* si le slot *n* est libre
  ou si le pointeur faible ne pointe plus sur l'objet (L'operation
  suivante consiste alors a liberer ce slot par *array_free weak_array n*
  )

  L'avantage de cette semantique est la rapidite des operations *array_add*
  (couplee a *array_copy* en cas de failure) , *array_get* et *array_free*.
  L'operation la plus couteuse est *set weak_t n (Some p)* quand n est 
  un element libre non ordonne (ce qui est rare, mais peut arriver. 
  Dans un tel cas, on utilisera un *'a t* de toute facon
  plutot qu'un *'a array* ).


*)

type 'a array;;

external array_create: int -> 'a array = "weakarray_create"
external array_copy: 'a array -> int -> 'a array = "weakarray_realloc";;
external array_add : 'a array -> 'a option -> int = "weakarray_add";;
external array_set : 'a array -> int -> 'a option -> unit = "weak_set";;
external array_get: 'a array -> int -> 'a option = "weak_get";;
external array_free: 'a array -> int -> unit = "weakarray_free";;
let array_length = length;;

let ref_add array option =
  try
    array_add (!array) option
  with
    Failure _ ->
      array := array_copy (!array) 
           ((array_length !array)*2);
      array_add (!array) option
