(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Damien Doligez, projet Para, INRIA Rocquencourt          *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: weak.mli,v 1.1.1.1 1998/04/20 11:32:18 lefessan Exp $ *)

(* Module [Weak]: arrays of weak pointers *)

type 'a t;;
        (* The type of arrays of weak pointers (weak arrays).  A weak
           pointer is an object that the garbage collector may erase at
           any time.
           A weak pointer is said to be full if it points to an object,
           empty if the object was erased by the GC.
         *)

val create : int -> 'a t;;
        (* [Weak.create n] returns a new weak array of length [n].
           All the pointers are initially empty.
         *)
val length : 'a t -> int;;
        (* [Weak.length ar] returns the length (number of elements) of
           [ar].
         *)
val set : 'a t -> int -> 'a option -> unit;;
        (* [Weak.set ar n (Some el)] sets the [n]th cell of [ar] to be a
           (full) pointer to [el]; [Weak.set ar n None] sets the [n]th
           cell of [ar] to empty.
           Raise [Invalid_argument "Weak.set"] if [n] is not in the range
           0 to [Weak.length a - 1].
         *)
val get : 'a t -> int -> 'a option;;
        (* [Weak.get ar n] returns None if the [n]th cell of [ar] is
           empty, [Some x] (where [x] is the object) if it is full.
           Raise [Invalid_argument "Weak.get"] if [n] is not in the range
           0 to [Weak.length a - 1].
         *)
val fill: 'a t -> int -> int -> 'a option -> unit;;
        (* [Weak.fill ar ofs len el] sets to [el] all pointers of [ar] from
           [ofs] to [ofs + len - 1].  Raise [Invalid_argument "Weak.fill"]
           if [ofs] and [len] do not designate a valid subarray of [a].
        *)
val blit : 'a t -> int -> 'a t -> int -> int -> unit;;
        (* [Weak.blit ar1 off1 ar2 off2 len] copies [len] weak pointers
           from [ar1] (starting at [off1]) to [ar2] (starting at [off2]).
           It works correctly even if [ar1] and [ar2] are the same.
           Raise [Invalid_argument "Weak.blit"] if [off1] and [len] do
           not designate a valid subarray of [ar1], or if [off2] and [len]
           do not designate a valid subarray of [ar2].
         *)


type 'a set;;
        (*
          The type of a set of weak pointers. 
        *)
val set_create: int -> 'a set;;
        (* [Weak.set_create len] creates a set of weak pointers which an
           initial size of [len] weak pointers. The set can contains more
           than [len] weak pointers at a time, but then the set will be 
           reallocated.
        *)
val set_add: 'a set -> 'a -> 'a set;;
        (* [Weak.set_add set obj] adds a weak pointer to [obj] to a set
           [set] of weak pointers and returns the new set. Indeed, if the
           set [set] is full, a new set is allocated.
        *)
val set_array: 'a set -> 'a array;;
        (* [Weak.set_array set] returns an array containing all life objects
           in the set [set].
        *)
val set_remove: 'a set -> 'a -> bool;;
        (* [Weak.set_remove set obj] removes [obj] from the set [set]. This
           operation can be expensive, since all the set must be scanned.
        *)
val set_info: 'a set -> int * int;;
        (* [Weak.set_info set] returns a couple ([size],[nobjs]) where
           [size] is the total size of [set] and [nobjs] the number of
           life objects in [set].
        *)

type 'a array;;
val array_create: int -> 'a array;;
val array_copy: 'a array -> int -> 'a array;;
val array_add : 'a array -> 'a option -> int;;
val array_set : 'a array -> int -> 'a option -> unit;;
val array_get: 'a array -> int -> 'a option;;
val array_free: 'a array -> int -> unit;;
val array_length : 'a array -> int;;
        (* [Weak.array_length ar] returns the length (number of elements) of
           [ar].
         *)

val ref_add : 'a array ref -> 'a option -> int
