(***********************************************************************)
(*                                                                     *)
(*                      Join Objective Caml                            *)
(*                                                                     *)
(*        Fabrice LE FESSANT, Project PARA, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Typemisc

(*
  Analyse des join patterns pour la fusion de canaux.

Entree: 
  (nchans : int) nombre de channels differents dans la definition
  (patterns : int list list) les join-patterns

Sortie:
  (channels: int -> int) donne a chaque channel sont index
  (s: structure) definition structure
*)

(*
1. a chaque channel, on associe la liste des semi-patterns avec lesquels il
    s'associe.
2. on trie les listes de semi-patterns
3. on place les channels dans une table de hashage avec comme cles leurs 
    listes de semi-patterns.
4. pour chaque cle, on extrait l'ensemble des channels equivalents

a | b | c = p1
a | b | d = p2
a | e | c = p3
a | e | d = p4

donne:

let patterns =
  [ [0;1;2]; [0;1;3]; [0;4;2]; [0;4;3] ]
let nchans = 5

;;


 *)


let fusion nchans patterns =
  let table = Array.create nchans [] in
  List.iter (fun pattern ->
    let rec remove i =
      let t = Array.of_list pattern in
      let channel = t.(i) in
      t.(i) <- -1;
      table.(channel) <- (Sort.list (<) (Array.to_list t)) :: table.(channel);
      if i <> 0 then remove (i-1)
    in
    remove ((List.length pattern) - 1)
    ) patterns;
  let htable = Hashtbl.create (nchans*2) in
  for i = 0 to nchans - 1 do
    table.(i) <- Sort.list (<) table.(i);
    Hashtbl.add htable table.(i) i
  done;
  let singles = ref [] in
  let sets = ref [] in
  for i = 0 to nchans - 1 do
      match Hashtbl.find_all htable table.(i) with
        [channel] -> singles := channel :: !singles;
      | [] -> ()
      | set -> 
          sets := set :: !sets;
          List.iter (fun channel -> 
            table.(channel) <- [];
            ) set;
  done;
  let f = Array.create nchans (-1) in
  let count = ref 0 in
  List.iter (fun set ->
    let p = ref 0 in
    List.iter (fun channel ->
      f.(channel) <- !count + (!p lsl 10);
      incr p
      ) set;
    incr count
    ) !sets;
  let nmult = !count in
  List.iter (fun channel -> f.(channel) <- !count; incr count) !singles;
  if nmult > 0 then
    Debug.print3 'f' "Nmult: %d Nchans: %d/%d\n" nmult !count nchans;
  nmult,!count,f
;;
(*
to_heap [1,1;2,2;3,3;4,4;5,5];;

*)
let rec log2 n =
  if n <> 0 then 1 + log2 (n/2) else 0
;;

let rec flatten_tree pos tab posh heap size =
  match size with
    0 -> 
      heap.(posh) <- fst tab.(pos);
      heap.(2*posh) <- snd tab.(pos)
  | 1 ->
      heap.(posh) <- fst tab.(pos);
      heap.(2*posh) <- snd tab.(pos);
      heap.(2*posh+1) <- fst tab.(pos+1);
      heap.(4*posh+2) <- snd tab.(pos+1);
  | _ ->
      let pivot = pos+(1 lsl (size-1)) in
      heap.(posh) <- fst tab.(pivot-1);
      flatten_tree pos tab (2*posh) heap (size-1);
      flatten_tree pivot tab (2*posh+1) heap (size-1)
;;

let to_heap list =
  let list = Sort.list (fun x y -> fst x < fst y) list in
  let size = log2 (List.length list - 1) in
  let heap = Array.create (4*(1 lsl size)) 0 in
  let statuses = Array.create (1 lsl  size) (0,0) in
  let rec set_status i = 
    function s::tl -> statuses.(i) <- s; set_status (i+1) tl
      | [] -> () in
  set_status 0 list;
  heap.(0) <- size;
  flatten_tree 0 statuses 1 heap size;
  heap
;;  

(* on construit la structure de recherche des matchs

let (nmult,nchans,f) = fusion nchans patterns;;
 *)

type process =
    Process of int (* process number *) * int (* thread number + 1 *)
  | Dispatch of (int * process ref) list
  | Undefined

type matching =
  {
    num: int;
    args: int list;
    mutable process: process
  } 

let list_pos thread list =
  let rec pos i = function
      [] -> raise Not_found
    | h :: tl when h == thread -> i
    | _ :: tl -> pos (i+1) tl
  in
  pos 0 list

let structure env nmult nchans f processes =
  try
  let index i = f.(i) land 63 in
  let dispatch i = f.(i) lsr 10 in
  let bit i = 1 lsl (index i)  in
  let nprocesses = ref 0 in
  let nmatchs = ref 0 in
  let matchs = ref [] in
  let trees = Array.create nchans [] in
  List.iter (fun (join,process,thread) ->
    let status = ref 0 in
    List.iter (fun i -> status := !status lor (bit i)) join.join_pattern;
    let arguments = Sort.list (fun i j -> index i < index j) 
        join.join_pattern in
    let matching =
      try
        List.assoc !status !matchs
      with
        Not_found ->
          let matching = { 
            num = !nmatchs;
            args = List.map index arguments;
            process = Undefined;
          } in
          incr nmatchs;
          matchs := (!status,matching) :: !matchs;
          matching
    in
    let rec insert_process process args = 
      match args with
        h::args when index h < nmult -> 
          begin
            match process with
              Undefined -> 
                Dispatch ([dispatch h,ref (insert_process Undefined args)])
            | Process _ -> failwith "insert_process: Process _ -> 1"
            | Dispatch l ->
                try
                  let p = List.assoc (dispatch h) l in
                  p := insert_process !p args;
                  Dispatch l
                with
                  Not_found ->
                    Dispatch
                      ((dispatch h,ref (insert_process Undefined args))::l)
          end
      | _ -> 
          begin
            match process with
              Undefined -> Process (!nprocesses, 
                                    match thread with
                                      None -> 0
                                    | Some -1 -> -1
                                    | Some chan -> 1+(list_pos chan arguments))
            | _ -> failwith "insert_process: _ -> 2"
          end
    in
    matching.process <- insert_process matching.process arguments;
    incr nprocesses;
    List.iter (fun channel ->
      let chan = index channel in
      try
        List.assoc !status trees.(chan); ()
      with
        Not_found -> 
          trees.(chan) <- (!status,matching.num) :: trees.(chan)
                                                     ) join.join_pattern
      ) processes;
  let trees = Array.to_list trees in
  let matches = List.rev (List.map snd !matchs) in
  let mkint i = mkint i env in
  mktuple (
   (mkint nmult)::(mkint nchans)::
    (List.map (fun tree -> 
      mktuple (List.flatten 
                 (List.map (fun (s,m) -> [mkint s;mkint m]) tree)) env
     ) trees)@
    (List.map (fun matching ->
      mktuple ((
               let rec to_tuple = function
                   Process (i,j) -> mktuple [mkint i;mkint j] env
                 | Undefined -> failwith "to_tuple: Undefined"
                 | Dispatch l ->
                     mktuple (List.map (fun (i,p) -> to_tuple !p) 
                                (Sort.list (fun p1 p2 -> fst p1 < fst p2) l)
                                ) env
               in
               to_tuple matching.process
               )::(List.map mkint matching.args)) env
      ) matches)
    ) env
    with
    e -> printnl_string "Exception in janalyse";
      raise e
