(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: typecore.ml,v 1.5 1998/05/11 07:28:30 lefessan Exp $ *)

(* Typechecking for the core language *)

open Misc;;
open Asttypes;;
open Longident;;
open Parsetree;;
open Types;;
open Typedtree;;
open Btype;;
open Ctype;;
open Typemisc;;

let (type_defs: (Env.t -> Location.t ->
  (Ident.t option * (Parsetree.pattern * Parsetree.expression) list *
     Location.t) list ->
       (Typedtree.pattern * Typedtree.expression) list *
         (Ident.t option * (Ident.t * Types.type_expr) list) list * Env.t) 
       ref) = ref (fun _ _ _ -> failwith "type_defs not implemented")
let (type_process: (Env.t ->
  Parsetree.expression ->
  (string * Ident.t) list * string list * bool ->
  Typedtree.expression * string list *  Ident.t list *
      (string * Ident.t) option) ref)
    = ref (fun _ _ _ -> failwith "type_process not implemented")
let (type_locations: (Env.t ->
  Location.t ->
  (string * Parsetree.automatas * Parsetree.expression) list ->
  (Typedtree.pattern * Typedtree.expression) list *
  (Typedtree.pattern * Typedtree.expression) list * Typedtree.expression *
  Env.t * Env.t) ref)
    = ref (fun _ _ _ -> failwith "type_locations not implemented")







let rec type_exp env sexp =
  match sexp.pexp_desc with
    Pexp_ident lid ->
      begin try
        let (path, desc) = Env.lookup_value lid env in
        match desc.val_kind, lid with
          Val_ivar _, Longident.Lident lab ->
            let (path_self, _) =
              Env.lookup_value (Longident.Lident "*self*") env
            in
            { exp_desc = Texp_instvar (path_self, path);
              exp_loc = sexp.pexp_loc; exp_type = instance desc.val_type;
              exp_env = env}
(* we could also test its type ... *)
        | _ ->
            { exp_desc = Texp_ident (path, desc);
              exp_loc = sexp.pexp_loc; 
              exp_type = instance (
              match desc.val_type.desc with
                Trecord (ty,lbls) -> ty
              | _ -> desc.val_type);
              exp_env = env}
      with
        Not_found -> raise (Error (sexp.pexp_loc, Unbound_value lid))
      end
  | Pexp_constant cst ->
      {exp_desc = Texp_constant cst; exp_loc = sexp.pexp_loc;
       exp_type = type_constant cst; exp_env = env}
  | Pexp_let (rec_flag, spat_sexp_list, sbody) ->
      let (pat_exp_list, new_env) = type_let env rec_flag spat_sexp_list in
      let body = type_exp new_env sbody in
      {exp_desc = Texp_let (rec_flag, pat_exp_list, body);
       exp_loc = sexp.pexp_loc; exp_type = body.exp_type; exp_env = env}
  | Pexp_function caselist ->
      let ty_arg = newvar ()
      and ty_res = newvar () in
      let cases = type_cases env ty_arg ty_res caselist in
      Parmatch.check_unused cases;
      Parmatch.check_partial sexp.pexp_loc cases;
      {exp_desc = Texp_function cases; exp_loc = sexp.pexp_loc;
       exp_type = newty (Tarrow (ty_arg, ty_res)); exp_env = env}
  | Pexp_apply (sfunct, sargs) ->
      let funct = type_exp env sfunct in
      let rec type_args ty_fun =
        function
          [] -> [], ty_fun
        | sarg1 :: sargl ->
            let (ty1, ty2) =
              try filter_arrow env ty_fun with
                Unify _ ->
                  raise
                    (Error
                       (sfunct.pexp_loc, Apply_non_function funct.exp_type))
            in
            let arg1 = type_expect env sarg1 ty1 in
            let (argl, ty_res) = type_args ty2 sargl in arg1 :: argl, ty_res
      in
      let (args, ty_res) = type_args funct.exp_type sargs in
      {exp_desc = Texp_apply (funct, args); exp_loc = sexp.pexp_loc;
       exp_type = ty_res; exp_env = env}
  | Pexp_match (sarg, caselist) ->
      let arg = type_exp env sarg in
      let ty_res = newvar () in
      let cases = type_cases env arg.exp_type ty_res caselist in
      Parmatch.check_unused cases;
      Parmatch.check_partial sexp.pexp_loc cases;
      {exp_desc = Texp_match (arg, cases); exp_loc = sexp.pexp_loc;
       exp_type = ty_res; exp_env = env}
  | Pexp_try (sbody, caselist) ->
      let body = type_exp env sbody in
      let cases =
        type_cases env (instance Predef.type_exn) body.exp_type caselist
      in
      Parmatch.check_unused cases;
      {exp_desc = Texp_try (body, cases); exp_loc = sexp.pexp_loc;
       exp_type = body.exp_type; exp_env = env}
  | Pexp_tuple sexpl ->
      let expl = List.map (type_exp env) sexpl in
      {exp_desc = Texp_tuple expl; exp_loc = sexp.pexp_loc;
       exp_type = newty (Ttuple (List.map (fun exp -> exp.exp_type) expl));
       exp_env = env}
  | Pexp_construct (lid, sarg, explicit_arity) ->
      let constr =
        try Env.lookup_constructor lid env with
          Not_found -> raise (Error (sexp.pexp_loc, Unbound_constructor lid))
      in
      let sargs =
        match sarg with
          None -> []
        | Some {pexp_desc = Pexp_tuple sel} when explicit_arity -> sel
        | Some {pexp_desc = Pexp_tuple sel} when constr.cstr_arity > 1 -> sel
        | Some se -> [se]
      in
      let (ty_args, ty_res) = instance_constructor constr in
      begin
        match ty_args with
          Tconstr_tuple ty_args -> 
            if List.length sargs <> constr.cstr_arity then
              raise
                (Error
                   (sexp.pexp_loc,
                    Constructor_arity_mismatch
                      (lid, constr.cstr_arity, List.length sargs)));
            let args = List.map2 (type_expect env) sargs ty_args in
            {exp_desc = Texp_construct (constr, args); 
              exp_loc = sexp.pexp_loc;
              exp_type = ty_res; 
              exp_env = env}
        | Tconstr_record lbls ->
            match sargs with
              [ { pexp_desc = Pexp_record lid_sexp_list }] ->
                let labels = ref [] in
                let args = List.map
                  (fun (name,mut,ty_arg) ->
                    try
                      let sarg = List.assoc (Lident name) lid_sexp_list in
                      type_expect env sarg ty_arg
                    with
                      Not_found ->
                        jerror sexp.pexp_loc ("Missing label "^name)
                      ) lbls in
                if List.length lid_sexp_list <> List.length lbls then
                  begin
                    let labels = ref [] in
                    List.iter (fun (lid,exp) ->
                      let (n,mut,_) = 
                        find_prefixed_label sexp.pexp_loc lbls lid
                      in
                      if List.mem n !labels then
                        raise(Error(sexp.pexp_loc,
                                    Label_multiply_defined lid));
                      labels := n :: !labels
                                       ) lid_sexp_list;
                  end;
                { exp_desc = Texp_construct (constr, args); 
                  exp_loc = sexp.pexp_loc;
                  exp_type = ty_res; 
                  exp_env = env}
            | _ -> raise 
                  (Error(sexp.pexp_loc,
                         NewFailure "Bad arguments for Record constructor"))
      end
  | Pexp_record lid_sexp_list ->
      let ty = newvar () in
      let num_fields = ref 0 in
      let type_label_exp (lid, sarg) =
        let label =
          try Env.lookup_label lid env with
            Not_found -> raise (Error (sexp.pexp_loc, Unbound_label lid))
        in
        let (ty_arg, ty_res) = instance_label label in
        begin try unify env ty_res ty with
          Unify trace ->
            raise (Error (sexp.pexp_loc, Label_mismatch (lid, trace)))
        end;
        let arg = type_expect env sarg ty_arg in
        num_fields := Array.length label.lbl_all; label, arg
      in
      let lbl_exp_list = List.map type_label_exp lid_sexp_list in
      let rec check_duplicates =
        function
          [] -> ()
        | (lid, sarg) :: remainder ->
            if List.mem_assoc lid remainder then
              raise (Error (sexp.pexp_loc, Label_multiply_defined lid))
            else check_duplicates remainder
      in
      check_duplicates lid_sexp_list;
      if List.length lid_sexp_list <> !num_fields then
        raise (Error (sexp.pexp_loc, Label_missing));
      {exp_desc = Texp_record lbl_exp_list; exp_loc = sexp.pexp_loc;
       exp_type = ty; exp_env = env}
  | Pexp_field (srecord, lid) ->
      let (record,label,mut,ty_arg) = 
        type_record env sexp.pexp_loc srecord lid in
      {exp_desc = Texp_field (record, label);
        exp_loc = sexp.pexp_loc;
       exp_type = ty_arg; exp_env = env}
  | Pexp_setfield (srecord, lid, snewval) ->
      let (record,label,mut,ty_arg) = 
        type_record env sexp.pexp_loc srecord lid in
      if mut = Immutable then
        raise (Error (sexp.pexp_loc, Label_not_mutable lid));
      let newval = type_expect env snewval ty_arg in
      {exp_desc = Texp_setfield (record, label, newval);
       exp_loc = sexp.pexp_loc; exp_type = instance Predef.type_unit;
       exp_env = env}
  | Pexp_array sargl ->
      let ty = newvar () in
      let argl = List.map (fun sarg -> type_expect env sarg ty) sargl in
      {exp_desc = Texp_array argl; exp_loc = sexp.pexp_loc;
       exp_type = instance (Predef.type_array ty); exp_env = env}
  | Pexp_ifthenelse (scond, sifso, sifnot) ->
      let cond = type_expect env scond (instance Predef.type_bool) in
      begin match sifnot with
        None ->
          let ifso = type_expect env sifso (instance Predef.type_unit) in
          {exp_desc = Texp_ifthenelse (cond, ifso, None);
           exp_loc = sexp.pexp_loc; exp_type = instance Predef.type_unit;
           exp_env = env}
      | Some sifnot ->
          let ifso = type_exp env sifso in
          let ifnot = type_expect env sifnot ifso.exp_type in
          {exp_desc = Texp_ifthenelse (cond, ifso, Some ifnot);
           exp_loc = sexp.pexp_loc; exp_type = ifso.exp_type; exp_env = env}
      end
  | Pexp_sequence (sexp1, sexp2) ->
      let exp1 = type_statement env sexp1 in
      let exp2 = type_exp env sexp2 in
      {exp_desc = Texp_sequence (exp1, exp2); exp_loc = sexp.pexp_loc;
       exp_type = exp2.exp_type; exp_env = env}
  | Pexp_while (scond, sbody) ->
      let cond = type_expect env scond (instance Predef.type_bool) in
      let body = type_statement env sbody in
      {exp_desc = Texp_while (cond, body); exp_loc = sexp.pexp_loc;
       exp_type = instance Predef.type_unit; exp_env = env}
  | Pexp_for (param, slow, shigh, dir, sbody) ->
      let low = type_expect env slow (instance Predef.type_int) in
      let high = type_expect env shigh (instance Predef.type_int) in
      let (id, new_env) =
        Env.enter_value param
          {val_type = instance Predef.type_int; val_kind = Val_reg} env
      in
      let body = type_statement new_env sbody in
      {exp_desc = Texp_for (id, low, high, dir, body);
       exp_loc = sexp.pexp_loc; exp_type = instance Predef.type_unit;
       exp_env = env}
  | Pexp_constraint (sarg, sty, sty') ->
      let (arg, ty') =
        match sty, sty' with
          None, None -> let arg = type_exp env sarg in arg, arg.exp_type
        | Some sty, None ->
            let ty = Typetexp.transl_simple_type env false sty in
            type_expect env sarg ty, ty
        | None, Some sty' ->
            let (ty', force) = Typetexp.transl_simple_type_delayed env sty' in
            let ty = enlarge_type env ty' in
            force ();
            let arg = type_exp env sarg in
            begin try Ctype.unify env arg.exp_type ty with
              Unify trace ->
                raise
                  (Error
                     (sarg.pexp_loc,
                      Coercion_failure (ty', full_expand env ty', trace)))
            end;
            arg, ty'
        | Some sty, Some sty' ->
            let (ty, force) = Typetexp.transl_simple_type_delayed env sty
            and (ty', force') =
              Typetexp.transl_simple_type_delayed env sty'
            in
            begin try
              let force'' = subtype env ty ty' in
              force (); force' (); force'' ()
            with
              Subtype (tr1, tr2) ->
                raise (Error (sexp.pexp_loc, Not_subtype (tr1, tr2)))
            end;
            type_expect env sarg ty, ty'
      in
      {exp_desc = arg.exp_desc; exp_loc = arg.exp_loc; exp_type = ty';
       exp_env = env}
  | Pexp_when (scond, sbody) ->
      let cond = type_expect env scond (instance Predef.type_bool) in
      let body = type_exp env sbody in
      {exp_desc = Texp_when (cond, body); exp_loc = sexp.pexp_loc;
       exp_type = body.exp_type; exp_env = env}
  | Pexp_send (e, met) ->
      let object = type_exp env e in
      begin try
        let (exp, typ) =
          match object.exp_desc with
            Texp_ident (path, {val_kind = Val_self meths}) ->
              begin try
                let (id, typ) = Meths.find met !meths in
                Texp_send (object, Tmeth_val id), typ
              with
                Not_found ->
                  let id = Ident.create met in
                  let typ = filter_method env met Private object.exp_type in
                  meths := Meths.add met (id, typ) !meths;
                  Texp_send (object, Tmeth_val id), typ
              end
          | Texp_ident (path, {val_kind = Val_anc methods}) ->
              let typ = filter_method env met Public object.exp_type in
              let (path, desc) =
                Env.lookup_value (Longident.Lident "*self*") env
              in
              let method_id = List.assoc met methods in
              let method_type = newvar () in
              let (obj_ty, res_ty) = filter_arrow env method_type in
              unify env obj_ty desc.val_type;
              unify env res_ty typ;
              Texp_apply
                ({exp_desc =
                    Texp_ident
                      (Path.Pident method_id,
                       {val_type = method_type; val_kind = Val_reg});
                  exp_loc = sexp.pexp_loc; exp_type = method_type;
                  exp_env = env},
                 [{exp_desc = Texp_ident (path, desc);
                   exp_loc = object.exp_loc; exp_type = desc.val_type;
                   exp_env = env}]),
              typ
          | _ ->
              Texp_send (object, Tmeth_name met),
              filter_method env met Public object.exp_type
        in
        {exp_desc = exp; exp_loc = sexp.pexp_loc; exp_type = typ;
         exp_env = env}
      with
        Unify _ -> raise (Error (e.pexp_loc, Undefined_method_err met))
      end
  | Pexp_new cl ->
      let (cl_path, cl_typ) =
        try Env.lookup_class cl env with
          Not_found -> raise (Error (sexp.pexp_loc, Unbound_class cl))
      in
      begin match cl_typ.cty_new with
        None -> raise (Error (sexp.pexp_loc, Virtual_class cl))
      | Some ty ->
          {exp_desc = Texp_new cl_path; exp_loc = sexp.pexp_loc;
           exp_type = instance ty; exp_env = env}
      end
  | Pexp_setinstvar (lab, snewval) ->
      begin try
        let (path, desc) = Env.lookup_value (Longident.Lident lab) env in
        match desc.val_kind with
          Val_ivar Mutable ->
            let newval = type_expect env snewval desc.val_type in
            let (path_self, _) =
              Env.lookup_value (Longident.Lident "*self*") env
            in
            {exp_desc = Texp_setinstvar (path_self, path, newval);
             exp_loc = sexp.pexp_loc; exp_type = instance Predef.type_unit;
             exp_env = env}
        | Val_ivar _ ->
            raise (Error (sexp.pexp_loc, Instance_variable_not_mutable lab))
        | _ -> raise (Error (sexp.pexp_loc, Unbound_instance_variable lab))
      with
        Not_found ->
          raise (Error (sexp.pexp_loc, Unbound_instance_variable lab))
      end
  | Pexp_override lst ->
      List.fold_right
        (fun (lab, _) l ->
           if List.exists (( = ) lab) l then
             raise (Error (sexp.pexp_loc, Value_multiply_overridden lab));
           lab :: l)
        lst [];
      let (path_self, {val_type = self_ty}) =
        try Env.lookup_value (Longident.Lident "*self*") env with
          Not_found -> raise (Error (sexp.pexp_loc, Outside_class))
      in
      let type_override (lab, snewval) =
        try
          let (path, desc) = Env.lookup_value (Longident.Lident lab) env in
          match desc.val_kind with
            Val_ivar _ -> path, type_expect env snewval desc.val_type
          | _ -> raise (Error (sexp.pexp_loc, Unbound_instance_variable lab))
        with
          Not_found ->
            raise (Error (sexp.pexp_loc, Unbound_instance_variable lab))
      in
      let modifs = List.map type_override lst in
      {exp_desc = Texp_override (path_self, modifs); exp_loc = sexp.pexp_loc;
       exp_type = self_ty; exp_env = env}
  | Pexp_call (e1, e2) ->
      type_call env sexp.pexp_loc Sync "send_synchrone" [] e1 e2
  | Pexp_def (spat_sexp_list, sbody) ->
      let (pat_exp_list, new_env) =
        type_def env sexp.pexp_loc spat_sexp_list
      in
      let body = type_exp new_env sbody in
      {exp_desc = Texp_let (Nonrecursive, pat_exp_list, body);
       exp_loc = sexp.pexp_loc; exp_type = body.exp_type; exp_env = env}
  | Pexp_thismodule ->
      {exp_desc = Texp_thismodule; exp_loc = sexp.pexp_loc;
       exp_type = mkmodty env "Join" "module_data"; 
        exp_env = env}
(* inutile de creer un nouveau thread pour un call asynchrone sans calculs *)
  | Pexp_process ({ pexp_desc = Pexp_call(e1,e2) } as sexp)
      when not_computed env e1 && not_computed env e2
    ->
        type_call env sexp.pexp_loc Async "send_asynchrone" [] e1 e2
  | Pexp_process process ->
      let (process, _, _, _) = (!type_process) env process ([],[],true) in
      mkexp
        (Texp_apply
           (type_exp env (mkmodpident "Join" "launch"),
            [mkprocedure process env]))
        (instance Predef.type_unit) env
  | Pexp_location (loc_list,sbody) ->
      let (loc_defs,chan_defs,threads_exp,loc_env,new_env) =
        (!type_locations)  env sexp.pexp_loc loc_list
      in
      let body =  type_exp new_env sbody in
      mkexp
        (Texp_let
           (Nonrecursive,
            loc_defs,
            mkexp
              (Texp_let
                 (Nonrecursive,
                  chan_defs,
                  mkexp
                    (Texp_sequence
                       (threads_exp,
                        body))
                    body.exp_type new_env))
              body.exp_type loc_env
              ))
        body.exp_type env
  | Pexp_vartype  ->
      let ty = newvar () in
      {
      exp_desc = 
      Texp_metatype (function () ->
        Runtypes.to_meta env sexp.pexp_loc ty);
        exp_loc = sexp.pexp_loc;
        exp_type = instance (Predef.type_metatype ty);
        exp_env = env}

  | _ -> failwith "typecore.ml: Pexp_... not implemented"

and type_record env pexp_loc srecord label =
  try
    match srecord.pexp_desc with
      Pexp_ident lid ->
        let (path, desc) = Env.lookup_value lid env in
        begin
          match desc.val_kind,desc.val_type.desc with
            Val_ivar _, _ -> raise Not_found
          | _, Trecord (ty_cstr,lbls)  ->

                let (num,mut,ty) = 
                  find_prefixed_label pexp_loc lbls label in
                ({exp_desc =
                   Texp_ident (path, desc);
                   exp_loc = pexp_loc; 
                   exp_type = instance desc.val_type;
                   exp_env = env}
                ,Tlabel_val num, mut, ty)
                  
          | _ -> raise Not_found
        end
    | _ -> raise Not_found
  with
    Not_found ->
      let record = type_exp env srecord in
      let label =
        try Env.lookup_label label env with
          Not_found -> raise (Error (pexp_loc, Unbound_label label))
      in
      let (ty_arg, ty_res) = instance_label label in
      unify_exp env record ty_res;
      (record,Tlabel_name label,label.lbl_mut, ty_arg)

and join_apply prim args env =
  let exp = type_exp env (mkmodpident "Join" prim) in
  mkexp (Texp_apply (exp,args))
    (let rec iter args ty =
      match args with
        [] -> ty
      | arg::args -> 
          match (repr ty).desc with
            Tarrow (_,ty) -> iter args ty
          | _ -> jerror arg.exp_loc ("Join."^prim^" apply to too many args")
    in
    instance (iter args exp.exp_type))
    env
   
and type_call env pexp_loc sync prim extra_args e1 e2 =
  let channel = type_exp env e1 in
  let (tyl, rty) =
    try
      filter_channel env sync (instance channel.exp_type)
    with
      Failure s -> raise (Error (pexp_loc, 
                                 SimpleTypeError (s,channel.exp_type)))
  in
  let arg_exp = type_expect env e2 tyl in
  match rty with
    Tasync -> 
      join_apply prim
        (channel::arg_exp::extra_args) env
  | Tsync ty ->
      {exp_desc =
        Texp_apply
          (type_exp env (mkmodpident "Join" prim),
           channel::arg_exp::extra_args);
        exp_loc = pexp_loc; exp_type = ty; exp_env = env}

and type_expect env sexp ty_expected =
  match sexp.pexp_desc with
    Pexp_constant (Const_string s as cst) ->
      let exp =
        {exp_desc = Texp_constant cst; exp_loc = sexp.pexp_loc;
         exp_type =
           begin match (repr ty_expected).desc with
             Tconstr (path, _, _) when Path.same path Predef.path_format ->
               type_format sexp.pexp_loc s
           | _ -> instance Predef.type_string
           end;
         exp_env = env}
      in
      unify_exp env exp ty_expected; exp
  | Pexp_let (rec_flag, spat_sexp_list, sbody) ->
      let (pat_exp_list, new_env) = type_let env rec_flag spat_sexp_list in
      let body = type_expect new_env sbody ty_expected in
      {exp_desc = Texp_let (rec_flag, pat_exp_list, body);
       exp_loc = sexp.pexp_loc; exp_type = body.exp_type; exp_env = env}
  | Pexp_def (spat_sexp_list, sbody) ->
      let (pat_exp_list, new_env) =
        type_def env sexp.pexp_loc spat_sexp_list
      in
      let body = type_expect new_env sbody ty_expected in
      {exp_desc = Texp_let (Nonrecursive, pat_exp_list, body);
       exp_loc = sexp.pexp_loc; exp_type = body.exp_type; exp_env = env}
  | Pexp_sequence (sexp1, sexp2) ->
      let exp1 = type_statement env sexp1 in
      let exp2 = type_expect env sexp2 ty_expected in
      {exp_desc = Texp_sequence (exp1, exp2); exp_loc = sexp.pexp_loc;
       exp_type = exp2.exp_type; exp_env = env}
  | _ -> let exp = type_exp env sexp in unify_exp env exp ty_expected; exp

and type_statement env sexp =
  let exp = type_exp env sexp in
  match (repr exp.exp_type).desc with
    Tarrow (_, _) ->
      Location.print_warning sexp.pexp_loc
        "\
this function application is partial,
maybe some arguments are missing."
          ;
      exp
  | _ -> exp

and type_cases env ty_arg ty_res caselist =
  List.map
    (fun (spat, sexp) ->
       let (pat, ext_env) = type_pattern env spat in
       unify_pat env pat ty_arg;
       let exp = type_expect ext_env sexp ty_res in pat, exp)
    caselist

and type_let env rec_flag spat_sexp_list =
  begin_def ();
  let (pat_list, new_env) =
    type_pattern_list env (List.map (fun (spat, sexp) -> spat) spat_sexp_list)
  in
  let exp_env =
    match rec_flag with
      Nonrecursive -> env
    | Recursive -> new_env
  in
  let exp_list =
    List.map2 (fun (spat, sexp) pat -> type_expect exp_env sexp pat.pat_type)
      spat_sexp_list pat_list
  in
  List.iter2 (fun pat exp -> Parmatch.check_partial pat.pat_loc [pat, exp])
    pat_list exp_list;
  end_def ();
  List.iter2
    (fun pat exp ->
       if not (is_nonexpansive exp) then
         iter_pattern (fun pat -> make_nongen pat.pat_type) pat)
    pat_list exp_list;
  List.iter (fun pat -> iter_pattern (fun pat -> generalize pat.pat_type) pat)
    pat_list;
  List.combine pat_list exp_list, new_env

and type_def env syntax_loc def_list =
  let (decls,idents,new_env) = (!type_defs) env syntax_loc
      (List.map (fun d -> (None,(d:(Parsetree.pattern * Parsetree.expression) list),syntax_loc)) def_list)
  in
  decls,new_env

(* Typing of toplevel bindings *)

let type_binding env rec_flag spat_sexp_list =
  Typetexp.reset_type_variables (); 
  type_let env rec_flag spat_sexp_list
;;

(* Typing of toplevel expressions *)

let type_expression env sexp =
  Typetexp.reset_type_variables ();
  begin_def ();
  let exp = type_exp env sexp in
  end_def ();
  if is_nonexpansive exp then generalize exp.exp_type
  else make_nongen exp.exp_type;
  exp
;;

(* Typing of methods *)

let rec type_expect_fun env sexp ty_expected =
  match sexp.pexp_desc with
    Pexp_function caselist ->
      let (ty_arg, ty_res) =
        try filter_arrow env ty_expected with
          Unify _ -> raise (Error (sexp.pexp_loc, Too_many_arguments))
      in
      let cases =
        List.map
          (fun (spat, sexp) ->
             let (pat, ext_env) = type_pattern env spat in
             unify_pat env pat ty_arg;
             let exp = type_expect_fun ext_env sexp ty_res in pat, exp)
          caselist
      in
      Parmatch.check_unused cases;
      Parmatch.check_partial sexp.pexp_loc cases;
      {exp_desc = Texp_function cases; exp_loc = sexp.pexp_loc;
       exp_type = newty (Tarrow (ty_arg, ty_res)); exp_env = env}
  | _ -> type_expect env sexp ty_expected
;;

let type_method env self self_name meths sexp ty_expected =
  let meths = ref meths in
  let (obj, env) =
    Env.enter_value "*self*" {val_type = self; val_kind = Val_reg} env
  in
  let pattern =
    {pat_desc = Tpat_var obj; pat_loc = Location.none; pat_type = self}
  in
  let (pattern, env) =
    match self_name with
      None -> pattern, env
    | Some name ->
        let (self_name, env) =
          Env.enter_value name {val_type = self; val_kind = Val_self meths}
            env
        in
        {pat_desc = Tpat_alias (pattern, self_name); pat_loc = Location.none;
         pat_type = self},
        env
  in
  let exp = type_expect_fun env sexp ty_expected in
  {exp_desc = Texp_function [pattern, exp]; exp_loc = sexp.pexp_loc;
   exp_type = newty (Tarrow (pattern.pat_type, exp.exp_type)); exp_env = env},
  !meths
;;
