(***********************************************************************)
(*                                                                     *)
(*                      Join Objective Caml                            *)
(*                                                                     *)
(*        Fabrice LE FESSANT, Project PARA, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: typejoin.ml,v 1.8 1998/05/11 15:45:25 lefessan Exp $ *)

(* Typechecking for the core language *)

open Misc;;
open Asttypes;;
open Longident;;
open Parsetree;;
open Types;;
open Typedtree;;
open Btype;;
open Ctype;;
open Typemisc;;
open Typecore;;

let error = jerror
let add_value name ty env =
  Env.add_value name { val_type = ty; val_kind = Val_reg } env

let rec translate_automata env  (location,def_list,syntax_loc) =
  let nchannels = ref (-1) in
  let (channels, processes) =
    List.fold_left
      (fun (channels, processes) (pattern, process) ->
        let join =
          {join_pattern = []; join_args = []; join_vars = [];
            join_replies = []; join_id = Ident.create "#process#";
            join_types = []}
        in
        let rec iter channels pattern =
          match pattern.ppat_desc with
            Ppat_or (pat1, pat2) -> iter (iter channels pat2) pat1
          | Ppat_channel (name, pat, sync) ->
              let _ = pattern_variables := [] in
              let (channel, channels) =
                try
                  let channel = List.assoc name channels in
                  match sync, channel.chan_sync with
                    Sync, Tasync | Async, Tsync _ ->
                      error pattern.ppat_loc 
                        ("Different synchrony for channel "^name)
                  | _ -> channel, channels
                with
                  Not_found ->
                    incr nchannels;
                    let tyl = newvar () in
                    let treply =
                      match sync with
                        Sync -> Tsync (newvar ())
                      | Async -> Tasync
                    in
                    let channel =
                      {chan_ext_ident = Ident.create name;
                        chan_in_ident = Ident.create name;
                        chan_args = tyl;
                        chan_type = newty (Tchannel (tyl, treply));
                        chan_types = []; 
                        chan_ord = !nchannels; 
                        chan_sync = treply}
                    in
                    channel, (name, channel) :: channels
              in
              if List.mem channel.chan_ord join.join_pattern then
                error pattern.ppat_loc 
                  ("Twice channel "^name^" in one join pattern");
              join.join_pattern <- channel.chan_ord :: join.join_pattern;
              let args = pat in
              let pat_args = type_pat env args in
              let _ =
                unify env pat_args.pat_type channel.chan_args 
              in
              let pat_args =
                match channel.chan_sync with
                  Tasync -> pat_args
                | Tsync ty ->
                    let tyj = join_type "reply_event" env in
                    let id =
                      enter_variable !synthetic_loc ("#reply#" ^ name) 
                        (ty,Val_reg)
                    in
                    join.join_replies <- (name, id) :: join.join_replies;
                    mkpattuple [mkpat (Tpat_var id) tyj; pat_args]
               in
               let _ =
                 join.join_args <-
                    (channel, pat_args) :: join.join_args;
                 join.join_vars <- !pattern_variables @ join.join_vars;
                 join.join_types <-
                    List.map (fun (v,(ty,_)) -> ty) !pattern_variables 
                    :: join.join_types
               in
               channels
          | _ -> failwith "Unknown pattern in join pattern"
        in
        let channels = iter channels pattern in
        join.join_pattern <- Sort.list (<) join.join_pattern;
        List.iter (fun (pjoin,_) ->
           if pjoin.join_pattern = join.join_pattern then
             error syntax_loc "Twice processes with same synchronization"
               ) processes;
         channels, (join, process) :: processes)
      ([], []) def_list
  in
  let channels = Sort.list (fun (_,x) (_,y) -> x.chan_ord <= y.chan_ord) (channels) in
  let automata = Ident.create "#automata#" in
  let nmult,nchans,f = Janalyse.fusion (List.length channels) 
      (List.map (fun (join,_) -> join.join_pattern) processes) in
  (automata,location,channels,processes,syntax_loc,nmult,nchans,f)
    
and
    type_defs env syntax_loc automata_list =
  try
  begin_def ();  
  let automata_type = mkmodty env "Join" "automata"
  and process_type = mkmodty env "Join" "process" in
  let automata_list = List.map (translate_automata env) automata_list
  in
  let tuple_type =
    newty
      (Ttuple (List.flatten
                 (List.map (fun (_,_,channels,_,_,_,_,_) ->
                   (List.map (fun (name, channel) -> channel.chan_type) 
                      channels))
                    automata_list)))
  in
  let chans_env =
    List.fold_right
      (fun (automata,_,channels,_,_,_,_,_) env ->
        List.fold_right
          (fun (name, channel) env ->
            add_value channel.chan_in_ident channel.chan_type env)
          channels (add_value automata  automata_type env)
          ) automata_list env
  in
  let exp =
    List.fold_left 
      (fun
        exp 
          (automata,location,channels,processes,syntax_loc,nmult,nchans,f) ->
            let chans_env =
              List.fold_right
                (fun (name, channel) env ->
                  add_value channel.chan_in_ident channel.chan_type env)
                channels 
                (add_value automata automata_type chans_env)
            in
            let (decls,processes) =
              List.fold_left (fun (decls,processes) (join,process) ->
                let (process_exp,thread) = 
                  make_process_def chans_env automata f (join, process)
                in
                let thread =
                  match thread with
                    None ->
                      if simple_thread chans_env process then
                        Some (-1)
                      else
                        None
                  | Some (name,reply_name) -> 
                      Some
                        (try List.assoc name channels
                        with
                          Not_found -> failwith "fill_automata"
                              ).chan_ord
                in
                (mkpat (Tpat_var join.join_id)
                   process_type,process_exp)::decls,
                (join,process_exp,thread)::processes)
                  ([],[]) processes
            in
            let (procs_env : Env.t) =
              List.fold_right
                (fun (join, process,thread) env ->
                  add_value join.join_id process_type env)
                processes chans_env
            in
            let structure = Janalyse.structure procs_env nmult nchans f processes in
            let trigger_list = List.map (fun (join,process,thread) -> process) processes
            in                     
            mksequence
              (mklet decls
                 (join_apply "fill_automata"
                    [(ident_to_tident automata procs_env);
                      mktuple trigger_list procs_env;
                      structure
                    ]
                    chans_env)
                 chans_env)
              exp
              chans_env
              ) 
      (mktuple
         (List.flatten
            (List.map (fun (_,_,channels,_,_,_,_,_) ->
              (List.map (fun (name, channel) -> 
                ident_to_tident channel.chan_in_ident
                   chans_env
                  ) 
                 channels))
               automata_list)) chans_env) automata_list
  in
  end_def();
  let idents = 
    List.map
      (fun (automata,location,channels,processes,syntax_loc,_,_,_) ->
        List.iter
          (fun (join, process) ->
            let tys = ref [] in
            List.iter
              (fun ty ->
                let newtys = ref [] in
                let rec f ty =
                  if List.memq ty !tys then make_nongen ty
                  else 
                    if not (List.memq ty !newtys) then
                    begin
                    newtys := ty :: !newtys; 
                    iter_type_expr f ty; () end
                in
                List.iter f ty; tys := !newtys @ !tys)
              join.join_types)
          processes;
        
        List.iter (fun (name, channel) -> 
          generalize channel.chan_type) channels;
        (location,
         List.map
           (fun (name, channel) ->
             channel.chan_ext_ident, channel.chan_type
               ) channels)
          ) automata_list
  in
  check_multiple_define syntax_loc idents;
    [(mkpattuple
        (List.flatten
           (List.map
              (fun (_,_,channels,_,_,_,_,_) ->
                List.map
                  (fun (name, channel) ->
                    mkpat (Tpat_var channel.chan_ext_ident) 
                      channel.chan_type)
                  channels)
              automata_list
              )
           )
        )
        ,
    mklet 
         (List.map 
            (fun
              (automata,location,channels,processes,
               syntax_loc,nmult,nchans,f) ->
                let chan_list =
                  (mkpat (Tpat_var automata) automata_type)::
                  (List.map
                     (fun (name, channel) ->
                       mkpat (Tpat_var channel.chan_in_ident) 
                         channel.chan_type)
                     channels)
                in
                let indexes =  
                  mktuple
                    (List.map (fun i-> mkint i env)
                       (Array.to_list f)) env in
                (mkpattuple chan_list,
                 match location with
                   None ->
                     join_apply "create_empty_automata"
                       [mkint (List.length processes) env;
                         mkint nchans env;
                         indexes
                       ]
                       env
                 | Some location ->
                     join_apply "create_empty_automata_in"
                       [ident_to_tident location env;
                         mkint (List.length processes) env; 
                         mkint nchans env;
                         indexes
                       ] env
                       )
                  )
            automata_list)
        exp
        env

  ], 
    idents,
    List.fold_left 
    (fun env (_,loc_idents) ->
      List.fold_left
        (fun env (ident,idtype) -> add_value ident idtype env) env loc_idents)
    env idents
  with
    Not_found ->
      failwith "in type_defs"


and make_process_def env automata f (join, process) =
  let new_env =
    List.fold_right
      (fun (id, (ty,vl)) env ->
         Env.add_value id {val_type = ty; val_kind = vl} env)
      join.join_vars env
  in
  let (process_exp, replies, traps, thread) =
    begin_def (); 
    type_process_and_trap_exc new_env process (join.join_replies,[],true)
  in
  end_def ();
  List.iter
    (fun (name, id) ->
      if not (List.mem name replies) then
        error process.pexp_loc ("No reply to channel " ^ name))
    join.join_replies;
  mkexp
    (Texp_function
       (let cases = [(match join.join_args with
          [ord, pat] -> pat
        | list ->
            let sorted_args =
              Sort.list (fun x y -> 
                (f.((fst x).chan_ord) land 63) <= 
                (f.((fst y).chan_ord) land 63))
                join.join_args
            in
            mkpattuple (List.map (fun (ord, pat) -> pat) sorted_args)),
                     process_exp]
        in
          Parmatch.check_partial (!synthetic_loc) cases;
          cases
       ))
    (mkmodty env "Join" "process") env,
    thread

and type_process_and_trap_exc env proc context = 
   let (exp,replies,traps,thread) = type_process env proc context 
   in
   match traps with
     [] -> (* pas de reponses: inutile de faire un trap *)
       (exp,replies,traps,thread)
   | traps ->
       (* des reponses : on place un handler d'exceptions autour du
          process dans lequel on expedie l'exception au canaux synchrones
          qui aurait du recevoir une reponse, puis on kill le thread *)
       let exn_id = Ident.create "exception" in
       let exn_type = instance Predef.type_exn in
       let new_env = add_value exn_id exn_type env in 
       (mkexp (Texp_try(exp,[
                        mkpat (Tpat_var exn_id) exn_type,
                        join_apply "send_exception" 
                          [ident_to_tident exn_id new_env;
                            mktuple (List.map (fun reply_name ->
                              ident_to_tident reply_name new_env
                              ) traps ) new_env 
                          ] new_env
                      ])) exp.exp_type env,
        replies, traps, thread)

(*
  Pour diminuer le nombre de threads lances, on peut ne pas forker 
  les threads non calculatoires, en posant thread = Some (-1).
  Toutefois, il faut noter que souvent, ces threads se terminent par
  un Join.exit ().
  L'autre solution est de ne forker que si necessaire, ie quand le scheduler
  ordonne la preemption du thread, ou que l'on fait un get_thread_stub.
  Et pourquoi pas les deux en meme temps ?

  En gros, on peut utiliser le <sync> pour reprendre le thread courant
  et l'utiliser a nouveau. Pour cela, il faut que trigger_def prenne la
  main immediatement, ce qui ne peut se faire que s'il est appele avec 
  un wrapper autour.

  Quelles criteres pour ces treads:
    Pas de reply
  
proc = { i:=!i+1; } | { k(j) }
 ---> i:=!i+1;k j;exit

Que faire des sends asynchrones terminaux ? Ils peuvent faire exit d'abord... 
mais au risque de vraiment tuer le thread !
*)

and simple_thread env proc =
  (* toutes les operations qu'il contient sont simples *)
  let rec simple_thread proc =
    match proc.pexp_desc with
      Pexp_null -> true
    | Pexp_par list -> List.for_all simple_thread list
    | Pexp_call (e1,e2) -> not_computed env e1 && not_computed env e2
    | Pexp_let (_,list,p)     -> simple_thread p &&
        List.for_all (function (pat,e) -> not_computed env e) list
    | Pexp_def (_,p) -> simple_thread p
    | Pexp_ifthenelse (e1,p,option)  -> option_simple_thread option &&
        not_computed env e1 && simple_thread p
    | Pexp_location _
    | Pexp_reply _
    |  _ -> false

  and option_simple_thread = function
      None -> true
    | Some e -> simple_thread e
  in
  simple_thread proc

and type_process env proc = function (channels,replies,autonomep) ->
  try
    match proc.pexp_desc with
      Pexp_null ->
          (join_apply "exit" [mkunit env] env,
           replies,[],None)
    | Pexp_par list ->
        let (sends,replys,threads,thread) =
          match List.fold_left (fun (s,r,a) p ->
              match p with
                { pexp_desc = Pexp_call (e1,e2) } 
                  when not_computed  env e1 &&  not_computed env e2
                -> p::s , r, a
              | { pexp_desc = Pexp_reply (e,_) } when not_computed env e
                -> s, p::r,a
              | _ -> s , r, p::a) ([],[],[]) list
            with
              [],[],[] -> error proc.pexp_loc "{}|{} not accepted"
            | p::s , [], [] ->  s, [], [], p
            | s, p::r, [] -> s, r, [], p
            | s, r, p::t -> s, r, t, p
        in
        let proc_sends =
          List.map 
            (fun p ->
              let (p,_,_,_) = type_process env p (channels,replies,false)
              in p
                ) sends 
        in
        let proc_replies =
          List.fold_left 
            (fun (ps,replies,all_traps) p -> 
              let (p,replies,traps,_) = 
                type_process env p (channels,replies,false)
              in
              p::ps, replies, traps@all_traps
                                       ) (proc_sends,replies,[]) replys
        in
          let (proc_threads,replies, all_traps) =
            List.fold_left 
              (fun (ps,replies,all_traps) p ->
                let (p,replies,traps,thread) = 
                  type_process_and_trap_exc env p (channels,replies,true)
                in
                let p =
                match thread with
                  None ->
                    join_apply "launch" [mkprocedure p env] env
                | Some (name,reply_name) ->
                    join_apply "reuse_thread"
                      [ident_to_tident reply_name env;
                        mkprocedure p env]
                      env
                in
                p::ps, replies, traps@all_traps
                  ) proc_replies threads
          in
          (* le thread principal *)
          let (p,replies,traps,thread) = 
            (match all_traps with
              [] -> type_process (* pas de replies supprimees *)
            | _ ->
                type_process_and_trap_exc)
              env thread (channels,replies,true)
          in
          (List.fold_left (fun exp_end p -> mksequence p exp_end env) 
             p proc_threads), 
          replies, traps@all_traps ,thread
            
      | Pexp_let (rec_flag, spat_sexp_list, sbody) ->
          let (pat_exp_list, new_env) = type_let env rec_flag spat_sexp_list in
          let (body, replies, traps, thread) = type_process new_env sbody 
                                          (channels,replies,true) in
            ({exp_desc = Texp_let (rec_flag, pat_exp_list, body);
              exp_loc = proc.pexp_loc; 
              exp_type = body.exp_type; 
              exp_env = env},
             replies, traps, thread)
      | Pexp_location (loc_list, sbody) ->
          let (loc_defs,chan_defs,threads_exp,loc_env,new_env) =
            type_locations  env proc.pexp_loc loc_list
          in
          let (body, replies, traps, thread) =  
            type_process new_env sbody (channels,replies,true)
          in
          mklet loc_defs 
            (mklet chan_defs
               (mksequence threads_exp body new_env)
               loc_env)
            env,
          replies, traps, thread
      | Pexp_def (spat_sexp_list, sbody) ->
          let (pat_exp_list, new_env) =
            type_def env proc.pexp_loc spat_sexp_list
          in
          let (body, replies, traps, thread) = type_process new_env sbody
                                          (channels,replies,true) in
            ({exp_desc = Texp_let (Nonrecursive, pat_exp_list, body);
              exp_loc = proc.pexp_loc;
              exp_type = body.exp_type; 
              exp_env = env},
             replies, traps, thread)
    | Pexp_ifthenelse (scond, p, q) ->
        let cond = type_expect env scond (instance Predef.type_bool) in
        let (proc_p, repl_p, traps_p, thread_p) = type_process env p 
                                          (channels,replies,true) in
        let (proc_q, repl_q, traps_q, thread_q) = 
          match q with
            Some q -> type_process env q (channels,replies,true) 
          | None ->
              (join_apply "exit" [mkunit env] env,
               replies, [], None)
        in
        if List.length repl_p <> List.length repl_q ||
           List.exists (fun e -> not (List.mem e repl_p)) repl_q then
          error proc.pexp_loc "Not the same replies in IF alternatives";
          ({exp_desc = Texp_ifthenelse (cond, proc_p, Some proc_q);
            exp_loc = proc.pexp_loc; exp_type = proc_p.exp_type;
             exp_env = env},
           repl_p, traps_p,
           if thread_p = thread_q then thread_p else None)
    | Pexp_call (e1, e2) ->
        (type_call env proc.pexp_loc Async       
           (if autonomep then "send_asynchrone_and_exit" 
           else "send_asynchrone")
           [] e1 e2,
         replies, [], None)
(* reply( let defs in e ) -> let defs in reply e *)
    | Pexp_reply ( 
      { pexp_desc = Pexp_let (rec_flag, spat_sexp_list, sbody);
        pexp_loc = exp_loc
      }, name) ->
        type_process env 
          { pexp_desc = Pexp_let(rec_flag, spat_sexp_list, 
                                   { pexp_desc = Pexp_reply(sbody,name); 
                                     pexp_loc = proc.pexp_loc});
            pexp_loc = exp_loc
          } (channels,replies,autonomep)

(* reply( let defs in e ) -> let defs in reply e *)
     | Pexp_reply ( 
      { pexp_desc = Pexp_def (spat_sexp_list, sbody);
        pexp_loc = exp_loc
      }, name) ->
        type_process env 
          { pexp_desc = Pexp_def(spat_sexp_list, 
                                   { pexp_desc = Pexp_reply(sbody,name); 
                                     pexp_loc = proc.pexp_loc});
            pexp_loc = exp_loc
          } (channels,replies,autonomep)

(* reply( if c then e1 else e2 ) -> if c then reply e1 else reply e2 *)
     | Pexp_reply ( 
      { pexp_desc = Pexp_ifthenelse(cond,e1,option);
        pexp_loc = exp_loc
      }, name) ->
        type_process env 
          { pexp_desc = 
            Pexp_ifthenelse(cond,  
                     { pexp_desc = Pexp_reply(e1,name); 
                       pexp_loc = proc.pexp_loc},
                     Some { pexp_desc = Pexp_reply(
                       (match option with
                         Some e2 -> e2
                       | None -> { pexp_desc = 
                                   Pexp_construct (Lident "()", None, false);
                                   pexp_loc = exp_loc}
                             ),name);
                       pexp_loc = proc.pexp_loc}
                       );
            pexp_loc = exp_loc
          } (channels,replies,autonomep)

(* reply( s << n >> ) -> tailcall *)
     | Pexp_reply ( 
       { pexp_desc = Pexp_apply({ pexp_desc = Pexp_ident lid} as e1,[e2]);
         pexp_loc = exp_loc
       }, name) when
       (try
         let (path,desc) = Env.lookup_value lid env
         in
         match desc.val_type.desc with
           Tchannel(t1,Tsync t2) -> true
         | _ -> false
       with
         Not_found -> false)
       ->
         let name,reply_name,lid = 
           match (name,channels) with
             (None,[name,reply_name]) ->
               (name,reply_name,(Longident.Lident name))
           | (Some name,_) ->
               (name,(
                try List.assoc name channels
                with
                  Not_found -> 
                    error proc.pexp_loc 
                      ("Reply to channel external from def "^name)
                      ),(Longident.Lident name) )
           | (None,_) ->
               error proc.pexp_loc "No channel for reply"
         in
         if List.mem name replies then
           error proc.pexp_loc ("Two replies to the same channel "^name);
         begin
           try
             let (path, desc) = Env.lookup_value lid env 
             in           
             match path with
               Path.Pident ident ->
                 begin match (Ctype.repr desc.val_type).desc with
                   Tchannel (ty1, ty2) ->
                     begin match ty2 with
                       Tasync -> error proc.pexp_loc 
                           "Reply to asynchronous channel"
                     | Tsync ty ->
                         Debug.print0 'f' "Tailcall\n";
                         let id = ident_to_tident reply_name env in
                         let exp =
                           type_call env exp_loc Sync "tail_synchrone"
                             [id] e1 e2
                         in
                         begin try
                           unify env ty exp.exp_type
                         with
                           Unify trace ->
                             raise(Error(proc.pexp_loc, 
                                         Expr_type_clash(trace)))
                         end;
                         {exp_desc = exp.exp_desc;
                           exp_loc = exp.exp_loc;
                           exp_type = ty; 
                           exp_env = env}
                           , name::replies, [reply_name],
                         (Some (name,reply_name))
                  end
               | _ -> error proc.pexp_loc 
                    "Reply to something not a channel"
              end
           | _ -> error proc.pexp_loc "Bad name in Reply"
         with
           Not_found -> raise (Error (exp_loc, Unbound_value lid))
         end
(* type_proc pexp_reply *)         
    
          (* normal replies *)
    | Pexp_reply (sexp, name) ->
        let name,reply_name,lid = 
          match (name,channels) with
            (None,[]) -> 
              error proc.pexp_loc "Reply with no synchronous channel"
          | (None,[name,reply_name]) ->
              (name,reply_name,(Longident.Lident name))
          | (None,_) -> error proc.pexp_loc 
                "Anonymous reply with several synchronous channels"
          | (Some name,_) ->
              (name,(
               try List.assoc name channels
               with
                 Not_found -> 
                   error proc.pexp_loc
                     ("Reply to channel external from def "^name)
                     ),(Longident.Lident name) )
        in
        if List.mem name replies then
          error proc.pexp_loc
            ("Two replies to the same channel "^name);
        try
          let (path, desc) = 
            try
              Env.lookup_value lid env 
            with
              Not_found ->
                failwith "lookup_value in Pexp_reply"
          in
          
          match path with
            Path.Pident ident ->
               begin match (Ctype.repr desc.val_type).desc with
                Tchannel (ty1, ty2) ->
                  begin match ty2 with
                    Tasync -> 
                      error proc.pexp_loc 
                        "Reply to asynchronous channel"
                  | Tsync ty ->
                      let id = ident_to_tident reply_name env in
                      let exp = type_expect env sexp ty in
                      let apply_Jrt = 
                        join_apply 
                          (if autonomep then "reply_to_and_exit"
                          else "reply_to")
                          [id; exp] env
                      in
                      apply_Jrt, name::replies, [reply_name],
                      (Some (name,reply_name))
                  end
               | _ -> 
                   raise 
                     (Error(proc.pexp_loc,
                            Expr_type_clash
                              (let t2 =
                                newty (Tchannel(newvar(),
                                                Tsync (newvar()))) in
                              [(desc.val_type,desc.val_type);
                                (t2,t2)])))
              end
          | _ -> error proc.pexp_loc "Bad name in Reply"
        with
          Not_found -> raise (Error (sexp.pexp_loc, Unbound_value lid))
  with
    Not_found ->
      failwith "not found in process"


and
    type_locations env syntax_loc location_list =
  let location_type = mkmodty env "Join" "location" in
  let location_list =
    List.map 
      (fun (loc_name,automatas,pexp) -> 
        let ident = Ident.create loc_name in
        (ident,automatas,pexp)
          ) location_list
  in
  let location_env = (* env + locations *)
    List.fold_left 
      (fun env (ident,automatas,exp) ->
        Env.add_value ident 
          {val_type = location_type; 
            val_kind = Val_reg} env
          ) env location_list
  in
  let (decls,idents,automata_env) = type_defs location_env syntax_loc 
      (List.flatten
         (List.map
            (fun (ident,automatas,exp) ->
              List.map 
                (fun automata ->
                  (Some ident,automata,syntax_loc)
                    ) automatas)
            location_list))
  in
  (List.map
     (fun (ident,automatas,exp) ->
       (mkpat (Tpat_var ident) location_type),
       (join_apply "create_agent"
          [mkexp Texp_thismodule (join_type "module_data" env) env]
          env)
         ) location_list),
  decls,
  (let start_thread (location,automatas,exp) =
    let new_env =
      add_value location location_type
        (List.fold_left
           (fun env (l,chans) ->
             match l with
               Some l when Ident.same l location ->
                 List.fold_left 
                   (fun env (ident,idtype) ->
                     add_value ident  idtype env) env chans
             | _ -> env) automata_env idents)
    in
    join_apply "agent_thread"
          [ident_to_tident location new_env;
            let exp = type_exp new_env exp
            in
            mkprocedure exp new_env
          ] new_env
  in
  List.fold_left
    (fun exp tuple ->
      mksequence  exp (start_thread tuple) automata_env)
    (start_thread (List.hd location_list))
    (List.tl location_list)
    ),
  location_env,
  automata_env
    
(* Typing of toplevel def bindings *)

let type_loc_binding env location loc_list =
  Typetexp.reset_type_variables (); 
  type_locations env location loc_list
;;
let type_join_binding env location ddefs =
  Typetexp.reset_type_variables (); 
  type_def env location ddefs
;;

let _ = 
  Typecore.type_defs := type_defs;
  Typecore.type_locations := type_locations;
  Typecore.type_process := type_process
