(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: typemisc.ml,v 1.5 1998/05/11 07:28:30 lefessan Exp $ *)

(* Typechecking for the core language *)

open Misc
open Asttypes
open Longident
open Parsetree
open Types
open Typedtree
open Btype
open Ctype
open Primitive

type error =
    Unbound_value of Longident.t
  | Unbound_constructor of Longident.t
  | Unbound_label of Longident.t
  | Constructor_arity_mismatch of Longident.t * int * int
  | Label_mismatch of Longident.t * (type_expr * type_expr) list
  | Pattern_type_clash of (type_expr * type_expr) list
  | Multiply_bound_variable
  | Orpat_not_closed
  | Expr_type_clash of (type_expr * type_expr) list
  | Apply_non_function of type_expr
  | Label_multiply_defined of Longident.t
  | Label_missing
  | Label_not_mutable of Longident.t
  | Bad_format of string
  | Undefined_method_err of string
  | Unbound_class of Longident.t
  | Virtual_class of Longident.t
  | Unbound_instance_variable of string
  | Instance_variable_not_mutable of string
  | Not_subtype of (type_expr * type_expr) list * (type_expr * type_expr) list
  | Outside_class
  | Value_multiply_overridden of string
  | Coercion_failure of type_expr * type_expr * (type_expr * type_expr) list
  | Too_many_arguments
  | NewFailure of string
  | SimpleTypeError of string * type_expr

exception Error of Location.t * error

let jerror sloc str = raise (Error(sloc,NewFailure str)) 

(* Typing of concrete records *)

let find_prefixed_label sloc lbls = function
    Lident label ->
      let rec find count = function
          [] -> raise (Error(sloc,NewFailure ("Unknown label "^label)))
        | (name,mut,ty)::rem ->
            if name = label then (count,mut,ty)
            else find (count+1) rem
      in
      find 0 lbls
  | _ -> raise 
        (Error(sloc,NewFailure "No Absolute label in record constructor "))

(* Typing of constants *)

let type_constant = function
    Const_int _ -> instance Predef.type_int
  | Const_char _ -> instance Predef.type_char
  | Const_string _ -> instance Predef.type_string
  | Const_float _ -> instance Predef.type_float

(* Typing of patterns *)

let unify_pat env pat expected_ty =
  try
    unify env pat.pat_type expected_ty
  with Unify trace ->
    raise(Error(pat.pat_loc, Pattern_type_clash(trace)))

let pattern_variables = ref ([]: (Ident.t * (type_expr * value_kind)) list)

let enter_variable location name ty =
  if List.exists (fun (id, ty) -> Ident.name id = name) !pattern_variables
  then raise(Error(location, Multiply_bound_variable));
  let id = Ident.create name in
  pattern_variables := (id, ty) :: !pattern_variables;
  id

let pat_tuple = function
    { ppat_desc = Ppat_tuple list } -> list
  | _ -> failwith "Typecore.ml: match failure in pat_tuple"




let rec type_pat env sp =
  match sp.ppat_desc with
    Ppat_any ->
      { pat_desc = Tpat_any;
        pat_loc = sp.ppat_loc;
        pat_type = newvar() }
  | Ppat_var name ->
      let ty = newvar() in
      let id = enter_variable sp.ppat_loc name (ty,Val_reg) in
      { pat_desc = Tpat_var id;
        pat_loc = sp.ppat_loc;
        pat_type = ty }
  | Ppat_alias(sp, name) ->
      let p = type_pat env sp in
      let id = enter_variable sp.ppat_loc name (p.pat_type,Val_reg) in
      { pat_desc = Tpat_alias(p, id);
        pat_loc = sp.ppat_loc;
        pat_type = p.pat_type }
  | Ppat_constant cst ->
      { pat_desc = Tpat_constant cst;
        pat_loc = sp.ppat_loc;
        pat_type = type_constant cst }
  | Ppat_tuple spl ->
      let pl = List.map (type_pat env) spl in
      { pat_desc = Tpat_tuple pl;
        pat_loc = sp.ppat_loc;
        pat_type = newty (Ttuple(List.map (fun p -> p.pat_type) pl)) }
  | Ppat_construct(lid, sarg, explicit_arity) ->
      let constr =
        try
          Env.lookup_constructor lid env
        with Not_found ->
          raise(Error(sp.ppat_loc, Unbound_constructor lid)) in
      let sargs =
        match sarg with
          None -> []
        | Some {ppat_desc = Ppat_tuple spl} when explicit_arity -> spl
        | Some {ppat_desc = Ppat_tuple spl} when constr.cstr_arity > 1 -> spl
        | Some({ppat_desc = Ppat_any} as sp) when constr.cstr_arity > 1 ->
            replicate_list sp constr.cstr_arity
        | Some sp -> [sp] in
      let (ty_args, ty_res) = instance_constructor constr in
      begin
        match ty_args with
          Tconstr_tuple ty_args ->
            if List.length sargs <> constr.cstr_arity then
              raise(Error(sp.ppat_loc,
                          Constructor_arity_mismatch(
                          lid,constr.cstr_arity, List.length sargs)));
            let args = List.map (type_pat env) sargs in
            List.iter2 (unify_pat env) args ty_args;
           { pat_desc = Tpat_construct(constr, args);
             pat_loc = sp.ppat_loc;
             pat_type = ty_res }
        | Tconstr_record lbls ->
            if List.length sargs <> 1 then
              raise(Error(sp.ppat_loc,
                          Constructor_arity_mismatch(
                          lid,1, List.length sargs)));
            type_pat_in_constr (constr,sp.ppat_loc,ty_res,lbls)
              env (List.hd sargs)
      end
  | Ppat_record lid_sp_list ->
      let ty = newvar() in
      let type_label_pat (lid, sarg) =
        let label =
          try
            Env.lookup_label lid env
          with Not_found ->
            raise(Error(sp.ppat_loc, Unbound_label lid)) in
        let (ty_arg, ty_res) = instance_label label in
        begin try
          unify env ty_res ty
        with Unify trace ->
          raise(Error(sp.ppat_loc, Label_mismatch(lid, trace)))
        end;
        let arg = type_pat env sarg in
        unify_pat env arg ty_arg;
        (label, arg)
      in
      { pat_desc = Tpat_record(List.map type_label_pat lid_sp_list);
        pat_loc = sp.ppat_loc;
        pat_type = ty }
  | Ppat_or(sp1, sp2) ->
      let initial_pattern_variables = !pattern_variables in
      let p1 = type_pat env sp1 in
      let p2 = type_pat env sp2 in
      if !pattern_variables != initial_pattern_variables then
        raise(Error(sp.ppat_loc, Orpat_not_closed));
      unify_pat env p2 p1.pat_type;
      { pat_desc = Tpat_or(p1, p2);
        pat_loc = sp.ppat_loc;
        pat_type = p1.pat_type }
  | Ppat_constraint(sp, sty) ->
      let p = type_pat env sp in
      let ty = Typetexp.transl_simple_type env false sty in
      unify_pat env p ty;
      p
  | Ppat_channel(name,pat,sync) -> 
      failwith "Channel pattern outside join-pattern"

and type_pat_in_constr (constr,ppat_loc,ty_res,lbls) env sp =
  match sp.ppat_desc with
    Ppat_any ->
      { pat_desc = Tpat_construct
          (constr, 
           Array.to_list (Array.init (List.length lbls)
                            (fun i ->
                              { pat_desc = Tpat_any;
                                pat_loc = sp.ppat_loc;
                                pat_type = newvar ()}
                                )));
        pat_loc = ppat_loc;
        pat_type = ty_res }

  | Ppat_var name ->
      let p = 
        { pat_desc = Tpat_construct
            (constr, 
             Array.to_list (Array.init (List.length lbls)
                              (fun i ->
                                { pat_desc = Tpat_any;
                                  pat_loc = sp.ppat_loc;
                                  pat_type = newvar ()}
                                  )));
          pat_loc = ppat_loc;
          pat_type = ty_res }
      in
      let id = enter_variable sp.ppat_loc name
          (newty (Trecord (ty_res,lbls)), Val_record 0)
      in
      { pat_desc = Tpat_alias(p, id);
        pat_loc = sp.ppat_loc;
        pat_type = ty_res
      }
        
  | Ppat_alias(sp, name) ->
      let p = type_pat_in_constr (constr,ppat_loc,ty_res,lbls) env sp in
      let id = enter_variable sp.ppat_loc name
          (newty (Trecord (ty_res,lbls)), Val_record 0)
      in
      { pat_desc = Tpat_alias(p, id);
        pat_loc = sp.ppat_loc;
        pat_type = ty_res }

  | Ppat_or(sp1, sp2) ->
      let initial_pattern_variables = !pattern_variables in
      let p1 = type_pat_in_constr (constr,ppat_loc,ty_res,lbls) env sp1 in
      let p2 = type_pat_in_constr (constr,ppat_loc,ty_res,lbls) env sp2 in
      if !pattern_variables != initial_pattern_variables then
        raise(Error(sp.ppat_loc, Orpat_not_closed));
      unify_pat env p2 p1.pat_type;
      { pat_desc = Tpat_or(p1, p2);
        pat_loc = sp.ppat_loc;
        pat_type = ty_res }  

  | Ppat_constant _
  | Ppat_tuple _
  | Ppat_construct _
  | Ppat_constraint _
  | Ppat_channel _ ->
      jerror sp.ppat_loc "Unexpected pattern in Record Constructor"

  | Ppat_record lid_sp_list ->
      let numfields = ref 0 in
      let absents = ref 0 in
      let labels = ref [] in
      let args = Array.init (List.length lbls)
          (fun i ->
            { pat_desc = Tpat_any;
              pat_loc = sp.ppat_loc;
              pat_type = newvar ()}
              )
      in
      List.iter (fun (lid,pat) -> 
        let (n,mut,ty) = find_prefixed_label sp.ppat_loc  lbls lid in
        if List.mem n !labels then
          raise(Error(sp.ppat_loc,
                      Label_multiply_defined lid));
        labels := n :: !labels;
        let p = type_pat env pat in
        unify_pat env p ty;
        args.(n) <- p
        ) lid_sp_list;
      { pat_desc = Tpat_construct(constr, Array.to_list args);
        pat_loc = sp.ppat_loc;
        pat_type = ty_res }

let add_pattern_variables env =
  let pv = !pattern_variables in
  pattern_variables := [];
  List.fold_right
    (fun (id, (ty,vl)) env ->
      Env.add_value id {val_type = ty; val_kind = vl} env)
    pv env

let type_pattern env spat =
  pattern_variables := [];
  let pat = type_pat env spat in
  let new_env = add_pattern_variables env in
  (pat, new_env)

let type_pattern_list env spatl =
  pattern_variables := [];
  let patl = List.map (type_pat env) spatl in
  let new_env = add_pattern_variables env in
  (patl, new_env)

let rec iter_pattern f p =
  f p;
  match p.pat_desc with
    Tpat_any | Tpat_var _ | Tpat_constant _ ->
      ()
  | Tpat_alias (p, _) ->
      iter_pattern f p
  | Tpat_tuple pl ->
      List.iter (iter_pattern f) pl
  | Tpat_construct (_, pl) ->
      List.iter (iter_pattern f) pl
  | Tpat_record fl ->
      List.iter (fun (_, p) -> iter_pattern f p) fl
  | Tpat_or (p, p2) ->
      iter_pattern f p;
      iter_pattern f p2


(* Generalization criterion for expressions *)

let rec is_nonexpansive exp =
  match exp.exp_desc with
    Texp_ident(_,_) -> true
  | Texp_constant _ -> true
  | Texp_let(rec_flag, pat_exp_list, body) ->
      List.for_all (fun (pat, exp) -> is_nonexpansive exp) pat_exp_list &
      is_nonexpansive body
  | Texp_function _ -> true
  | Texp_tuple el ->
      List.for_all is_nonexpansive el
  | Texp_construct(_, el) ->
      List.for_all is_nonexpansive el
  | Texp_record lbl_exp_list ->
      List.for_all
        (fun (lbl, exp) -> lbl.lbl_mut = Immutable & is_nonexpansive exp)
        lbl_exp_list
  | Texp_field(exp, lbl) -> is_nonexpansive exp
  | Texp_array [] -> true
  | Texp_new _ -> true
  | _ -> false

(* Typing of printf formats *)

let type_format location fmt =
  let len = String.length fmt in
  let ty_input = newvar()
  and ty_result = newvar() in
  let rec skip_args j =
    if j >= len then j else
      match fmt.[j] with
        '0' .. '9' | ' ' | '.' | '-' -> skip_args (j+1)
      | _ -> j in
  let rec scan_format i =
    if i >= len then ty_result else
    match fmt.[i] with
      '%' ->
        let j = skip_args(i+1) in
        begin match String.unsafe_get fmt j with
        (* We're using unsafe_get here so that if j = String.length fmt,
           we'll fall in the catch-all case of the match *)
          '%' ->
            scan_format (j+1)
        | 's' ->
            newty (Tarrow(instance Predef.type_string, scan_format (j+1)))
        | 'c' ->
            newty (Tarrow(instance Predef.type_char, scan_format (j+1)))
        | 'd' | 'o' | 'x' | 'X' | 'u' ->
            newty (Tarrow(instance Predef.type_int, scan_format (j+1)))
        | 'f' | 'e' | 'E' | 'g' | 'G' ->
            newty (Tarrow(instance Predef.type_float, scan_format (j+1)))
        | 'b' ->
            newty (Tarrow(instance Predef.type_bool, scan_format (j+1)))
        | 'a' ->
            let ty_arg = newvar() in
            newty (Tarrow (newty (Tarrow(ty_input,
                                         newty (Tarrow (ty_arg, ty_result)))),
                           newty (Tarrow (ty_arg, scan_format (j+1)))))
        | 't' ->
            newty (Tarrow(newty (Tarrow(ty_input, ty_result)),
                          scan_format (j+1)))
        | c ->
            raise(Error(location, Bad_format(String.sub fmt i (j-i+1))))
        end
    | _ -> scan_format (i+1) in
  newty
    (Tconstr(Predef.path_format, [scan_format 0; ty_input; ty_result], ref Mnil))

(* Typing of expressions *)

let unify_exp env exp expected_ty =
  try
    unify env exp.exp_type expected_ty
  with Unify trace ->
    raise(Error(exp.exp_loc, Expr_type_clash(trace)))

(* Error report *)

open Format
open Printtyp

let report_error = function
    Unbound_value lid ->
      print_string "Unbound value "; longident lid
  | Unbound_constructor lid ->
      print_string "Unbound constructor "; longident lid
  | Unbound_label lid ->
      print_string "Unbound label "; longident lid
  | Constructor_arity_mismatch(lid, expected, provided) ->
      open_box 0;
      print_string "The constructor "; longident lid;
      print_space(); print_string "expects "; print_int expected;
      print_string " argument(s),"; print_space();
      print_string "but is here applied to "; print_int provided;
      print_string " argument(s)";
      close_box()
  | Label_mismatch(lid, trace) ->
      unification_error trace
        (function () ->
           print_string "The label "; longident lid;
           print_space(); print_string "belongs to the type")
        (function () ->
           print_string "but is here mixed with labels of type")
  | Pattern_type_clash trace ->
      unification_error trace
        (function () ->
           print_string "This pattern matches values of type")
        (function () ->
           print_string "but is here used to match values of type")
  | Multiply_bound_variable ->
      print_string "This variable is bound several times in this matching"
  | Orpat_not_closed ->
      print_string "A pattern with | must not bind variables"
  | Expr_type_clash trace ->
      unification_error trace
        (function () ->
           print_string "This expression has type")
        (function () ->
           print_string "but is here used with type")
  | Apply_non_function typ ->
      begin match (repr typ).desc with
        Tarrow(_, _) ->
          print_string "This function is applied to too many arguments"
      | _ ->
          print_string
            "This expression is not a function, it cannot be applied"
      end
  | Label_multiply_defined lid ->
      print_string "The label "; longident lid;
      print_string " is defined several times"
  | Label_missing ->
      print_string "Some labels are undefined"
  | Label_not_mutable lid ->
      print_string "The label "; longident lid;
      print_string " is not mutable"
  | Bad_format s ->
      print_string "Bad format `"; print_string s; print_string "'"
  | Undefined_method_err me ->
      print_string "This expression has no method ";
      print_string me
  | Unbound_class cl ->
      print_string "Unbound class "; longident cl
  | Virtual_class cl ->
      print_string "One cannot create instances of the virtual class ";
      longident cl
  | Unbound_instance_variable v ->
      print_string "Unbound instance variable ";
      print_string v
  | Instance_variable_not_mutable v ->
      print_string " The instance variable "; print_string v;
      print_string " is not mutable"
  | Not_subtype(tr1, tr2) ->
      reset ();
      List.iter
        (function (t, t') -> mark_loops t; if t != t' then mark_loops t')
        tr1;
      List.iter
        (function (t, t') -> mark_loops t; if t != t' then mark_loops t')
        tr2;
      trace true (fun _ -> print_string "is not a subtype of type") tr1;
      trace false (fun _ -> print_string "is not compatible with type") tr2
  | Outside_class ->
      print_string "Object duplication outside a class definition."
  | Value_multiply_overridden v ->
      print_string "The instance variable "; print_string v;
      print_string " is overridden several times"
  | Coercion_failure (ty, ty', trace) ->
      unification_error trace
        (function () ->
           mark_loops ty; if ty' != ty then mark_loops ty';
           print_string "This expression cannot be coerced to type";
           print_break 1 2;
           type_expansion ty ty';
           print_string ";";
           print_space ();
           print_string "it has type")
        (function () ->
           print_string "but is here used with type")
  | Too_many_arguments ->
      print_string "This function has too many arguments"
  | NewFailure s ->
      print_string s
  | SimpleTypeError (s,ty) ->
      print_string s;
      type_expr ty

(***********************************************************************)
(*                                                                     *)
(*                   Join-calculus extension                           *)
(*                                                                     *)
(*      Fabrice Le Fessant, projet SOR/PARA INRIA Rocquencourt         *)
(*                                                                     *)
(*                 mail: fabrice.le_fessant@inria.fr                   *)
(*                 web:  http://www-sor.inria.fr/~lefessan             *)
(*                                                                     *)
(*  Copyright 1997 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(*                                                                     *)
(*         File written in the Objective-CAML language                 *)
(***********************************************************************)

type channel =
    { 
      chan_ext_ident: Ident.t;
      chan_in_ident: Ident.t;
      chan_args: type_expr;
      chan_type: type_expr;
      mutable chan_types: type_expr list;
      chan_ord : int;
      chan_sync: reply_type
    }


type join = {
  mutable join_pattern: int list;
  mutable join_args: (channel * pattern) list;
  mutable join_vars: (Ident.t * (type_expr*value_kind)) list;
  mutable join_replies: (string * Ident.t) list;
  mutable join_id: Ident.t;
  mutable join_types: (type_expr list) list

}

let synthetic_loc = ref Location.none

let mkpat pat typ =
  { pat_desc = pat; pat_loc = !synthetic_loc; pat_type = typ }
let mkexp exp typ env =
  { exp_desc = exp; exp_loc = !synthetic_loc; exp_type = typ; exp_env = env }

let mkunit env =
  let lid =(Lident "()") in
  let constr =
    try
      Env.lookup_constructor lid env
    with Not_found ->
      raise(Error(!synthetic_loc, Unbound_constructor lid)) in
  let (ty_args, ty_res) = instance_constructor constr in
    { exp_desc = Texp_construct(constr, []);
      exp_loc = !synthetic_loc;
      exp_type = ty_res;
      exp_env = env }


let mkpexp exp =
  { pexp_desc = exp;
    pexp_loc = !synthetic_loc
  }

let mkppat pat =
  { ppat_desc = pat;
    ppat_loc = !synthetic_loc
  }

let rec mklistpexp = function
    [] ->
      mkpexp(Pexp_construct(Lident "[]", None, false))
  | e1 :: el ->
      mkpexp(Pexp_construct(Lident "::",
                           Some(mkpexp(Pexp_tuple[e1; mklistpexp el])),
                           false))

let mkmodpident modul ident =
  mkpexp (Pexp_ident (Ldot (Lident modul,ident)))

let mkmodty env modul ident =
  Typetexp.transl_simple_type env true
    { ptyp_desc = Ptyp_constr (Ldot (Lident modul,ident),[]);
      ptyp_loc = !synthetic_loc}
(*
let build_end processes channels =
  let pattern2tuple (join,process,thread) =
    mkpexp(Pexp_tuple [
	     mkpexp (Pexp_constant (Const_int join.join_pattern));
	     mkpexp (Pexp_ident (Lident ("#process#"^
                                         (string_of_int join.join_pattern))));
             mkpexp (Pexp_constant
                       (Const_int 
                          (match thread with
                              None -> 0 
                            | Some (name,reply_name) -> 
                                1+(List.assoc name channels).chan_ord)))
           ]
    )
  in
  let apply_Jfa = 
    mkpexp (Pexp_apply (mkmodpident "Join" "fill_automata",
                        [mkpexp (Pexp_ident (Lident "#automata#"));
                         mkpexp(Pexp_array (List.map pattern2tuple processes))
                        ]))
  in
  let tupleexp =
    mkpexp (Pexp_tuple (List.map (function (name,chan) -> 
                                    mkpexp (Pexp_ident (Lident name))
                        ) channels))
  in
    mkpexp (Pexp_sequence (apply_Jfa,tupleexp))
*)

let mkprocedure exp env =
    mkexp(
      Texp_function [
        type_pat env (mkppat (Ppat_construct (Lident "()",None,false)))
        ,exp]
    ) (mkmodty env "Join" "process") env

let mktuple list env =
  mkexp (Texp_tuple list)
    (newty (Ttuple (List.map (fun exp -> exp.exp_type) list)))
    env

let mkpattuple list =
  mkpat (Tpat_tuple list)
    (newty (Ttuple (List.map (fun pat -> pat.pat_type) list)))
    

let mkarray list env =
  mkexp (Texp_array list)
    (instance (Predef.type_array (List.hd list).exp_type))
    env

let mkint const env =
  mkexp (Texp_constant (Const_int const)) (instance Predef.type_int) env

let mksequence exp1 exp2 env =
  mkexp (Texp_sequence (exp1,exp2)) (instance exp2.exp_type) env
;;

let mklet decls exp env =
  mkexp (Texp_let(Nonrecursive,decls,exp)) exp.exp_type env

let join_type name env = mkmodty env "Join" name

let not_computed env e =
  let rec not_computed e = 
    match e.pexp_desc with
      Pexp_ident _ 
    | Pexp_constant _
    | Pexp_thismodule
    | Pexp_function _
    | Pexp_null         -> true
          
    | Pexp_match _
    | Pexp_try _
    | Pexp_new _
    | Pexp_setinstvar _
    | Pexp_override _
    | Pexp_process _ -> false
          
    | Pexp_field (e,_)        -> not_computed e
    | Pexp_constraint (e,_,_) -> not_computed e
    | Pexp_send (e,_)         -> not_computed e
    | Pexp_def (_,e)          -> not_computed e
    | Pexp_let (_,list,e)     -> not_computed e &&
        List.for_all (function (pat,e) -> not_computed e) list
    | Pexp_record list   -> 
        List.for_all (function (pat,e) -> not_computed e) list
    | Pexp_tuple list -> List.for_all not_computed list
    | Pexp_array list -> List.for_all not_computed list
          
    | Pexp_construct (_,option,_)     -> option_not_computed option
    | Pexp_ifthenelse (e1,e2,option)  -> option_not_computed option &&
        not_computed e1 && not_computed e2
    | Pexp_setfield (e1,_,e2) -> not_computed e1 && not_computed e2
    | Pexp_sequence (e1,e2)   -> not_computed e1 && not_computed e2
    | Pexp_while    (e1,e2)   -> not_computed e1 && not_computed e2
    | Pexp_when     (e1,e2)   -> not_computed e1 && not_computed e2
    | Pexp_for (_,e1,e2,_,e3) -> not_computed e1 && not_computed e2
          && not_computed e3
          
    | Pexp_call (channel,list) -> false

    | Pexp_apply (sexp,list)       -> 
        begin
          match sexp.pexp_desc with
          Pexp_ident lid ->
            begin try
              let (path, desc) = Env.lookup_value lid env in
              begin
                match desc.val_kind with
                  Val_prim prim ->
                    begin
                      match prim.prim_name with

                        "%identity" | "%field0" | "%field1" |
                        "%setfield0" | "%makeblock" | "%makemutable" |
                        "%sequand" | "%sequor" | "%boolnot" |
                        "%negint" | "%succint" | "%predint" |
                        "%addint" | "%subint" | "%mulint" | "%andint"
                        | "%orint" | "%xorint" | "%lslint" | "%lsrint"
                        | "%asrint" | "%eq" | "%noteq" | "%ltint" |
                        "%leint" | "%gtint" | "%geint" | "%incr" |
                        "%decr" | "%intoffloat" | "%floatofint" |
                        "%negfloat" | "%absfloat" | "%addfloat" |
                        "%subfloat" | "%mulfloat" | "%eqfloat" |
                        "%noteqfloat" | "%ltfloat" | "%lefloat" |
                        "%gtfloat" | "%gefloat" | "%string_length" |
                        "%string_unsafe_get" | "%string_unsafe_set" |
                        "%array_length" | "%array_unsafe_get" |
                        "%array_unsafe_set" | "%obj_size" |
                        "%obj_field" | "%obj_set_field" ->
                        if List.for_all not_computed list then
                            true
                        else
                          false

                      | _ -> false
                    end
                | _ -> false
              end
            with
              Not_found -> false
            end
        | _ -> false
      end
    | _ -> false

  and option_not_computed = function
      None -> true
    | Some e -> not_computed e
  in
  not_computed e

let ident_to_tident ident env =
  let path =Path.Pident ident in
  let desc = 
    try
      Env.find_value path env
    with Not_found ->
      failwith ("Can not find ident "^(Ident.name ident)) 
  in
  mkexp (Texp_ident (path,desc)) desc.val_type env

let check_multiple_define location idents =
  let chan_names = ref [] in
(*  let loc_names = ref []
  in *)
  let add_name names ident =
    let name = Ident.name ident in
    if List.mem name !names then
      raise(Error(location, NewFailure("Ident "^ name ^ " is defined several times")))
    else
      names := name :: !names
  in
  List.iter 
    (fun (loc_option,channels) ->
      List.iter (fun (name,ty) -> add_name chan_names name) channels;
(*      match loc_option with
        Some loc_ident -> add_name loc_names loc_ident
      | _ -> ()
*)            ) 
    idents
    
